@@ -1,4 +1,4 @@
-/*  $Id: syspred.pl,v 1.30 1997/02/10 12:56:23 jan Exp $
+/*  $Id: syspred.pl,v 1.34 1997/08/07 07:57:26 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     jan@swi.psy.uva.nl
@@ -50,9 +50,6 @@
 	, garbage_collect/0
 	, arithmetic_function/1
         , default_module/2
-	, save_program/1
-	, save_program/2
-	, save/1
 	, absolute_file_name/2
 	, absolute_file_name/3
 	, require/1
@@ -77,18 +74,19 @@ $map_bits(Pred, ?Name, Old, Old) :-		% ask a bit
 	call(Pred, Name, Bits),
 	Old /\ Bits > 0.
 
-$port_bit(     call, 2'00000001).
-$port_bit(     exit, 2'00000010).
-$port_bit(     fail, 2'00000100).
-$port_bit(     redo, 2'00001000).
-$port_bit(    unify, 2'00010000).
-$port_bit(    break, 2'00100000).
-$port_bit( cut_call, 2'01000000).
-$port_bit( cut_exit, 2'10000000).
-$port_bit(      cut, 2'11000000).
-$port_bit(      all, 2'00111111).
-$port_bit(     full, 2'00101111).
-$port_bit(     half, 2'00101101).
+$port_bit(      call, 2'000000001).
+$port_bit(      exit, 2'000000010).
+$port_bit(      fail, 2'000000100).
+$port_bit(      redo, 2'000001000).
+$port_bit(     unify, 2'000010000).
+$port_bit(     break, 2'000100000).
+$port_bit(  cut_call, 2'001000000).
+$port_bit(  cut_exit, 2'010000000).
+$port_bit( exception, 2'100000000).
+$port_bit(       cut, 2'011000000).
+$port_bit(       all, 2'000111111).
+$port_bit(      full, 2'000101111).
+$port_bit(      half, 2'000101101).
 
 leash(Ports) :-
 	$leash(Old, Old),
@@ -552,16 +550,21 @@ statistics :-
 
 	format('~2f seconds cpu time for ~D inferences~n',
 				    [Cputime, Inferences]),
-	format('~D atoms, ~D functors, ~D predicates, ~D modules, ~D byte-codes~n~n',
+	format('~D atoms, ~D functors, ~D predicates, ~D modules, ~D VM-codes~n~n',
 				    [Atoms, Functors, Predicates, Modules, Codes]),
-	format('                      Limit    Allocated       In use~n'),
-	format('Heap         :                  ~t~D~53| Bytes~n',
-	       [Heapused]),
-	format('Local  stack :~t~D~27| ~t~D~40| ~t~D~53| Bytes~n',
+	format('                       Limit    Allocated       In use~n'),
+	(   statistics(heap, Heap),
+	    statistics(heaplimit, HeapLimit)
+	->  format('Heap         :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
+		   [HeapLimit, Heap, Heapused])
+	;   format('Heap         :                  ~t~D~54| Bytes~n',
+		   [Heapused])
+	),
+	format('Local  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
 	       [LocalLimit, Local, LocalUsed]),
-	format('Global stack :~t~D~27| ~t~D~40| ~t~D~53| Bytes~n',
+	format('Global stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
 	       [GlobalLimit, Global, GlobalUsed]),
-	format('Trail  stack :~t~D~27| ~t~D~40| ~t~D~53| Bytes~n',
+	format('Trail  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
 	       [TrailLimit, Trail, TrailUsed]),
 
 	gc_statistics,
@@ -707,22 +710,6 @@ absolute_file_name(Term, Abs) :-
 garbage_collect :-
 	$garbage_collect(0).
 
-%	save_program(+Name[, +Options])
-%	Save the currently running image in file
-
-save_program(Name) :-
-	save_program(Name, []).
-
-save_program(Name, Options) :-
-	$autoload:clear_library_index,
-	$save_program(Name, Options).
-
-%	save(+File)
-%	Create saved-state in file
-
-save(File) :-
-	save(File, _).
-
 %	arithmetic_function(Spec)
 %	Register a predicate as an arithmetic function.  Takes Name/Arity
 %	and a term as argument.
