@@ -1,4 +1,4 @@
-/*  $Id: toplevel.pl,v 1.29 1997/03/10 09:36:09 jan Exp $
+/*  $Id: toplevel.pl,v 1.34 1997/11/04 10:38:22 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     jan@swi.psy.uva.nl
@@ -35,7 +35,7 @@ $welcome :-
 	Patch is Version mod 100,
 	$ttyformat('Welcome to SWI-Prolog (Version ~w.~w.~w)~n',
 		   [Major, Minor, Patch]),
-	$ttyformat('Copyright (c) 1993-1996 University of Amsterdam.  '),
+	$ttyformat('Copyright (c) 1993-1997 University of Amsterdam.  '),
 	$ttyformat('All rights reserved.~n~n'),
 	$ttyformat('For help, use ?- help(Topic). or ?- apropos(Word).~n~n').
 
@@ -65,15 +65,6 @@ $load_system_init_file :-
 	).
 $load_system_init_file.
 
-$check_novice :-
-	$novice(on, on), 
-	getenv('PROLOGCHILD', _), !, 
-	format('Cannot start Prolog from a child process running under Prolog~n'), 
-	format('Please type Control-D or `exit'' to return to Prolog~n'), 
-	halt.
-$check_novice.
-
-
 $load_gnu_emacs_interface :-
 	getenv('EMACS', t),
 	$argv(Args),
@@ -215,7 +206,6 @@ $load_associated_file.
 :- flag($qid, _, 1).
 
 $initialise :-
-	$check_novice, 
 	$clean_history,
 	$set_file_search_paths,
 	$run_at_initialization,
@@ -281,9 +271,7 @@ prolog :-
 		$system_prompt(TypeIn, BreakLev, Prompt),
 		prompt(Old, '|    '), 
 		trim_stacks,
-		read_history(h, '!h', 
-			      [trace, end_of_file], 
-			      Prompt, Goal, Bindings), 
+		read_query(Prompt, Goal, Bindings),
 		prompt(_, Old),
 		call_expand_query(Goal, ExpandedGoal,
 				  Bindings, ExpandedBindings)
@@ -291,6 +279,55 @@ prolog :-
 	    ), !.
 
 
+read_query(Prompt, Goal, Bindings) :-
+	feature(history, N),
+	N =< 0, !,
+	remove_history_prompt(Prompt, Prompt1),
+	repeat,				% over syntax errors
+	prompt1(Prompt1),
+	(   feature(readline, true)
+	->  $raw_read(user_input, Line),
+	    atom_chars(Line, LineChars),
+	    append(LineChars, ".", CompleteLine),
+	    call(rl_add_history(CompleteLine)),
+	    $term_to_atom(Goal, Line, Bindings, 1)
+	;   read_term(user_input, Goal, [variable_names(Bindings)])
+	), !.
+read_query(Prompt, Goal, Bindings) :-
+	seeing(Old), see(user_input),
+	(   read_history(h, '!h', 
+			 [trace, end_of_file], 
+			 Prompt, Goal, Bindings)
+	->  see(Old)
+	;   see(Old),
+	    fail
+	).
+
+remove_history_prompt(Prompt0, Prompt) :-
+	atom_chars(Prompt0, Chars0),
+	clean_history_prompt_chars(Chars0, Chars1),
+	delete_leading_blanks(Chars1, Chars),
+	atom_chars(Prompt, Chars).
+
+clean_history_prompt_chars([], []).
+clean_history_prompt_chars([0'%, 0'!|T], T) :- !.
+clean_history_prompt_chars([H|T0], [H|T]) :-
+	clean_history_prompt_chars(T0, T).
+ 
+delete_leading_blanks([32|T0], T) :- !,
+	delete_leading_blanks(T0, T).
+delete_leading_blanks(L, L).
+
+
+set_default_history :-
+	(   feature(readline, true)
+	->  set_feature(history, 0)
+	;   set_feature(history, 15)
+	).
+
+:- initialization set_default_history.
+
+
 		/********************************
 		*            PROMPTING		*
 		********************************/
@@ -301,7 +338,7 @@ prolog :-
 $prompt("%m%l%! ?- ").
 
 $set_prompt(P) :-
-	name(P, S),
+	atom_chars(P, S),
 	retractall($prompt(_)),
 	assert($prompt(S)).
 
@@ -316,7 +353,7 @@ $system_prompt(Module, BrekLev, Prompt) :-
 	->   $substitute("%l", ["[", BrekLev, "] "], P1, P2)
 	;    $substitute("%l", [], P1, P2)
 	),
-	name(Prompt, P2).
+	atom_chars(Prompt, P2).
 	
 $substitute(From, T, Old, New) :-
 	convert_to(T, T0),
@@ -495,6 +532,14 @@ $time_call(Goal, yes) :-
 	Goal, !.
 $time_call(_Goal, no).
 
+unhandled_exception(false, Term) :- !,
+        $warning('Unhandled exception'),
+	print_message(error, Term),
+	$ttyformat('~nNo~n').
+unhandled_exception(true, _Term) :-
+	$warning('Unhandled exception'),
+	$ttyformat('~nNo~n').
+
 
 		 /*******************************
 		 *	    EXPANSION		*
