@@ -1,4 +1,4 @@
-/*  $Id: latex2html.pl,v 1.1 1997/01/20 16:36:17 jan Exp $
+/*  $Id: latex2html.pl,v 1.8 1997/07/11 08:29:26 jan Exp $
 
     Designed and implemented by Jan Wielemaker
     E-mail: jan@swi.psy.uva.nl
@@ -20,21 +20,35 @@
 	    translate_environment/4,	% +In, +ModeIn, -ModeOut, -HTML
 	    translate_reference/4,	% +Kind, +RefPrefix, +Label, -HTML
 	    translate_footnote/2,	% +Text, -HTML
+	    translate_table/3,		% +Format, +BodyTokens, -HTML
 	    current_setting/1,		% +Type(-Value ...)
 	    do_float/2,			% +Float(+Number), :Goal
 	    tex_load_commands/1,	% +BaseName
+	    add_to_index/1,		% +Term
 	    add_to_index/2,		% +Term, +Tag
-	    clean_tt/2			% +Raw, -Clean
+	    clean_tt/2,			% +Raw, -Clean
+	    downcase_atom/2,		% +Atom, -LowerCase
+	    upcase_atom/2		% +Atom, -UpperCase
 	  ]).
 :- use_module(library(quintus)).
 
-version('0.5').
+version('0.10').
 
 :- dynamic			
 	html_output_dir/1,		% output relative to this dir
-	html_file_base/1,		% Basename of the main file
+	tex_file_base/1,		% Basename of the main file
+	html_file_base/1,		% Basename of main output file
+	html_split_level/1,		% Split upto this level
 	title/1,			% \title{} storage
-	author/1.
+	author/1.			% \auther{} command storage
+:- discontiguous
+	cmd/2,
+	cmd/3,
+	cmd/4.
+
+
+html_split_level(2).
+html_file_base('Title').
 
 :- multifile
 	user:file_search_path/2.
@@ -59,10 +73,15 @@ read_tex_inputs :-
 	getenv('TEXINPUTS', Val), !,
 	split(Val, ":", PathElement),
 	retractall(user:file_search_path(tex, _)),
-	forall(member(E, PathElement),
-	       asserta(user:file_search_path(tex, E))).
+	reverse(PathElement, RevPath),
+	forall(member(E, RevPath), assert_tex_input(E)).
 read_tex_inputs.
 
+assert_tex_input('') :- !,
+	asserta(user:file_search_path(tex, '.')).
+assert_tex_input(Dir) :-
+	asserta(user:file_search_path(tex, Dir)).
+
 		 /*******************************
 		 *       EXTENSION MODULES	*
 		 *******************************/
@@ -81,7 +100,8 @@ latex2html_module :-
 	->  true
 	;   asserta(tex:tex_extension_module(M))
 	),
-	M:dynamic((cmd/2, cmd/3, cmd/4, env/2, (#)/2)).
+	M:dynamic((cmd/2, cmd/3, cmd/4, env/2, (#)/2)),
+	M:discontiguous((cmd/2, cmd/3, cmd/4, env/2, (#)/2)).
 
 %	Load a tex command file
 
@@ -117,6 +137,7 @@ current_setting(html_output_dir(Dir)) :-
 
 keep_figures(true).
 onefile(false).
+makeindex(false).
 title('No title').
 author('Anonymous').
 
@@ -133,6 +154,7 @@ latex2html(TeXFile, Base) :-
 	reset_index,
 	reset_labels,
 	reset_cites,
+	reset_output,
 	tex_tokens(TeXFile, TeXTokens),
 	(   member(env(document, _, _), TeXTokens)
 	->  translate(TeXTokens, preamble, HTML)
@@ -165,27 +187,45 @@ latex2html(Spec) :-
 			   ],
 			   TheTeXFile),
 	asserta(html_output_dir(Base)),
-	asserta(html_file_base(Base)),
+	asserta(tex_file_base(Base)),
 	(   exists_directory(Base)
 	->  true
 	;   sformat(Cmd, 'mkdir ~w~n', [Base]),
 	    shell(Cmd)
 	),
-	latex2html(TheTeXFile, Base),
+	html_file_base(Main),
+	latex2html(TheTeXFile, Main),
 	goodbye,
-	retract(html_file_base(_)),
+	retract(tex_file_base(_)),
 	retract(html_output_dir(_)).
 
 
+:- dynamic
+	current_html_output_db/1,	% Base
+	in_head/1.
+
+reset_output :-
+	retractall(current_html_output_db(_)),
+	retractall(in_head(_)).
+
 open_output(Base) :-
 	html_output_dir(Dir), !,
 	concat_atom([Dir, /, Base, '.html'], HtmlFile),
-	tex_tell(HtmlFile).
-
+	tex_tell(HtmlFile),
+	put_html_token(html('<HTML>')),
+	asserta(current_html_output_db(Base)).
 
 close_output :-
+	(   retract(current_html_output_db(_))
+	->  put_html_token(html('</BODY>')),
+	    put_html_token(html('</HTML>'))
+	;   true
+	),
 	tex_told.
 
+current_html_output(Raw) :-
+	current_html_output_db(File), !,
+	File = Raw.
 
 welcome :-
 	version(Version),
@@ -193,11 +233,11 @@ welcome :-
 
 goodbye :-
 	html_output_dir(Dir),
-	html_file_base(Base),
+	html_file_base(Html),
 	format(user_error, '~*t~72|~n', [0'*]),
 	format(user_error,
 	       'Tranlation completed; output written to "~w/~w.html".~n',
-	       [Dir, Base]),
+	       [Dir, Html]),
 	format(user_error, 'Prolog statistics:~n~n', []),
 	statistics.
 
@@ -245,6 +285,9 @@ translate_2(verbatim(_Cmd, Text), Mode, Mode, 		% \begin{verbatim} ...
 translate_2(verb(_, Text), pcecode, pcecode,
 	    [verb('\n'), verb(Text)]).
 translate_2(vskip(_), pcecode, pcecode, verb('\n')).
+translate_2('\n', Mode, Mode, [html('<BR>')]) :-
+	Mode = group(Atts),
+	memberchk(obeylines, Atts), !.
 :- set_feature(character_escapes, false).
 translate_2(verb(_, Text), Mode, Mode, #code(Text)).	% \verbX...X
 translate_2([Atom], Mode, Mode, nospace(Atom)) :-	% {foo} 
@@ -255,14 +298,16 @@ translate_2(Group, Mode, Mode, HTML) :-			% {...}
 translate_2(~, Mode, Mode, html('&nbsp;')).		% ~
 translate_2(Atom0, Mode, Mode, Atom) :-			% Normal word
 	atomic(Atom0),
-	(   Mode = group(sc)
+	(   Mode = group(Atts),
+	    memberchk(font(sc), Atts)
 	->  upcase_atom(Atom0, Atom)
 	;   Atom = Atom0
 	).
 
 translate_group(Group, [HTML, Close]) :-		% {...}
-	translate(Group, group(-), MEnd, HTML),
-	(   MEnd = group(EndFont),
+	translate(Group, group([]), MEnd, HTML),
+	(   MEnd = group(Attributes),
+	    member(font(EndFont), Attributes),
 	    html_font(EndFont, _, Close)
 	->  true
 	;   Close = []
@@ -321,6 +366,10 @@ language_map(table,	'Table').
 
 #(tell(_File),		[]) :- onefile(true).
 #(tell(File),		tell(File)).
+#(head(Head),		[html('<HEAD>'), HtmlHead, html('</HEAD>')]) :-
+	expand_macros(Head, HtmlHead).
+#(beginbody,		html('<BODY>')).
+#(endbody,		html('</BODY>')).
 #(thetitle,		Title) :-
 	title(Title).
 #(theauthor,		Author) :-
@@ -391,6 +440,13 @@ language_map(table,	'Table').
 	node_header(Tag, HTML).
 #(footer(Tag),		HTML) :-
 	node_footer(Tag, HTML).
+#(next_and_prev_references,
+        [ #iflref(prevfile,  	    '[Previous]'), ' ',
+	  #iflref(nextfile,  	    '[Next]')
+	]).
+#(home_reference,
+ 	[ #iflref(home,  	    '[Home]')
+	]).
 #(Macro, []) :-
 	functor(Macro, Name, Arity),
 	format(user_error,
@@ -499,13 +555,19 @@ do_float(Term, Goal) :-
 		 *******************************/
 
 :- dynamic
-	label/3.			% Label, File, Index
+	label/3,			% Label, File, Index
+	next_file/2.			% File, NextFile
 
 reset_labels :-
-	retractall(label(_, _, _)).
+	retractall(label(_, _, _)),
+	retractall(next_file(_, _)).
 
 collect_labels([], _) :- !.
-collect_labels([tell(File)|T], _) :- !,
+collect_labels([tell(File)|T], PreviousFile) :- !,
+	(   PreviousFile \== File
+	->  assert(next_file(PreviousFile, File))
+	;   true
+	),
 	collect_labels(T, File).
 collect_labels([label(Label, _, Tag)|T], File) :- !,
 	assert(label(Label, File, Tag)),
@@ -534,7 +596,7 @@ label_tag(Tag) :-
 %
 %	Translate an environment.
 
-env(pcecode(_, Tokens), #xmp(HTML)) :-
+env(pcecode(_, Tokens), #pre(HTML)) :-
 	translate(Tokens, pcecode, HTML).
 env(summarylist(_, Summary),
     [ html('<TABLE>'),
@@ -546,8 +608,9 @@ env(htmlonly(_, Tokens), HTML) :-
 env(document(_, Contents), HTML) :- !,
 	translate(Contents, document, HTML).
 env(quote(_, Tokens), #quote(Quote)) :-
-	translate(Tokens, normal, Quote).
-env(abstract(_, Tokens), #quote(Quote)) :-
+	translate_group(Tokens, Quote).
+%env(abstract(_, Tokens), [html('<HR>'), #quote(Quote), html('<HR>')]) :-
+env(abstract(_, Tokens), #quote([html('<HR>'), Quote, html('<HR>')])) :-
 	translate(Tokens, normal, Quote).
 env(center(_, Tokens), #center(Center)) :-
 	translate(Tokens, normal, Center).
@@ -660,7 +723,8 @@ translate_items([H0|T0], List, [H1|T1]) :-
 		 *******************************/
 
 prolog_function(\(usepackage, [_,{File},_])) :-
-	(   absolute_file_name(tex(File),
+	(   member(Term, [tex(File), library(File)]),
+	    absolute_file_name(Term,
 			       [ extensions([pl, qlf]),
 				 access(read)
 			       ],
@@ -668,6 +732,14 @@ prolog_function(\(usepackage, [_,{File},_])) :-
 	->  ensure_loaded(user:PlFile)
 	;   true
 	).
+prolog_function(\(newcommand, [{Name}, [], {Expanded}])) :-
+	declare_command(Name, 0, Expanded).
+prolog_function(\(newcommand, [{Name}, [Args], {Expanded}])) :-
+	declare_command(Name, Args, Expanded).
+prolog_function(\(renewcommand, [{Name}, [], {Expanded}])) :-
+	declare_command(Name, 0, Expanded).
+prolog_function(\(renewcommand, [{Name}, [Args], {Expanded}])) :-
+	declare_command(Name, Args, Expanded).
 	
 
 		 /*******************************
@@ -681,7 +753,7 @@ prolog_function(\(usepackage, [_,{File},_])) :-
 cmd(onefile, preamble, []) :-
 	retractall(onefile(_)),
 	assert(onefile(true)).
-cmd(htmlpackage({File}), []) :-
+cmd(htmlpackage({File}), preamble, []) :-
 	(   absolute_file_name(tex(File),
 			       [ extensions([pl, qlf]),
 				 access(read)
@@ -692,16 +764,9 @@ cmd(htmlpackage({File}), []) :-
 	).
 cmd(documentclass(_, _), preamble, []).
 cmd(usepackage(_, {_File}, _), preamble, []) :- !.
-/*	(   absolute_file_name(tex(File),
-			       [ extensions([pl, qlf]),
-				 access(read)
-			       ],
-			       PlFile)
-	->  ensure_loaded(user:PlFile)
-	;   true
-	).
-*/
-cmd(makeindex, preamble, []).
+cmd(makeindex, preamble, []) :-
+	retractall(makeindex(_)),
+	asserta(makeindex(true)).
 cmd(newcommand({Name}, [], {Expanded}), []) :-
 	declare_command(Name, 0, Expanded).
 cmd(newcommand({Name}, [Args], {Expanded}), []) :-
@@ -713,21 +778,30 @@ cmd(renewcommand({Name}, [Args], {Expanded}), []) :-
 cmd(def({'\booktitle'}, {Title}), HTML) :-
 	cmd(title({[Title]}), HTML).
 cmd(def(_, _), []).
+cmd(sloppy, []).
 cmd(noindent, []).
 cmd(clearpage, []).
 cmd(cleardoublepage, []).
 cmd(nopagebreak, []).
 cmd(pagebreak, []).
+cmd(linebreak, [html('<BR>')]).
 cmd(newpage, []).
 cmd(hyphenpenalty(_), []).
 cmd(newlength(_), []).
+cmd(setlength(_,_), []).
 cmd(settowidth(_, _), []).
 cmd(setcounter({page}, _), []).
 cmd(vskip(_), [html('<P>')]).
 cmd(vspace(_), [html('<P>')]).
-cmd(vfill, []).
+cmd(hspace(_), []).
+cmd(headheight(_), []).
+cmd(footheight(_), []).
+cmd(vfill, [html('<P>')]).
+cmd(vfil, [html('<P>')]).
+cmd(hfill, []).
 cmd(parskip(_), []).
 cmd(parindent(_), []).
+cmd(raggedright, []).			% Always in HTML
 cmd(tableofcontents,
     [ #tell('Contents'),
       #header,
@@ -757,6 +831,7 @@ cmd(subparagraph({Title}), [#b(+Title), ' ']).
 cmd(label({Label}), #label(Label, [], Tag)) :-	% \label and \xyzref
 	label_tag(Tag).
 cmd(ref({RefName}), #lref(RefName, ref(RefName))).
+cmd(pageref({RefName}), #lref(RefName, ref(RefName))).
 	
 cmd(secref({Label}), HTML) :-
 	translate_reference(section, sec, Label, HTML).
@@ -783,18 +858,16 @@ cmd(title({Title}), []) :-			% \title
 	retractall(title(_)),
 	translate(Title, normal, HTML),
 	assert(title(HTML)).
-cmd(booktitle({Title}), HTML) :-		% \booktitle
-	cmd(title({Title}), HTML).
 cmd(author({Author}), []) :-			% \author
 	retractall(author(_)),
 	translate(Author, normal, HTML),
 	assert(author(HTML)).
 cmd(maketitle,					% \maketitle
-    [ #center(#h(1, #thetitle)),
-      html('<HR>'),
-      #center(#i(#theauthor)),
-      html('<HR>')
-    ]).
+    #quote(#quote(#quote(#quote([ #center(#h(1, #thetitle)),
+				  html('<HR>'),
+				  #center(#i(#theauthor)),
+				  html('<HR>')
+				]))))).
 
 cmd(newblock, []).				% BiBTeX \newblock
 cmd(protect, []).				% BiBTeX produced?
@@ -808,7 +881,7 @@ cmd(cite({Key}),	#cite(Key)).		% \cite
 cmd(yearcite({Key}),	#yearcite(Key)).	% \yearcite
 cmd(opencite({Key}),	#opencite(Key)).	% \opencite
 cmd(bibliography({_}), HTML) :-			% \bibliography
-	html_file_base(File),
+	tex_file_base(File),
 	(   absolute_file_name(tex(File), [ extensions([bbl]),
 					    access(read)
 					  ],
@@ -869,6 +942,8 @@ cmd(mbox({Boxed}), HTML) :-
 	translate_group(Boxed, HTML).
 cmd(makebox(_, _, {Boxed}), HTML) :-
 	translate_group(Boxed, HTML).
+cmd(parbox(_, _, {Boxed}), HTML) :-
+	translate_group(Boxed, HTML).
 cmd(string({Text}), nospace(Text)).
 cmd(ldots, '...').
 cmd(cline(_), []).
@@ -924,7 +999,8 @@ cmd(psdirectories({Spec}), []) :-
 cmd(psfig({Spec}), html(Img)) :-
 	psfig_options(Spec, Options),
 	member(figure(File), Options),
-	file_name_extension(Base, ps, File),
+	file_name_extension(Base, Ext, File),
+	ps_extension(Ext),
 	file_base_name(Base, GifBase),
 	file_name_extension(GifBase, gif, GifFile),
 	sformat(Img, '<IMG SRC="~w">', GifFile),
@@ -978,6 +1054,9 @@ cmd(postscriptfig(_Options, {File}, Title),
 	;   ps2gif(psfig(File), OutFile, [margin(5)])
 	).
 
+ps_extension(eps).
+ps_extension(ps).
+
 %
 %	HTML documentation
 %
@@ -1038,28 +1117,31 @@ cmd(pow({A1}, {A2}), math, [math+A1, '**', +A2]).
 
 cmd(Cmd, Mode0, Mode, HTML) :-
 	user_cmd(Cmd, Mode0, Mode, HTML), !.
-cmd(Font, group(Old), group(Font), HTML) :-
+cmd(obeylines, group(Atts), group([obeylines|Atts]), []).
+cmd(Font, group(Old), group([font(Font)|Old1]), HTML) :-
 	html_font(Font, Open, _), !,
-	(   html_font(Old, _, Close)
+	(   delete(Old, font(OldFont), Old1),
+	    html_font(OldFont, _, Close)
 	->  HTML = [Close,Open]
-	;   HTML = Open
+	;   Old1 = Old,
+	    HTML = Open
 	).
 
-html_font(em, html('<EM>'), html('</EM>')).
-html_font(bf, html('<B>'),  html('</B>')).
-html_font(it, html('<I>'),  html('</I>')).
-html_font(tt, html('<TT>'), html('</TT>')).
-html_font(sf, html('<B>'),  html('</B>')).
-html_font(sc, [],           []).
-html_font(rm, [],           []).
-html_font(sl, html('<B>'),  html('</B>')).
-html_font(scriptsize,   [], []).
-html_font(footnotesize, [], []).
-html_font(small,        [], []).
-html_font(normalsize,   [], []).
-html_font(large,        [], []).
-html_font('Large',      [], []).
-html_font('Huge',       [], []).
+html_font(em, 		html('<EM>'), 		html('</EM>')).
+html_font(bf, 		html('<B>'),  		html('</B>')).
+html_font(it, 		html('<I>'),  		html('</I>')).
+html_font(tt, 		html('<TT>'), 		html('</TT>')).
+html_font(sf, 		html('<B>'),  		html('</B>')).
+html_font(sc, 		html('<font size=-1>'), html('</font>')).
+html_font(rm, 		[],           		[]).
+html_font(sl, 		html('<B>'),  		html('</B>')).
+html_font(scriptsize,   [], 			[]).
+html_font(footnotesize, html('<font size=-1>'), html('</font>')).
+html_font(small,        html('<font size=-1>'), html('</font>')).
+html_font(normalsize,   [],			[]).
+html_font(large,        html('<font size=+1>'), html('</font>')).
+html_font('Large',      html('<font size=+2>'), html('</font>')).
+html_font('Huge',       html('<font size=+3>'), html('</font>')).
 
 
 		 /*******************************
@@ -1143,6 +1225,11 @@ downcase_atom(Up, Low) :-
 	maplist(downcase, S0, S1),
 	atom_chars(Low, S1).
 
+capitalise_atom(In, Out) :-
+	atom_chars(In, S0),
+	capitalise(S0, S1, up),
+	atom_chars(Out, S1).
+
 upcase(L, U) :-
 	between(0'a, 0'z, L), !,
 	U is L + 0'A - 0'a.
@@ -1153,9 +1240,27 @@ downcase(L, U) :-
 	U is L + 0'a - 0'A.
 downcase(C, C).
 
+capitalise([], [], _).
+capitalise([H0|T0], [H|T], up) :-
+	islower(H0), !,
+	upcase(H0, H),
+	capitalise(T0, T, down).
+capitalise([H|T0], [H|T], up) :-
+	capitalise(T0, T, up).
+capitalise([H|T0], [H|T], down) :-
+	isletter(H), !,
+	capitalise(T0, T, down).
+capitalise([H|T0], [H|T], down) :-
+	capitalise(T0, T, up).
+	
 isletter(C) :- between(0'a, 0'z, C), !.
 isletter(C) :- between(0'A, 0'Z, C), !.
-
+isdigit(C)  :- between(0'0, 0'9, C), !.
+islower(C)  :- between(0'a, 0'z, C), !.
+isspace(32).
+isspace(9).
+isspace(10).
+isspace(13).
 
 		 /*******************************
 		 *	    FOOTNOTES		*
@@ -1220,18 +1325,26 @@ appendix :-
 
 translate_section(Level, -, TeXTitle,
 	[ Footer,
-	  #tell(NodeFile),
-	  #header(Tag),
+	  Tell,
+	  Header,
 	  #h(Level, #label(RefName, [Tag, ' ', Title]))
 	]) :- !,
 	translate(TeXTitle, normal, Title),
 	section_tag(OldTag),
 	increment_section_counter(Level, Tag, FirstSubSection),
-	(   FirstSubSection
-	->  Footer = #footer(OldTag)
-	;   Footer = []
+	(   html_split_level(Split),
+	    Level =< Split
+	->  Tell = #tell(NodeFile),
+	    Header = #header(Tag),
+	    section_file(Tag, Title, NodeFile),
+	    (   FirstSubSection
+	    ->  Footer = #footer(OldTag)
+	    ;   Footer = []
+	    )
+	;   Tell = [],
+	    Footer = [],
+	    Header = []
 	),
-        section_file(Tag, Title, NodeFile),
 	sformat(RefName, 'sec:~w', [Tag]),
 	assert(section(Level, Tag, Title)).
 translate_section(Level, *, Title, #h(Level, Title)).
@@ -1317,9 +1430,14 @@ tableofcontents(TagPrefix, [Sections, CloseUL]) :-
 section_html(TagPrefix,
 	     [ FixLevel,
 	       html('<LI>'),
-	       #lref(fileof(RefName), [Open, Tag, ' ', Title, Close])
+	       #lref(Ref, [Open, Tag, ' ', Title, Close])
 	     ]) :-
 	section(Level, Tag, Title),
+	html_split_level(Split),
+	(   Level =< Split
+	->  Ref = fileof(RefName)
+	;   Ref = RefName
+	),
 	concat(TagPrefix, _, Tag),
 	fix_level(Level, FixLevel),
 	sformat(RefName, 'sec:~w', [Tag]),
@@ -1390,6 +1508,8 @@ get_lower_letters([_|T0], T) :-
 	get_lower_letters(T0, T).
 
 
+make_index([]) :-
+	makeindex(false), !.
 make_index([]) :-
 	\+ index(_, _, _), !.
 make_index(HTML) :-
@@ -1413,13 +1533,15 @@ index_html([SortKey|T0], CL0, [Sep, TermHTML|TH]) :-
 	index_html(T0, CL, TH).
 	
 index_terms([], []).
-index_terms([Term|T0], [ html('<DT>'),
+index_terms([Term0|T0], [ html('<DT>'),
 			 HtmlTerm,
 			 html('<DD>'),
 			 Where
 		       | TH
 		       ]) :-
-	findall(Tag, index(_, Term, Tag), Tags),
+	findall(Tag, index(_, Term0, Tag), Tags),
+	tex_atom_to_tokens(Term0, Tokens),
+	translate(Tokens, index, _, Term),
 	(   member(+PrimeTag, Tags)
 	->  HtmlTerm = #lref(PrimeTag, Term)
 	;   HtmlTerm = Term
@@ -1447,25 +1569,51 @@ add_separator(Term, _, CL, [ html('<DT>'),
 		 *        STANDARD LINKS	*
 		 *******************************/
 
-node_header([]) :-
+node_header([#head(#title(#thetitle)),
+	     #beginbody
+	    ]) :-
 	onefile(true), !.
-node_header([#title(#thetitle),
+node_header([#head([#title(#thetitle),
+		    link(home),
+		    link(contents),
+		    link(index),
+		    link(summary),
+		    link(previous),
+		    link(next)]),
+	     #beginbody,
 	     html('<HR>'),
-	     #center( [#iflref(fileof('document-contents'), '[Contents]'), ' ',
-		       #iflref(fileof('document-index'),    '[Index]'), ' ',
-		       #iflref(fileof('sec:summary'),  	    '[Summary]')]),
+	     #center( [ body_link(home),
+			body_link(contents),
+			body_link(index),
+			body_link(summary),
+			body_link(previous),
+			body_link(next)
+		      ]),
 	     html('<HR>')
 	    ]).
 	
 node_header(_, []) :-
 	onefile(true), !.
 node_header(SectionTag,
-	    [#title([#thetitle, nospace(:), ' ', 'Section', SectionTag]),
+	    [#head([#title([#thetitle, nospace(:), ' ',
+			    'Section', SectionTag]),
+		    link(home),
+		    link(contents),
+		    link(index),
+		    link(summary),
+		    link(up(UpRef)),
+		    link(previous),
+		    link(next)]),
+	     #beginbody,
 	     html('<HR>'),
-	     #center( [#iflref(fileof(UpRef),	    	    '[Up]'), ' ',
-		       #iflref(fileof('document-contents'), '[Contents]'), ' ',
-		       #iflref(fileof('document-index'),    '[Index]'), ' ',
-		       #iflref(fileof('sec:summary'),  	    '[Summary]')]),
+	     #center( [ body_link(home),
+			body_link(contents),
+			body_link(index),
+			body_link(summary),
+			body_link(up(UpRef)),
+			body_link(previous),
+			body_link(next)
+		      ]),
 	     html('<HR>')
 	    ]) :-
 	parent_tag(SectionTag, UpTag), !,
@@ -1529,24 +1677,28 @@ translate_table(Format, Body, HTML) :-
 	atom_chars(Format, Fmt),
 	table_frame(Fmt, Body, FrameAttributes, Fmt2, Body2),
 	expand_table_commands(Body2, Body3),
-	(   table_columns(Fmt2, Ncols, ColHTML)
+	(   table_columns(Fmt2, Ncols, ColAtts)
 	->  true
 	;   format(user_error, 'Failed to parse tabular spec "~s"~n', [Fmt2]),
-	    ColHTML = [],
+	    ColAtts = [],
 	    Ncols = 0
 	),
-	(   table_body(Body3, BodyHTML)
+	(   table_body(Body3, ColAtts, BodyHTML)
 	->  true
 	;   format(user_error, 'Failed to translate table body~n', []),
 	    trace, fail
 	),
 	HTML = [ html(Head),
-		 ColHTML,
+%		 ColHTML,
 		 BodyHTML,
 		 html('</TABLE>')
 	       ],
-	sformat(Head, '<TABLE COLS=~d FRAME=~w RULES=group>',
-		[Ncols, FrameAttributes]).
+	(   FrameAttributes == void
+	->  Border = 0
+	;   Border = 2
+	),
+	sformat(Head, '<TABLE BORDER=~d COLS=~d FRAME=~w RULES=group>',
+		[Border, Ncols, FrameAttributes]).
 
 %	expand_table_commands(+BodyIn, -BodyOut)
 %
@@ -1614,27 +1766,33 @@ table_columns(Fmt, Ncols, Cols) :-
 	table_columns(Fmt, 0, Ncols, Cols).
 
 table_columns([],      Ncols, Ncols, []).
-table_columns([0'l|T], NC0,   NC,    [html('<COL>')|TH]) :-
+table_columns([0'l|T], NC0,   NC,    [[]|TH]) :-
 	NC1 is NC0 + 1,
 	table_columns(T, NC1, NC, TH).
-table_columns([0'c|T], NC0,   NC,    [html('<COL ALIGN=center>')|TH]) :-
+table_columns([0'c|T], NC0,   NC,    [['ALIGN'=center]|TH]) :-
 	NC1 is NC0 + 1,
 	table_columns(T, NC1, NC, TH).
-table_columns([0'r|T], NC0,   NC,    [html('<COL ALIGN=right>')|TH]) :-
+table_columns([0'r|T], NC0,   NC,    [['ALIGN'=right]|TH]) :-
 	NC1 is NC0 + 1,
 	table_columns(T, NC1, NC, TH).
-table_columns([0'p|T0], NC0,   NC,   [html(Col)|TH]) :-
-	phrase(parbox_width(W), T0, T),
-	(   integer(W)
-	->  sformat(Col, '<COL WIDTH=~d%>', [W])
+table_columns([0'p|T0], NC0,   NC,   [Col|TH]) :-
+	phrase(parbox_width(_W), T0, T),
+	Col = [],
+/*	(   integer(W)
+	->  Col = ['WIDTH'=W]
 	;   append(Done, T, T0),
 	    format(user_error,
 		   'Could not determing column width from "~s"~n', [Done]),
-	    Col = '<COL>'
+	    Col = []
 	),
+*/
 	NC1 is NC0 + 1,
 	table_columns(T, NC1, NC, TH).
-table_columns([0'||T], NC0,  NC,     [html('<COLGROUP>')|TH]) :-
+table_columns([0'D|T0], NC0,   NC,   [['ALIGN'=char, 'CHAR'=Chr]|TH]) :-
+	phrase(align_char(Chr), T0, T),
+	NC1 is NC0 + 1,
+	table_columns(T, NC1, NC, TH).
+table_columns([0'||T], NC0,  NC,     TH) :-
 	table_columns(T, NC0, NC, TH).
 
 parbox_width(W) -->
@@ -1648,6 +1806,17 @@ parbox_width(-) -->
 	string_without("{}", _),
 	"}", !.
 
+align_char(Chr) -->		% D{inputsep}{outputsep}{decimal places}
+	"{",
+	string_without("{}", W),
+	"}{",
+	string_without("{}", _),
+	"}{",
+	number(_),
+	"}",
+	{ atom_chars(Chr, W)
+	}.
+
 string_without(L, [C|T]) -->
 	[C],
 	{\+ member(C, L)}, !,
@@ -1655,14 +1824,24 @@ string_without(L, [C|T]) -->
 string_without(_, []) --> [].
 
 number(N) -->
+	optional_sign(S),
 	digit(C0),
 	float_digits(C),
-	{number_chars(N, [C0|C])}.
+	{ append(S, [C0|C], Chars),
+	  number_chars(N, Chars)
+	}.
 
 integer(N) -->
+	optional_sign(S),
 	digit(C0),
 	digits(C),
-	{number_chars(N, [C0|C])}.
+	{ append(S, [C0|C], Chars),
+	  number_chars(N, Chars)
+	}.
+
+optional_sign("-") --> "-", !.
+optional_sign("+") --> "+", !.
+optional_sign("")  --> "".
 
 digit(C) -->
 	[C],
@@ -1698,48 +1877,81 @@ relative_width(mm, N, W) :- W is integer(100 * (N/25.4) / 6).
 
 :- set_feature(character_escapes, true).
 
-table_body([], []).
-table_body([\(\, _)|T], []) :-
-	 all_white_space(T).
-table_body([' '|T0], T) :-
-	 table_body(T0, T).
-table_body(['\n'|T0], T) :- !,
-	table_body(T0, T).
-table_body([\hline|T0], [html('<TBODY>')|T]) :-
-	 table_body(T0, T).
-table_body(Body, [[ html('<TR VALIGN=top>'),
-		    Row,
-		    html('</TR>')
-		  ]|Rest]) :-
-	table_row(Body, BodyRest, Row),
-	table_body(BodyRest, Rest).
+table_body([], _, []).
+table_body([\(\, _)|T], _, []) :-
+	all_white_space(T).
+table_body([' '|T0], ColAtts, T) :-
+	table_body(T0, ColAtts, T).
+table_body(['\n'|T0], ColAtts, T) :- !,
+	table_body(T0, ColAtts, T).
+table_body([\hline|T0], ColAtts, [html('<TBODY>')|T]) :-
+	table_body(T0, ColAtts, T).
+table_body(Body, ColAtts, [[ html('<TR VALIGN=top>'),
+			     Row,
+			     html('</TR>')
+			  ]|Rest]) :-
+	table_row(Body, 1, ColAtts, BodyRest, Row),
+	table_body(BodyRest, ColAtts, Rest).
 
-table_row([], [], []) :- !.
-table_row([' '|T0], T, TH) :- !,
-	table_row(T0, T, TH).
-table_row(['\n'|T0], T, TH) :- !,
-	table_row(T0, T, TH).
-table_row([\(\, _)|Rest], Rest, []) :- !.
-table_row([\(multicolumn, [{N}, {A}, {Tokens}])|R0], R,
+table_row([], _, _, [], []) :- !.
+table_row([' '|T0], C, ColAtts, T, TH) :- !,
+	table_row(T0, C, ColAtts, T, TH).
+table_row(['\n'|T0], C, ColAtts, T, TH) :- !,
+	table_row(T0, C, ColAtts, T, TH).
+table_row([\(\, _)|Rest], _, _, Rest, []) :- !.
+table_row([\(multicolumn, [{N}, {A}, {Tokens}])|R0], C, ColAtts, R,
 	  [html(MC), Item|THtml]) :-
 	column_alignment(A, Alignment), !,
-	sformat(MC, '<TD COLSPAN=~w HALIGN=~w>', [N, Alignment]),
+	sformat(MC, '<TD COLSPAN=~w ALIGN=~w>', [N, Alignment]),
 	translate_group(Tokens, Item),
-	table_row(R0, R, THtml).
-table_row(L,  R,  [html('<TD>'), Chtml|THtml]) :-
+	to_integer(N, N2),
+	C2 is C + N2,
+	table_cell(R0, R1, _),		% discard tokens upto &
+	table_row(R1, C2, ColAtts, R, THtml).
+table_row(L, C, ColAtts, R,  [html(CellHeader), Chtml|THtml]) :-
+	cell_header(C, ColAtts, CellHeader),
 	table_cell(L, T, Tokens),
 	translate_group(Tokens, Chtml),
-	table_row(T, R, THtml).
+	C2 is C + 1,
+	table_row(T, C2, ColAtts, R, THtml).
 
-column_alignment(l, left).
-column_alignment(c, center).
-column_alignment(r, right).
-column_alignment(X, left) :-
-	format(user_error, 'Unknown multicolumn alignment: "~w"~n', [X]).
+cell_header(C, ColAtts, Header) :-
+	nth1(C, ColAtts, Spec),
+	maplist(sgml_attribute, Spec, Attributes),
+	concat_atom(['<TD'|Attributes], ' ', H0),
+	concat(H0, '>', Header).
+
+sgml_attribute(Name=Value, Att) :-
+	concat_atom([Name, =, Value], Att).
+
+to_integer(Atom, Integer) :-
+	atom_chars(Atom, Chars),
+	number_chars(Integer, Chars).
+
+column_alignment(X, Alignment) :-
+	atom_chars(X, Chars),
+	phrase(column_alignment(Alignment), Chars).
+
+column_alignment(A) -->
+	vlines,
+	calignment(A), !,
+	vlines.
+
+vlines -->
+	"|", !,
+	vlines.
+vlines -->
+	[].
+
+calignment(left) --> "l".
+calignment(center) --> "c".
+calignment(right) --> "r".
+calignment(left) --> [X],
+	{format(user_error, 'Unknown multicolumn alignment: "~w"~n', [X])}.
 
 table_cell([], [], []).
-table_cell([' '|T0], R, T) :- !,
-	 table_cell(T0, R, T).
+%table_cell([' '|T0], R, T) :- !,
+%	 table_cell(T0, R, T).
 table_cell([&|L], L, []) :- !.
 table_cell([\(\, A)|L], [\(\, A)|L], []) :- !.
 table_cell([H|T0], R, [H|T]) :-
@@ -1923,6 +2135,13 @@ option(tmp,	Tmp) :-
 ps2gif(In, Out) :-
 	ps2gif(In, Out, []).
 
+ps2gif(In, Out, _Options) :-
+	absolute_file_name(In, [ access(read),
+				 extensions([gif])
+			       ],
+			   InFile), !,
+	concat_atom(['cp ', InFile, ' ', Out], Cmd),
+	shell(Cmd).
 ps2gif(In, Out, Options) :-
 	get_option(Options, tmp(Tmp)),
 	get_option(Options, res(Res0)),
@@ -2025,6 +2244,21 @@ write_html(ref(Label)) :- !,
 write_html(label(Label, Text, _)) :- !,
 	sformat(Anchor, '<A NAME="~w">', [Label]),
 	write_html([html(Anchor), Text, html('</A>')]).
+write_html(body_link(Link)) :- !,
+	(   translate_ref(Link, Ref, Type)
+	->  sformat(Anchor, '<A HREF="~w">', [Ref]),
+	    capitalise_atom(Type, Text),
+	    write_html([html(Anchor), Text, html('</A>')]),
+	    nl_html
+	;   true
+	).
+write_html(link(Link)) :- !,
+	(   translate_ref(Link, Ref, Type)
+	->  sformat(Html, '<LINK REL=~w HREF="~w">', [Type, Ref]),
+	    write_html(html(Html)),
+	    nl_html
+	;   true
+	).
 write_html(iflref(fileof(Label), Text)) :- !,
 	(   label(Label, _, _)
 	->  write_html(lref(fileof(Label), Text))
@@ -2053,7 +2287,8 @@ write_html(lref(Label, Text)) :-
 	write_html(lref(FixedLabel, Text)).
 write_html(lref(Label, Text)) :- !,
 	format(user_error, 'No label for reference "~w"~n', [Label]),
-	write_html(Text).
+	macro_expand(#b(Text), Expanded),
+	write_html(Expanded).
 write_html(lforw(Label, Text)) :- !,
 	concat('back-to-', Label, BackName),
 	(   label(Label, File, _)
@@ -2104,6 +2339,40 @@ write_html(H) :-
 	put_html_token(H), !.
 write_html(_).
 	
+:- set_feature(character_escapes, true).
+nl_html :-
+	write_html(verb('\n')).
+:- set_feature(character_escapes, false).
+
+%	translate_ref(+Label, -Anchor, -TextLabel)
+
+translate_ref(next, Anchor, next) :-
+	current_html_output(Current),
+	next_file(Current, Next),
+	concat(Next, '.html', Anchor).
+translate_ref(previous, Anchor, previous) :-
+	current_html_output(Current),
+	next_file(Prev, Current),
+	concat(Prev, '.html', Anchor).
+translate_ref(up(Label), Anchor, up) :-
+	label(Label, File, _),
+	concat(File, '.html', Anchor).
+translate_ref(home, Anchor, home) :-
+	html_file_base(Home),
+	\+ current_html_output(Home),
+	onefile(false),
+	concat(Home, '.html', Anchor).
+translate_ref(contents, Anchor, contents) :-
+	label('document-contents', File, _),
+	concat(File, '.html', Anchor).
+translate_ref(index, Anchor, index) :-
+	label('document-index', File, _),
+	concat(File, '.html', Anchor).
+translate_ref(summary, Anchor, summary) :-
+	label('sec:summary', File, _),
+	concat(File, '.html', Anchor).
+
+
 cmd_layout('<P>',    2, 0).
 cmd_layout('<DL>',   2, 1).
 cmd_layout('</DL>',  1, 2). 
@@ -2133,7 +2402,13 @@ cmd_layout('<LISTING>',	 2, 0).
 cmd_layout('</LISTING>', 0, 2). 
 cmd_layout('<XMP>',	 2, 0). 
 cmd_layout('</XMP>', 	 0, 2). 
-     
+cmd_layout('<BODY>',	 2, 1). 
+cmd_layout('</BODY>',	 1, 1). 
+cmd_layout('</HEAD>',	 0, 1). 
+cmd_layout('<HTML>',	 0, 1). 
+cmd_layout('</HTML>',	 1, 1). 
+cmd_layout(Cmd,		 1, 1) :-
+	concat('<TABLE', _, Cmd).
 
 :- initialization
    read_tex_inputs.
