@@ -1,4 +1,4 @@
-/*  $Id: pl-arith.c,v 1.31 1997/02/10 12:56:40 jan Exp $
+/*  $Id: pl-arith.c,v 1.34 1997/08/07 07:57:34 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -35,18 +35,20 @@ day.
 #define M_E (2.7182818284590452354)
 #endif
 
+#if !defined(HAVE_ISNAN) && defined(NaN)
+#define isnan(f)  ((f) == NaN)
+#define HAVE_ISNAN
+#endif
+
 #ifdef WIN32
 #include <excpt.h>
 #endif
 
-#define MAXARITHFUNCTIONS (100)
-
-typedef struct arithFunction * 	ArithFunction;
 typedef int (*ArithF)();
 
 struct arithFunction
 { ArithFunction next;		/* Next of chain */
-  FunctorDef	functor;	/* Functor defined */
+  functor_t	functor;	/* Functor defined */
   ArithF	function;	/* Implementing function */
   Module	module;		/* Module visibility module */
 #if O_PROLOG_FUNCTIONS
@@ -57,12 +59,13 @@ struct arithFunction
 #endif
 };
 
-forwards ArithFunction	isCurrentArithFunction(FunctorDef, Module);
-static void		promoteToRealNumber(Number n);
+#define arithFunctionTable	(GD->arith.table)
+#define function_array		(&GD->arith.functions)
+#define FunctionFromIndex(n)	fetchBuffer(function_array, n, ArithFunction)
 
-static ArithFunction arithFunctionTable[ARITHHASHSIZE];
-static code next_index;
-static ArithFunction functions;
+static ArithFunction	isCurrentArithFunction(functor_t, Module);
+static void		registerFunction(ArithFunction f);
+static void		promoteToRealNumber(Number n);
 
 
 		/********************************
@@ -75,9 +78,10 @@ pl_between(term_t low, term_t high, term_t n, word b)
   { case FRG_FIRST_CALL:
       { long l, h, i;
 
-	if ( !PL_get_long(low, &l) ||
-	     !PL_get_long(high, &h) )
-	  return warning("between/3: instantiation fault");
+	if ( !PL_get_long(low, &l) )
+	  return PL_error("between", 3, NULL, ERR_TYPE, ATOM_integer, low);
+	if ( !PL_get_long(high, &h) )
+	  return PL_error("between", 3, NULL, ERR_TYPE, ATOM_integer, high);
 
 	if ( PL_get_long(n, &i) )
 	{ if ( i >= l && i <= h )
@@ -85,7 +89,7 @@ pl_between(term_t low, term_t high, term_t n, word b)
 	  fail;
 	}
 	if ( !PL_is_variable(n) )
-	  return warning("between/3: instantiation fault");
+	  return PL_error("between", 3, NULL, ERR_TYPE, ATOM_integer, n);
 	if ( h < l )
 	  fail;
 
@@ -111,31 +115,60 @@ pl_between(term_t low, term_t high, term_t n, word b)
 
 word
 pl_succ(term_t n1, term_t n2)
-{ long i;
+{ long i1, i2;
 
-  if ( PL_get_long(n1, &i) )
-    return PL_unify_integer(n2, i+1);
-  if ( PL_get_long(n2, &i) )
-    return PL_unify_integer(n1, i-1);
+  if ( PL_get_long(n1, &i1) )
+  { if ( PL_get_long(n2, &i2) )
+      return i1+1 == i2 ? TRUE : FALSE;
+    else if ( PL_unify_integer(n2, i1+1) )
+      succeed;
 
-  return warning("succ/2: instantiation fault");
+    return PL_error("succ", 2, NULL, ERR_TYPE, ATOM_integer, n2);
+  }
+  if ( PL_get_long(n2, &i2) )
+  { if ( PL_unify_integer(n1, i2-1) )
+      succeed;
+  }
+
+  return PL_error("succ", 2, NULL, ERR_TYPE, ATOM_integer, n1);
 }
 
+
+static int
+var_or_long(term_t t, long *l, int which, int *mask)
+{ if ( PL_get_long(t, l) )
+  { *mask |= which;
+    succeed;
+  } 
+  if ( PL_is_variable(t) )
+    succeed;
+    
+  return PL_error("plus", 3, NULL, ERR_TYPE, ATOM_integer, t);
+}
+
+
 word
 pl_plus(term_t a, term_t b, term_t c)
 { long m, n, o;
+  int mask = 0;
 
-  if ( PL_get_long(a, &m) )
-  { if ( PL_get_long(b, &n) )
+  if ( !var_or_long(a, &m, 0x1, &mask) ||
+       !var_or_long(b, &n, 0x2, &mask) ||
+       !var_or_long(c, &o, 0x4, &mask) )
+    fail;
+
+  switch(mask)
+  { case 0x7:
+      return m+n == o ? TRUE : FALSE;
+    case 0x3:				/* +, +, - */
       return PL_unify_integer(c, m+n);
-    if ( PL_get_long(c, &o) )
+    case 0x5:				/* +, -, + */
       return PL_unify_integer(b, o-m);
-  } else
-  { if ( PL_get_long(b, &n) && PL_get_long(c, &o) )
+    case 0x6:				/* -, +, + */
       return PL_unify_integer(a, o-n);
+    default:
+      return PL_error("succ", 2, NULL, ERR_INSTANTIATION);
   }
-
-  return warning("plus/3: instantiation fault");
 }
 
 
@@ -180,7 +213,7 @@ ar_compare(Number n1, Number n2, int what)
 }
 
 
-word
+static word
 compareNumbers(term_t n1, term_t n2, int what)
 { number left, right;
 
@@ -226,16 +259,16 @@ pl_equalNumbers(term_t n1, term_t n2)			/* =:=/2 */
 		*********************************/
 
 static ArithFunction
-isCurrentArithFunction(register FunctorDef f, register Module m)
-{ register ArithFunction a;
+isCurrentArithFunction(functor_t f, Module m)
+{ ArithFunction a;
   ArithFunction r = NULL;
   int level = 30000;
 
-  for(a = arithFunctionTable[pointerHashValue(f, ARITHHASHSIZE)];
+  for(a = arithFunctionTable[functorHashValue(f, ARITHHASHSIZE)];
       a && !isTableRef(a); a = a->next)
   { if ( a->functor == f )
-    { register Module m2;
-      register int l;
+    { Module m2;
+      int l;
 
       for( m2 = m, l = 0; m2; m2 = m2->super, l++ )
       { if ( m2 == a->module && l < level )
@@ -258,7 +291,7 @@ realExceptionHandler(int sig, int type, SignalContext scp, char *addr)
 #ifndef BSD_SIGNALS
   signal(sig, (OsSigHandler)realExceptionHandler);
 #endif
-  if ( status.arithmetic > 0 )
+  if ( LD->in_arithmetic > 0 )
   { warning("Floating point exception");
 #ifndef O_RUNTIME
     Sfprintf(Serror, "[PROLOG STACK:\n");
@@ -302,21 +335,30 @@ calling convention required by ar_func_n() below.
 static int
 prologFunction(ArithFunction f, term_t av, Number r)
 { int arity = f->proc->definition->functor->arity;
-  fid_t cid = PL_open_foreign_frame();
+  fid_t fid = PL_open_foreign_frame();
   qid_t qid;
   int rval;
 
-  qid = PL_open_query(NULL, TRUE, f->proc, av);
+  qid = PL_open_query(NULL, PL_Q_CATCH_EXCEPTION, f->proc, av);
 
   if ( PL_next_solution(qid) )
-    rval = valueExpression(av+arity-1, r);
-  else
-  { warning("Arithmetic function %s failed", procedureName(f->proc));
-    rval = FALSE;
-  }
+  { rval = valueExpression(av+arity-1, r);
+    PL_close_query(qid);
+    PL_discard_foreign_frame(fid);
+  } else
+  { term_t except;
 
-  PL_close_query(qid);
-  PL_discard_foreign_frame(cid);
+    if ( (except = PL_exception(qid)) )
+    { rval = PL_throw(except);		/* pass exception */
+    } else
+    { char *name = stringAtom(f->proc->definition->functor->name);
+
+      rval = PL_error(name, arity-1, NULL, ERR_FAILED, f->proc);
+    }
+
+    PL_cut_query(qid);			/* donot destroy data */
+    PL_close_foreign_frame(fid);	/* same */
+  }
 
   return rval;
 }
@@ -326,7 +368,7 @@ prologFunction(ArithFunction f, term_t av, Number r)
 int
 valueExpression(term_t t, Number r)
 { ArithFunction f;
-  FunctorDef fDef;
+  functor_t functor;
   Word p = valTermRef(t);
   word w;
 
@@ -343,43 +385,49 @@ valueExpression(term_t t, Number r)
       r->type = V_REAL;
       succeed;
     case TAG_VAR:
-      return warning("Unbound variable in arithmetic expression");
+      return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
     case TAG_ATOM:
-      fDef = lookupFunctorDef(w, 0);
+      functor = lookupFunctorDef(w, 0);
       break;
     case TAG_COMPOUND:
-      fDef = functorTerm(w);
+      functor = functorTerm(w);
       break;
     default:
-      return warning("Illegal data type in arithmetic expression");
+      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_number, t);
   }
 
-  if ( !(f = isCurrentArithFunction(fDef, contextModule(environment_frame))))
-  { if ( fDef == FUNCTOR_dot2 )		/* handle "a" (make function) */
-    { Word a, p = valTermRef(t);
+  if ( !(f = isCurrentArithFunction(functor,
+				    contextModule(environment_frame))))
+  { if ( functor == FUNCTOR_dot2 )	/* handle "a" (make function) */
+    { Word a, b, p = valTermRef(t);
 
       deRef(p);
       a = argTermP(*p, 0);
       deRef(a);
       if ( isTaggedInt(*a) )
-      { a = argTermP(*p, 1);
-	deRef(a);
-	if ( *a == ATOM_nil )
+      { b = argTermP(*p, 1);
+	deRef(b);
+	if ( *b == ATOM_nil )
 	{ r->value.i = valInt(*a);
 	  r->type = V_INTEGER;
 	  succeed;
+	} else
+	{ term_t a2 = PL_new_term_ref();
+	  PL_get_arg(2, t, a2);
+	  return PL_error(".", 2, NULL, ERR_TYPE, ATOM_nil, a2);
 	}
+      } else
+      { term_t a1 = PL_new_term_ref();
+	PL_get_arg(1, t, a1);
+	return PL_error(".", 2, NULL, ERR_TYPE, ATOM_integer, a1);
       }
-      return warning("./2: Bad character code");
     } else
-    { return warning("Unknown arithmetic operator: %s/%d",
-		     stringAtom(fDef->name), fDef->arity);
-    }
+      return PL_error(NULL, 0, NULL, ERR_NOT_EVALUABLE, functor);
   }
 
 #if O_PROLOG_FUNCTIONS
   if ( f->proc )
-  { int rval, n, arity = fDef->arity;
+  { int rval, n, arity = arityFunctor(functor);
     term_t h0 = PL_new_term_refs(arity+1); /* one extra for the result */
 
     for(n=0; n<arity; n++)
@@ -404,9 +452,9 @@ valueExpression(term_t t, Number r)
     __try
     {
 #else
-    status.arithmetic++;
+    LD->in_arithmetic++;
 #endif
-    switch(fDef->arity)
+    switch(arityFunctor(functor))
     { case 0:
 	rval = (*f->function)(r);
         break;
@@ -455,9 +503,21 @@ valueExpression(term_t t, Number r)
       pl_abort();
     }
 #else
-    status.arithmetic--;
+    LD->in_arithmetic--;
 #endif
 
+    if ( r->type == V_REAL )
+    {
+#ifdef HUGE_VAL
+      if ( r->value.f == HUGE_VAL )
+	return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
+#endif
+#ifdef HAVE_ISNAN
+      if ( isnan(r->value.f) )
+	return PL_error(NULL, 0, NULL, ERR_AR_UNDEF);
+#endif
+    }
+
     return rval;
   }
 }
@@ -583,10 +643,13 @@ overflow:
 
 /* Binary functions requiring integer argument */
 
-#define BINAIRY_INT_FUNCTION(name, op) \
+#define BINAIRY_INT_FUNCTION(name, plop, op) \
   static int \
   name(Number n1, Number n2, Number r) \
-  { TRY(toIntegerNumber(n1) && toIntegerNumber(n2)); \
+  { if ( !toIntegerNumber(n1) ) \
+      return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n1); \
+    if ( !toIntegerNumber(n2) ) \
+      return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n2); \
     r->value.i = n1->value.i op n2->value.i; \
     r->type = V_INTEGER; \
     succeed; \
@@ -602,27 +665,91 @@ overflow:
     succeed; \
   }
 
-UNAIRY_FLOAT_FUNCTION(ar_sqrt, sqrt)
 UNAIRY_FLOAT_FUNCTION(ar_sin, sin)
 UNAIRY_FLOAT_FUNCTION(ar_cos, cos)
 UNAIRY_FLOAT_FUNCTION(ar_tan, tan)
-UNAIRY_FLOAT_FUNCTION(ar_asin, asin)
-UNAIRY_FLOAT_FUNCTION(ar_acos, acos)
 UNAIRY_FLOAT_FUNCTION(ar_atan, atan)
-UNAIRY_FLOAT_FUNCTION(ar_log, log)
 UNAIRY_FLOAT_FUNCTION(ar_exp, exp)
-UNAIRY_FLOAT_FUNCTION(ar_log10, log10)
 
 BINAIRY_FLOAT_FUNCTION(ar_atan2, atan2)
 BINAIRY_FLOAT_FUNCTION(ar_pow, pow)
 
-BINAIRY_INT_FUNCTION(ar_mod, %)
-BINAIRY_INT_FUNCTION(ar_div, /)
-BINAIRY_INT_FUNCTION(ar_disjunct, |)
-BINAIRY_INT_FUNCTION(ar_conjunct, &)
-BINAIRY_INT_FUNCTION(ar_shift_right, >>)
-BINAIRY_INT_FUNCTION(ar_shift_left, <<)
-BINAIRY_INT_FUNCTION(ar_xor, ^)
+BINAIRY_INT_FUNCTION(ar_mod, "mod", %)
+BINAIRY_INT_FUNCTION(ar_disjunct, "\\/", |)
+BINAIRY_INT_FUNCTION(ar_conjunct, "/\\", &)
+BINAIRY_INT_FUNCTION(ar_shift_right, ">>", >>)
+BINAIRY_INT_FUNCTION(ar_shift_left, "<<", <<)
+BINAIRY_INT_FUNCTION(ar_xor, "xor", ^)
+
+static int
+ar_sqrt(Number n1, Number r)
+{ promoteToRealNumber(n1);
+  if ( n1->value.f < 0 )
+    return PL_error("sqrt", 1, NULL, ERR_AR_UNDEF);
+  r->value.f = sqrt(n1->value.f);
+  r->type    = V_REAL;
+  succeed;
+}
+
+
+static int
+ar_asin(Number n1, Number r)
+{ promoteToRealNumber(n1);
+  if ( n1->value.f < -1.0 || n1->value.f > 1.0 )
+    return PL_error("asin", 1, NULL, ERR_AR_UNDEF);
+  r->value.f = asin(n1->value.f);
+  r->type    = V_REAL;
+  succeed;
+}
+
+
+static int
+ar_acos(Number n1, Number r)
+{ promoteToRealNumber(n1);
+  if ( n1->value.f < -1.0 || n1->value.f > 1.0 )
+    return PL_error("acos", 1, NULL, ERR_AR_UNDEF);
+  r->value.f = acos(n1->value.f);
+  r->type    = V_REAL;
+  succeed;
+}
+
+
+static int
+ar_log(Number n1, Number r)
+{ promoteToRealNumber(n1);
+  if ( n1->value.f <= 0.0 )
+    return PL_error("log", 1, NULL, ERR_AR_UNDEF);
+  r->value.f = log(n1->value.f);
+  r->type    = V_REAL;
+  succeed;
+}
+
+
+static int
+ar_log10(Number n1, Number r)
+{ promoteToRealNumber(n1);
+  if ( n1->value.f <= 0.0 )
+    return PL_error("log10", 1, NULL, ERR_AR_UNDEF);
+  r->value.f = log10(n1->value.f);
+  r->type    = V_REAL;
+  succeed;
+}
+
+
+static int
+ar_div(Number n1, Number n2, Number r)
+{ if ( !toIntegerNumber(n1) )
+    return PL_error("//", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
+  if ( !toIntegerNumber(n2) )
+    return PL_error("//", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
+  if ( n2->value.i == 0 )
+    return PL_error("//", 2, NULL, ERR_DIV_BY_ZERO);
+
+  r->value.i = n1->value.i / n2->value.i;
+  r->type = V_INTEGER;
+
+  succeed;
+}
 
 static int
 ar_sign(Number n1, Number r)
@@ -640,9 +767,10 @@ static int
 ar_rem(Number n1, Number n2, Number r)
 { real f;
 
-  if ( !toIntegerNumber(n1) ||
-       !toIntegerNumber(n2) )
-    return warning("rem/2: arguments must be integers");
+  if ( !toIntegerNumber(n1) )
+    return PL_error("rem", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
+  if ( !toIntegerNumber(n2) )
+    return PL_error("rem", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
 
   f = (real)n1->value.i / (real)n2->value.i;
   r->value.f = f - (real)((long) f);
@@ -654,7 +782,10 @@ ar_rem(Number n1, Number n2, Number r)
 static int
 ar_divide(Number n1, Number n2, Number r)
 { if ( intNumber(n1) && intNumber(n2) )
-  { if ( n1->value.i % n2->value.i == 0)
+  { if ( n2->value.i == 0 )
+      return PL_error("/", 2, NULL, ERR_DIV_BY_ZERO);
+
+    if ( n1->value.i % n2->value.i == 0)
     { r->value.i = n1->value.i / n2->value.i;
       r->type = V_INTEGER;
       succeed;
@@ -663,6 +794,8 @@ ar_divide(Number n1, Number n2, Number r)
 
   promoteToRealNumber(n1);
   promoteToRealNumber(n2);
+  if ( n2->value.f == 0.0 )
+      return PL_error("/", 2, NULL, ERR_DIV_BY_ZERO);
 
   r->value.f = n1->value.f / n2->value.f;
   r->type = V_REAL;
@@ -729,7 +862,7 @@ ar_min(Number n1, Number n2, Number r)
 static int
 ar_negation(Number n1, Number r)
 { if ( !toIntegerNumber(n1) )
-    return warning("is/2: argument to \\/1 should be an integer");
+    return PL_error("\\", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   r->value.i = ~n1->value.i;
   r->type = V_INTEGER;
@@ -785,7 +918,7 @@ ar_integer(Number n1, Number r)
     r->type = V_REAL;
     succeed;
 #else
-    return warning("integer/1: argument too large");
+    return PL_error("integer", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
 #endif
   }
 }
@@ -885,7 +1018,7 @@ ar_truncate(Number n1, Number r)
 static int
 ar_random(Number n1, Number r)
 { if ( !toIntegerNumber(n1) )
-    return warning("random/1: n1->valueument should be an integer");
+    return PL_error("random", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   r->value.i = Random() % n1->value.i;
   r->type = V_INTEGER;
@@ -943,8 +1076,9 @@ word
 pl_arithmetic_function(term_t descr)
 { Procedure proc;
   Definition def;
-  FunctorDef fd;
-  register ArithFunction f;
+  functor_t fd;
+  FunctorDef fdef;
+  ArithFunction f;
   Module m = NULL;
   term_t head = PL_new_term_ref();
   int v;
@@ -952,28 +1086,28 @@ pl_arithmetic_function(term_t descr)
   PL_strip_module(descr, &m, head);
   if ( !PL_get_functor(head, &fd) )
     return warning("arithmetic_function/1: Illegal head");
-  if ( fd->arity < 1 )
+  fdef = valueFunctor(fd);
+  if ( fdef->arity < 1 )
     return warning("arithmetic_function/1: Illegal arity");
 
   proc = lookupProcedure(fd, m);
   def = proc->definition;
-  fd = lookupFunctorDef(fd->name, fd->arity - 1);
+  fd = lookupFunctorDef(fdef->name, fdef->arity - 1);
   if ( (f = isCurrentArithFunction(fd, m)) && f->module == m )
     succeed;				/* already registered */
 
-  if ( next_index >= MAXARITHFUNCTIONS )
-    return warning("Cannot handle more than %d arithmetic functions",
-		   MAXARITHFUNCTIONS);
-
-  v = pointerHashValue(fd, ARITHHASHSIZE);
-  f = &functions[next_index];
+  v = functorHashValue(fd, ARITHHASHSIZE);
+  f = allocHeap(sizeof(struct arithFunction));
   f->functor  = fd;
   f->function = NULL;
   f->module   = m;
   f->proc     = proc;
-  f->index    = next_index++;
+
+  startCritical;
   f->next     = arithFunctionTable[v];
   arithFunctionTable[v] = f;  
+  registerFunction(f);
+  endCritical;
 
   succeed;
 }
@@ -986,7 +1120,7 @@ pl_current_arithmetic_function(term_t f, word h)
 
   switch( ForeignControl(h) )
   { case FRG_FIRST_CALL:
-    { FunctorDef fd;
+    { functor_t fd;
 
       PL_strip_module(f, &m, head);
 
@@ -1030,9 +1164,14 @@ pl_current_arithmetic_function(term_t f, word h)
 
 #endif /* O_PROLOG_FUNCTIONS */
 
-#define ADD(functor, func) { (ArithFunction)NULL, functor, func }
+typedef struct
+{ functor_t	functor;
+  ArithF	function;
+} ar_funcdef;
 
-static struct arithFunction ar_functions[MAXARITHFUNCTIONS] = {
+#define ADD(functor, func) { functor, func }
+
+static const ar_funcdef ar_funcdefs[] = {
   ADD(FUNCTOR_plus2,		ar_add),
   ADD(FUNCTOR_minus2,		ar_minus),
   ADD(FUNCTOR_star2,		ar_times),
@@ -1083,16 +1222,48 @@ static struct arithFunction ar_functions[MAXARITHFUNCTIONS] = {
   ADD(FUNCTOR_e0,		ar_e),
 
   ADD(FUNCTOR_cputime0,		ar_cputime),
-
-  ADD((FunctorDef)NULL,		(ArithF)NULL)
 };
 
 #undef ADD
 
+static void
+registerFunction(ArithFunction f)
+{ f->index = entriesBuffer(function_array, ArithFunction);
+  addBuffer(function_array, f, ArithFunction);
+}
+
+
+static void
+registerBuiltinFunctions()
+{ int n, size = sizeof(ar_funcdefs)/sizeof(ar_funcdef);
+  ArithFunction f = allocHeap(size * sizeof(struct arithFunction));
+  const ar_funcdef *d;
+
+					/* grow to desired size immediately */
+  growBuffer(function_array, size * sizeof(ArithFunction));
+  memset(f, 0, size * sizeof(struct arithFunction));
+
+  for(d = ar_funcdefs, n=0; n<size; n++, f++, d++)
+  { int v = functorHashValue(d->functor, ARITHHASHSIZE);
+
+    f->functor  = d->functor;
+    f->function = d->function;
+    f->module   = MODULE_system;
+    f->next     = arithFunctionTable[v];
+    arithFunctionTable[v] = f;
+    registerFunction(f);
+    DEBUG(1, Sdprintf("Registered %s/%d at %d, index=%d\n",
+		      stringAtom(nameFunctor(f->functor)),
+		      arityFunctor(f->functor),
+		      v,
+		      f->index));
+  }			       
+}
+
 
 void
 initArith(void)
-{
+{ 
 #ifdef SIGFPE
   pl_signal(SIGFPE, (handler_t) realExceptionHandler);
 #endif
@@ -1100,30 +1271,16 @@ initArith(void)
   setmatherr(realExceptionHandler);
 #endif
 
+  initBuffer(function_array);
 					/* link the table to enumerate */
-  { register ArithFunction *f;
-    register int n;
+  { ArithFunction *f;
+    int n;
 
     for(n=0, f = arithFunctionTable; n < (ARITHHASHSIZE-1); n++, f++)
       *f = makeTableRef(f+1);
   }
 
-					/* initialise it */
-  { ArithFunction f;
-    int v;
-
-    functions = ar_functions;
-
-    for( f = functions, next_index = 0; f->functor; f++, next_index++ )
-    { v = pointerHashValue(f->functor, ARITHHASHSIZE);
-      f->module = MODULE_system;
-#if O_COMPILE_ARITH
-      f->index = next_index;
-#endif
-      f->next = arithFunctionTable[v];
-      arithFunctionTable[v] = f;
-    }
-  }
+  registerBuiltinFunctions();
 }
 
 #if O_COMPILE_ARITH
@@ -1133,19 +1290,19 @@ initArith(void)
 		*********************************/
 
 int
-indexArithFunction(register FunctorDef fdef, register Module m)
-{ register ArithFunction f;
+indexArithFunction(functor_t fdef, register Module m)
+{ ArithFunction f;
 
-  if ( (f = isCurrentArithFunction(fdef, m)) == (ArithFunction) NULL )
+  if ( !(f = isCurrentArithFunction(fdef, m)) )
     return -1;
 
   return (int)f->index;
 }
 
 
-FunctorDef
+functor_t
 functorArithFunction(int n)
-{ return functions[n].functor;
+{ return FunctionFromIndex(n)->functor;
 }
 
 
@@ -1153,7 +1310,7 @@ bool
 ar_func_n(code n, int argc, Number *stack)
 { number result;
   int rval;
-  ArithFunction f = &functions[(int)n];
+  ArithFunction f = FunctionFromIndex((int)n);
   Number sp = *stack;
 
   sp -= argc;
@@ -1188,7 +1345,19 @@ ar_func_n(code n, int argc, Number *stack)
   }
 
   if ( rval )
-  { *sp++ = result;
+  { if ( result.type == V_REAL )
+    {
+#ifdef HUGE_VAL
+      if ( result.value.f == HUGE_VAL )
+	return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
+#endif
+#ifdef HAVE_ISNAN
+      if ( isnan(result.value.f) )
+	return PL_error(NULL, 0, NULL, ERR_AR_UNDEF);
+#endif
+    }
+
+    *sp++ = result;
     *stack = sp;
   }
 
