@@ -1,4 +1,4 @@
-/*  $Id: pl-comp.c,v 1.45 1997/03/10 09:36:27 jan Exp $
+/*  $Id: pl-comp.c,v 1.53 1997/11/12 14:21:07 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -11,7 +11,7 @@
 
 #define CODE(c, n, a, e)	{ n, c, a, e }
 
-struct code_info codeTable[] = {
+const code_info codeTable[] = {
 /*     ID		name	     #args #xr */
   CODE(I_NOP,		"i_nop",	0, 0),
   CODE(I_ENTER,		"i_enter",	0, 0),
@@ -96,6 +96,9 @@ struct code_info codeTable[] = {
 #endif
   CODE(I_EXITFACT,	"i_exitfact",	0, 0),
   CODE(D_BREAK,		"d_break",	0, 0),
+#if O_CATCHTHROW
+  CODE(B_THROW,		"b_throw",	0, 0),
+#endif
 /*List terminator */
   CODE(0,		NULL,		0, 0)
 };
@@ -104,7 +107,7 @@ forwards void	checkCodeTable(void);
 
 static void
 checkCodeTable(void)
-{ CodeInfo ci;
+{ const code_info *ci;
   unsigned int n;
 
   for(ci = codeTable, n = 0; ci->name != NULL; ci++, n++ )
@@ -250,16 +253,16 @@ forwards int	analyseVariables2(Word, int, int, int);
 #define A_ERROR		2
 #endif /* O_COMPILE_ARITH */
 
-typedef struct
+typedef struct _varDef
 { word		functor;		/* mimic a functor (FUNCTOR_var1) */
   Word		address;		/* address of the variable */
   int		times;			/* occurences */
   int		offset;			/* offset in environment frame */
-} vardef, *VarDef;
+} vardef;
 
-static VarDef  *vardefs;		/* variables defined */
-static int      nvardefs;		/* # vardefs in array above */
-static int	filledVars;		/* vardef structures filled */
+#define vardefs		(LD->comp._vardefs)
+#define nvardefs	(LD->comp._nvardefs)
+#define filledVars	(LD->comp._filledVars)
 
 static VarDef
 getVarDef(int i)
@@ -277,7 +280,7 @@ getVarDef(int i)
       vardefs = malloc(sizeof(VarDef) * nvd);
     }
     if ( !vardefs )
-      fatalError("Not enough memory");
+      outOfCore();
 
     for(vdp = &vardefs[nvardefs], n=nvardefs; n++ < nvd; )
       *vdp++ = NULL;
@@ -287,7 +290,7 @@ getVarDef(int i)
   if ( !(vd = vardefs[i]) )
   { vd = vardefs[i] = allocHeap(sizeof(vardef));
     memset(vd, 0, sizeof(*vd));
-    vd->functor = FUNCTOR_var1->functor;
+    vd->functor = FUNCTOR_var1;
   }
 
   return vd;
@@ -367,7 +370,7 @@ analyse_variables(Word head, Word body, int arity, int *nv)
   for(n=0; n<arity+nvars; n++)
   { VarDef vd = vardefs[n];
 
-    assert(vd->functor == FUNCTOR_var1->functor);
+    assert(vd->functor == FUNCTOR_var1);
     if (vd->address == (Word) NULL)
       continue;
     if (vd->times == 1)				/* ISVOID */
@@ -395,27 +398,27 @@ analyseVariables2(Word head, int nvars, int arity, int argn)
     vd = getVarDef(index);
     vd->address = head;
     vd->times = 1;
-    *head = consPtr(vd, TAG_COMPOUND|STG_HEAP);
+    *head = (index<<7)|TAG_ATOM|STG_GLOBAL; /* special mark */
 
     return nvars;
   }
 
+  if ( tagex(*head) == (TAG_ATOM|STG_GLOBAL) )
+  { VarDef vd = vardefs[(*head) >> 7];
+
+    vd->times++;
+    return nvars;
+  }
+
   if ( isTerm(*head) )
   { Functor f = valueTerm(*head);
+    int ar = arityFunctor(f->definition);
 
-    if ( f->definition == FUNCTOR_var1->functor )
-    { VarDef vd = (VarDef)f;
-      vd->times++;
-      return nvars;
-    } else
-    { int ar = arityFunctor(f->definition);
+    head = f->arguments;
+    argn = ( argn < 0 ? 0 : arity );
 
-      head = f->arguments;
-      argn = ( argn < 0 ? 0 : arity );
-
-      for(; ar > 0; ar--, head++, argn++)
-	nvars = analyseVariables2(head, nvars, arity, argn);
-    }
+    for(; ar > 0; ar--, head++, argn++)
+      nvars = analyseVariables2(head, nvars, arity, argn);
   }
 
   return nvars;
@@ -475,7 +478,7 @@ typedef struct
   int		arity;			/* arity of top-goal */
   Clause	clause;			/* clause we are constructing */
   int		vartablesize;		/* size of the vartable */
-  buffer	codes;			/* scratch code table */
+  tmp_buffer	codes;			/* scratch code table */
   VarTable	used_var;		/* boolean array of used variables */
 } compileInfo;
 
@@ -498,8 +501,8 @@ forwards bool	compileArithArgument(Word, compileInfo *);
 
 static inline int
 isIndexedVarTerm(word w)
-{ if ( hasFunctor(w, FUNCTOR_var1) )
-  { VarDef v = (VarDef) valPtr(w);
+{ if ( tagex(w) == (TAG_ATOM|STG_GLOBAL) )
+  { VarDef v = vardefs[w>>7];
     return v->offset;
   }
 
@@ -560,16 +563,17 @@ orVars(VarTable valt1, VarTable valt2)
 
 static void
 setVars(register Word t, VarTable vt)
-{ deRef(t);
+{ int index;
+
+  deRef(t);
+  if ( (index = isIndexedVarTerm(*t)) >= 0 )
+  { isFirstVar(vt, index);
+    return;
+  }
 
   if ( isTerm(*t) )
-  { int index;
-    register int arity;
+  { int arity;
 
-    if ( (index = isIndexedVarTerm(*t)) >= 0 )
-    { isFirstVar(vt, index);
-      return;
-    }
     arity = arityTerm(*t);
     for(t = argTermP(*t, 0); arity > 0; t++, arity--)
       setVars(t, vt);
@@ -618,7 +622,9 @@ clause should belong to.
     return NULL;
 
   if ( (ci.arity = proc->definition->functor->arity) > MAXARITY )
-    return (Clause) warning("Compiler: arity too high (%d)\n", ci.arity);
+  { warning("Compiler: arity too high (%d)\n", ci.arity);
+    return (Clause) NULL;
+  }
 
   DEBUG(9, Sdprintf("Splitted and found proc\n"));
 
@@ -715,7 +721,7 @@ Finish up the clause.
 
     discardBuffer(&ci.codes);
 
-    statistics.codes += clause->code_size;
+    GD->statistics.codes += clause->code_size;
   }
 
   return clause;
@@ -755,7 +761,7 @@ A ; B, A -> B, A -> B ; C, \+ A
 
     NOTE: A tricky bit now is that we  can  reach  the  same  point  via
     different  paths.   Each of these paths may result in another set of
-    variabled  already  instantiated.   This  gives  troubles  with  the
+    variables  already  instantiated.   This  gives  troubles  with  the
     FIRSTVAR  type  of instructions.  to avoid such trouble the compiler
     generates  SETVAR  instructions  to  balance  both   brances.    See
     balanceVars();
@@ -766,7 +772,7 @@ compileBody(register Word body, code call, register compileInfo *ci)
 { deRef(body);
 
   if ( isTerm(*body) )
-  { FunctorDef fd = functorTerm(*body);
+  { functor_t fd = functorTerm(*body);
 
     if ( fd == FUNCTOR_comma2 )			/* A , B */
     { TRY( compileBody(argTermP(*body, 0), I_CALL, ci) );
@@ -903,6 +909,8 @@ be a variable, and thus cannot be removed if it is before an I_POPF.
       }
       /* FALLTHROUGH for tagged integers */
     case TAG_ATOM:
+      if ( tagex(*arg) == (TAG_ATOM|STG_GLOBAL) )
+	goto isvar;
       if ( isNil(*arg) )
       {	Output_0(ci, (where & A_BODY) ? B_NIL : H_NIL);
       } else
@@ -924,12 +932,11 @@ be a variable, and thus cannot be removed if it is before an I_POPF.
     }
   }
 
-  assert(isTerm(*arg));
-    
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Non-void variables. There are many cases for this.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+isvar:
   if ( (index = isIndexedVarTerm(*arg)) >= 0 )
   { first = isFirstVar(ci->used_var, index);
 
@@ -976,9 +983,11 @@ Non-void variables. There are many cases for this.
     return NONVOID;
   }
 
+  assert(isTerm(*arg));
+    
   { int ar;
     int lastnonvoid;
-    FunctorDef fdef;
+    functor_t fdef;
     int isright = (where & A_RIGHT);
 
     fdef = functorTerm(*arg);
@@ -1002,7 +1011,7 @@ Non-void variables. There are many cases for this.
       Output_1(ci, c, (word)fdef);
     }
     lastnonvoid = PC(ci);
-    ar = fdef->arity;
+    ar = arityFunctor(fdef);
     where &= ~A_RIGHT;
     for(arg = argTermP(*arg, 0); ar > 0; ar--, arg++)
     { where |= A_ARG;
@@ -1034,18 +1043,17 @@ compileSubClause(register Word arg, code call, compileInfo *ci)
 { Module tm = ci->module;
 
   deRef(arg);
-
-  if ( isTerm(*arg) )
-  {
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 A non-void variable. Create a I_USERCALL0 instruction for it.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-    if ( isIndexedVarTerm(*arg) >= 0 )
-    { compileArgument(arg, A_BODY, ci);
-      Output_0(ci, I_USERCALL0);
-      succeed;
-    }
+  if ( isIndexedVarTerm(*arg) >= 0 )
+  { compileArgument(arg, A_BODY, ci);
+    Output_0(ci, I_USERCALL0);
+    succeed;
+  }
 
+  if ( isTerm(*arg) )
+  {
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 If the argument is of the form <Module>:<Goal>, <Module> is an atom  and
 <Goal>  is  nonvar  then compile to the specified module.  Otherwise use
@@ -1081,7 +1089,7 @@ will use the meta-call mechanism for all these types of calls.
 /*  cont: */
 
 #if O_COMPILE_ARITH
-    if ( status.optimise )
+    if ( GD->cmdline.optimise )
     { switch( compileArith(arg, ci) )
       { case A_OK:	succeed;
 	case A_ERROR:	fail;
@@ -1094,8 +1102,9 @@ Term, not a variable and not a module call.  Compile the  arguments  and
 generate  the  call  instruction.   Note  this  codes traps the $apply/2
 operator.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-    { FunctorDef fdef = functorTerm(*arg);
-      Procedure proc = lookupProcedure(fdef, tm);
+    { functor_t functor = functorTerm(*arg);
+      FunctorDef fdef = valueFunctor(functor);
+      Procedure proc = lookupProcedure(functor, tm);
       int ar = fdef->arity;
 
 #ifdef O_INLINE_FOREIGNS
@@ -1108,7 +1117,7 @@ operator.
 	{ Word a = argTermP(*arg, n);
 
 	  deRef(a);
-	  if ( isTerm(*a) && (vars[n] = isIndexedVarTerm(*a)) >= 0 )
+	  if ( (vars[n] = isIndexedVarTerm(*a)) >= 0 )
 	    continue;
 
 	  goto non_fv;
@@ -1135,16 +1144,21 @@ operator.
       if ( fdef->name == ATOM_call )
       { Output_1(ci, I_USERCALLN, (code)(fdef->arity - 1));
 	succeed;
-      } else if ( fdef == FUNCTOR_apply2 )
+      } else if ( functor == FUNCTOR_apply2 )
       { Output_0(ci, I_APPLY);
 	succeed;
 #if O_BLOCK
-      } else if ( fdef == FUNCTOR_dcut1 )
+      } else if ( functor == FUNCTOR_dcut1 )
       { Output_0(ci, I_CUT_BLOCK);
 	succeed;
-      } else if ( fdef == FUNCTOR_dexit2 )
+      } else if ( functor == FUNCTOR_dexit2 )
       { Output_0(ci, B_EXIT);
 	succeed;
+#endif
+#if O_CATCHTHROW
+      } else if ( functor == FUNCTOR_dthrow1 )
+      { Output_0(ci, B_THROW);
+	succeed;
 #endif
       }
       Output_1(ci, call, (code) proc);
@@ -1161,11 +1175,11 @@ operator.
     } else if ( *arg == ATOM_fail )
     { Output_0(ci, I_FAIL);
     } else
-    { FunctorDef fdef = lookupFunctorDef(*arg, 0);
+    { functor_t fdef = lookupFunctorDef(*arg, 0);
       code cproc = (code) lookupProcedure(fdef, tm);
 
 #ifdef O_INLINE_FOREIGNS
-      if ( true(fdef, INLINE_F) )
+      if ( true(valueFunctor(fdef), INLINE_F) )
       { Output_1(ci, I_CALL_FV0, cproc);
       } else
 #endif /*O_INLINE_FOREIGNS*/
@@ -1209,7 +1223,7 @@ OUT-OF-DATE: now pushes *numbers* rather then tagged Prolog data structures.
 static int
 compileArith(Word arg, compileInfo *ci)
 { code a_func;
-  FunctorDef fdef = functorTerm(*arg);
+  functor_t fdef = functorTerm(*arg);
 
   if      ( fdef == FUNCTOR_ar_equals2 )	a_func = A_EQ;	/* =:= */
   else if ( fdef == FUNCTOR_ar_not_equal2 )	a_func = A_NE;	/* =\= */
@@ -1240,7 +1254,7 @@ compileArith(Word arg, compileInfo *ci)
 
 
 static bool
-compileArithArgument(register Word arg, register compileInfo *ci)
+compileArithArgument(Word arg, compileInfo *ci)
 { int index;
 
   deRef(arg);
@@ -1259,7 +1273,7 @@ compileArithArgument(register Word arg, register compileInfo *ci)
     succeed;
   }
 					/* variable */
-  if ( isTerm(*arg) && (index = isIndexedVarTerm(*arg)) >= 0 )
+  if ( (index = isIndexedVarTerm(*arg)) >= 0 )
   { int first = isFirstVar(ci->used_var, index);
 
     if ( index < ci->arity )		/* shared in the head */
@@ -1284,8 +1298,8 @@ compileArithArgument(register Word arg, register compileInfo *ci)
   if ( isVar(*arg) )			/* void variable */
     return warning("Compiler: void variable in arithmetic expression");
 
-  { FunctorDef fdef;
-    int ar;
+  { functor_t fdef;
+    int n, ar;
     Word a;
 
     if ( isAtom(*arg) )
@@ -1294,23 +1308,23 @@ compileArithArgument(register Word arg, register compileInfo *ci)
       a = NULL;
     } else if ( isTerm(*arg) )
     { fdef = functorTerm(*arg);
-      ar = fdef->arity;
+      ar = arityFunctor(fdef);
       a = argTermP(*arg, 0);      
     } else
       return warning("Illegal argument to arithmic function");
 
     if ( (index = indexArithFunction(fdef, ci->module)) < 0 )
       return warning("%s/%d: unknown arithmetic operator",
-		     stringAtom(fdef->name), fdef->arity);
+		     stringAtom(nameFunctor(fdef)), ar);
 
-    for(; ar > 0; a++, ar--)
+    for(n=0; n<ar; a++, n++)
       TRY( compileArithArgument(a, ci) );
 
-    switch(fdef->arity)
+    switch(ar)
     { case 0:	Output_1(ci, A_FUNC0, index); break;
       case 1:	Output_1(ci, A_FUNC1, index); break;
       case 2:	Output_1(ci, A_FUNC2, index); break;
-      default:  Output_2(ci, A_FUNC,  index, (code) fdef->arity); break;
+      default:  Output_2(ci, A_FUNC,  index, (code) ar); break;
     }
 
     succeed;
@@ -1337,7 +1351,7 @@ assert_term(term_t term, int where, SourceLoc loc)
 { Clause clause;
   Procedure proc;
   Definition def;
-  Module source_module = (loc ? modules.source : (Module) NULL);
+  Module source_module = (loc ? LD->modules.source : (Module) NULL);
   Module module = source_module;
   term_t tmp  = PL_new_term_ref();
   term_t head = PL_new_term_ref();
@@ -1506,7 +1520,7 @@ word
 pl_redefine_system_predicate(term_t pred)
 { Procedure proc;
   Module m = NULL;
-  FunctorDef fd;
+  functor_t fd;
   term_t head = PL_new_term_ref();
 
   if ( !PL_strip_module(pred, &m, head) ||
@@ -1544,7 +1558,7 @@ arg1Key(Clause clause, word *key)
     switch(c)
     { case H_FUNCTOR:
       case H_RFUNCTOR:
-	*key = ((FunctorDef)*PC)->functor;
+	*key = ((functor_t)*PC);
         succeed;
       case H_CONST:
 	*key = *PC;
@@ -1554,7 +1568,7 @@ arg1Key(Clause clause, word *key)
         succeed;
       case H_LIST:
       case H_RLIST:
-	*key = FUNCTOR_dot2->functor;
+	*key = FUNCTOR_dot2;
         succeed;
       case H_INTEGER:
       case H_FLOAT:			/* tbd */
@@ -1607,7 +1621,7 @@ typedef struct
 
 forwards bool	decompile_head(Clause, term_t, decompileInfo *);
 forwards bool	decompileBody(decompileInfo *, code, Code);
-forwards void	build_term(FunctorDef, decompileInfo *);
+forwards void	build_term(functor_t, decompileInfo *);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 decompileHead()  is  public  as  it  is   needed  to  update  the  index
@@ -1700,7 +1714,7 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
 
   DEBUG(5, Sdprintf("Decompiling head of %s\n", predicateName(def)));
   arity = def->functor->arity;
-  TRY( PL_unify_functor(head, def->functor) );
+  TRY( PL_unify_functor(head, def->functor->functor) );
   if ( arity > 0 )
     get_arg_ref(head, argp);
   PC = clause->codes;
@@ -1766,7 +1780,7 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
 	  continue;
 	}
       case H_FUNCTOR:
-	{ FunctorDef fdef = (FunctorDef) XR(*PC++);
+	{ functor_t fdef = (functor_t) XR(*PC++);
 	  term_t t2;
 
       common_functor:
@@ -1782,7 +1796,7 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
           goto common_functor;
 	}
       case H_RFUNCTOR:
-	{ FunctorDef fdef = (FunctorDef) XR(*PC++);
+	{ functor_t fdef = (functor_t) XR(*PC++);
 
       common_rfunctor:
 	  TRY(PL_unify_functor(argp, fdef));
@@ -1977,7 +1991,7 @@ decompileBody(register decompileInfo *di, code end, Code until)
 			    setVar(*ARGP++);
 			    continue;
       case B_FUNCTOR:
-      { FunctorDef fdef = (FunctorDef)XR(*PC++);
+      { functor_t fdef = (functor_t)XR(*PC++);
 
       common_bfunctor:
 	*ARGP = globalFunctor(fdef);
@@ -1991,7 +2005,7 @@ decompileBody(register decompileInfo *di, code end, Code until)
         goto common_bfunctor;
       }
       case B_RFUNCTOR:
-      { FunctorDef fdef = (FunctorDef)XR(*PC++);
+      { functor_t fdef = (functor_t)XR(*PC++);
 
       common_brfunctor:
 	*ARGP = globalFunctor(fdef);
@@ -2016,7 +2030,7 @@ decompileBody(register decompileInfo *di, code end, Code until)
       			    PC++;
 			    continue;
 #endif /* O_COMPILE_ARITH */
-      { FunctorDef f;
+      { functor_t f;
 #if O_COMPILE_ARITH
 	case A_LT:	    f = FUNCTOR_smaller2;	goto f_common;
 	case A_LE:	    f = FUNCTOR_smaller_equal2;	goto f_common;
@@ -2029,6 +2043,9 @@ decompileBody(register decompileInfo *di, code end, Code until)
 #if O_BLOCK
 	case I_CUT_BLOCK:   f = FUNCTOR_dcut1;		goto f_common;
 	case B_EXIT:	    f = FUNCTOR_dexit2;		goto f_common;
+#endif
+#if O_CATCHTHROW
+	case B_THROW:	    f = FUNCTOR_dthrow1;	goto f_common;
 #endif
         case I_USERCALLN:   f = lookupFunctorDef(ATOM_call, *PC++ + 1);
 							goto f_common;
@@ -2048,7 +2065,7 @@ decompileBody(register decompileInfo *di, code end, Code until)
 			    continue;
       case I_DEPART:
       case I_CALL:        { Procedure proc = (Procedure)XR(*PC++);
-			    build_term(proc->definition->functor, di);
+			    build_term(proc->definition->functor->functor, di);
 			    pushed++;
 			    continue;
 			  }
@@ -2067,7 +2084,7 @@ decompileBody(register decompileInfo *di, code end, Code until)
 	  
 	  *ARGP++ = makeVarRef(index);
 	}
-	build_term(((Procedure)XR(*PC))->definition->functor, di);
+	build_term(((Procedure)XR(*PC))->definition->functor->functor, di);
 	pushed++;
 	PC += vars+1;
 	continue;
@@ -2155,18 +2172,17 @@ stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-build_term(register FunctorDef f, register decompileInfo *di)
+build_term(functor_t f, decompileInfo *di)
 { word term;
-  int arity;
-  register Word a;
+  int arity = arityFunctor(f);
+  Word a;
 
-  if ( f->arity == 0 )
-  { *ARGP++ = f->name;
+  if ( arity == 0 )
+  { *ARGP++ = nameFunctor(f);
     return;
   }    
 
   term = globalFunctor(f);
-  arity = f->arity;
   a = argTermP(term, arity-1);
 
   ARGP--;
@@ -2187,7 +2203,7 @@ build_term(register FunctorDef f, register decompileInfo *di)
 #undef ARGP
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-unify_definition(?Head, +Def, -TheHead)
+unify_definition(?Head, +Def, -TheHead, flags)
     Given some definition, unify its Prolog reference (i.e. its head with
     optional module specifier) with ?Head.  If TheHead is specified, the
     plain head (i.e. without module specifier) will be referenced from
@@ -2196,11 +2212,26 @@ unify_definition(?Head, +Def, -TheHead)
     This function properly deals with module-inheritance, etc.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+static int
+unify_functor(term_t t, functor_t fd, int how)
+{ if ( how&GP_NAMEARITY )
+  { FunctorDef fdef = valueFunctor(fd);
+
+    return PL_unify_term(t,
+			 PL_FUNCTOR, FUNCTOR_divide2,
+			   PL_ATOM, fdef->name,
+			   PL_INTEGER, fdef->arity);
+  } else
+  { return PL_unify_functor(t, fd);
+  }
+}
+
+
 int
-unify_definition(term_t head, Definition def, term_t thehead)
+unify_definition(term_t head, Definition def, term_t thehead, int how)
 { if ( PL_is_variable(head) )
   { if ( def->module == MODULE_user )
-    { PL_unify_functor(head, def->functor);
+    { unify_functor(head, def->functor->functor, how);
       if ( thehead )
 	PL_put_term(thehead, head);
     } else
@@ -2210,7 +2241,7 @@ unify_definition(term_t head, Definition def, term_t thehead)
       PL_get_arg(1, head, tmp);
       PL_unify_atom(tmp, def->module->name);
       PL_get_arg(2, head, tmp);
-      PL_unify_functor(tmp, def->functor);
+      unify_functor(tmp, def->functor->functor, how);
       if ( thehead )
 	PL_put_term(thehead, tmp);
     }
@@ -2224,7 +2255,7 @@ unify_definition(term_t head, Definition def, term_t thehead)
 	 !isSuperModule(def->module, m) )
       fail;
 
-    if ( PL_unify_functor(h, def->functor) )
+    if ( unify_functor(h, def->functor->functor, how) )
     { if ( thehead )
 	PL_put_term(thehead, h);
       succeed;
@@ -2239,85 +2270,99 @@ word
 pl_clause4(term_t p, term_t term, term_t ref, term_t bindings, word h)
 { Procedure proc;
   Definition def;
-  Clause clause;
   ClauseRef cref;
+  Word argv;
   Module module = NULL;
 
-  if ( ForeignControl(h) == FRG_CUTTED )
-    succeed;
+  switch( ForeignControl(h) )
+  { case FRG_FIRST_CALL:
+    { Clause clause;
 
-  if ( PL_get_pointer(ref, (void **)&clause) ) /* clause(H, B, 2733843) */
-  { Module defModule;
-    term_t tmp  = PL_new_term_ref();
-    term_t head = PL_new_term_ref();
-    term_t body = PL_new_term_ref();
-    FunctorDef f;
-
-    if ( !inCore(clause) || !isClause(clause) )
-      return warning("clause/3: Invalid reference");
-	
-    if ( !decompile(clause, term, bindings) )
-      fail;
-
-    proc = clause->procedure;
-    def = proc->definition;
-    defModule = def->module;
-
-    if ( PL_get_functor(term, &f) && f == FUNCTOR_module2 )
-    { PL_strip_module(p, &module, tmp);
-      if ( module != defModule )
-	fail;
-    }
-
-    if ( !unify_definition(p, def, tmp) )
-      fail;
-
-    get_head_and_body_clause(term, head, body, NULL);
-
-    return PL_unify(tmp, head);
-  } else
-  { Word argv;
-
-    if ( ForeignControl(h) == FRG_FIRST_CALL)
-    { if ( !get_procedure(p, &proc, 0, GP_FIND) ||
+      if ( PL_get_pointer(ref, (void **)&clause) ) /* clause(H, B, 2733843) */
+      { Module defModule;
+	term_t tmp  = PL_new_term_ref();
+	term_t head = PL_new_term_ref();
+	term_t body = PL_new_term_ref();
+	functor_t f;
+    
+	if ( !inCore(clause) || !isClause(clause) )
+	  return warning("clause/3: Invalid reference");
+	    
+	if ( !decompile(clause, term, bindings) )
+	  fail;
+    
+	proc = clause->procedure;
+	def = proc->definition;
+	defModule = def->module;
+    
+	if ( PL_get_functor(term, &f) && f == FUNCTOR_module2 )
+	{ PL_strip_module(p, &module, tmp);
+	  if ( module != defModule )
+	    fail;
+	}
+    
+	if ( !unify_definition(p, def, tmp, 0) )
+	  fail;
+    
+	get_head_and_body_clause(term, head, body, NULL);
+    
+	return PL_unify(tmp, head);
+      }
+      if ( !get_procedure(p, &proc, 0, GP_FIND) ||
 	   true(proc->definition, FOREIGN) )
 	fail;
       def = proc->definition;
       cref = def->definition.clauses;
-    } else
+      enterDefinition(def);		/* reference the predicate */
+      break;
+    }
+    case FRG_REDO:
     { cref = ForeignContextPtr(h);
       proc = cref->clause->procedure;
       def  = proc->definition;
+      break;
     }
-  
-    if ( def->functor->arity > 0 )
-    { term_t head = PL_new_term_ref();
+    case FRG_CUTTED:
+    default:
+    { cref = ForeignContextPtr(h);
+      def  = cref->clause->procedure->definition;
 
-      PL_strip_module(p, &module, head);
-      argv = valTermRef(head);
-      deRef(argv);
-      argv = argTermP(*argv, 0);
-    } else
-      argv = NULL;
-
-    for(; cref; cref = cref->next)
-    { bool det;
-  
-      if ( !(cref = findClause(cref, argv, def, &det)) )
-	fail;
-  
-      if ( !decompile(cref->clause, term, bindings) )
-	continue;
-      if ( !PL_unify_pointer(ref, cref->clause) )
-	continue;
-  
-      if ( det == TRUE )
-	succeed;
-
-      ForeignRedoPtr(cref->next);
+      leaveDefinition(def);
+      succeed;
     }
   }
 
+  if ( def->functor->arity > 0 )
+  { term_t head = PL_new_term_ref();
+
+    PL_strip_module(p, &module, head);
+    argv = valTermRef(head);
+    deRef(argv);
+    argv = argTermP(*argv, 0);
+  } else
+    argv = NULL;
+
+  for(; cref; cref = cref->next)
+  { bool det;
+
+    if ( !(cref = findClause(cref, argv, def, &det)) )
+    { leaveDefinition(def);
+      fail;
+    }
+
+    if ( !decompile(cref->clause, term, bindings) )
+      continue;
+    if ( !PL_unify_pointer(ref, cref->clause) )
+      continue;
+
+    if ( det == TRUE )
+    { leaveDefinition(def);
+      succeed;
+    }
+
+    ForeignRedoPtr(cref->next);
+  }
+
   fail;
 }
 
@@ -2361,7 +2406,7 @@ pl_nth_clause(term_t p, term_t n, term_t ref, word h)
     for( cref = def->definition.clauses, i=1; cref; cref = cref->next, i++)
     { if ( cref->clause == clause )
       { if ( !PL_unify_integer(n, i) ||
-	     !unify_definition(p, def, 0) )
+	     !unify_definition(p, def, 0, 0) )
 	  fail;
 
 	succeed;
@@ -2452,7 +2497,7 @@ wouldBindToDefinition(Definition from, Definition to)
   Definition def = from;
   Procedure proc;
 
-  while(m)
+  for(;;)
   { if ( def )
     { if ( def == to )			/* found it */
 	succeed;
@@ -2463,9 +2508,11 @@ wouldBindToDefinition(Definition from, Definition to)
 	fail;
     }
 
-    m = m->super;			/* see in super module */
-    proc = isCurrentProcedure(from->functor, m);
-    def = proc ? proc->definition : (Definition)NULL;
+    if ( (m = m->super) )
+    { proc = isCurrentProcedure(from->functor->functor, m);
+      def = proc ? proc->definition : (Definition)NULL;
+    } else
+      break;
   }
 
   fail;
@@ -2507,11 +2554,11 @@ pl_xr_member(term_t ref, term_t term, word h)
       switch(codeTable[op].argtype)
       { case CA1_PROC:
 	{ Procedure proc = (Procedure) *PC;
-	  rval = unify_definition(term, proc->definition, 0);
+	  rval = unify_definition(term, proc->definition, 0, 0);
 	  break;
 	}
 	case CA1_FUNC:
-	{ FunctorDef fd = (FunctorDef) *PC;
+	{ functor_t fd = (functor_t) *PC;
 	  rval = PL_unify_functor(term, fd);
 	  break;
 	}
@@ -2542,7 +2589,7 @@ pl_xr_member(term_t ref, term_t term, word h)
     fail;
   } else				/* instantiated */
   { Procedure proc;
-    FunctorDef fd;
+    functor_t fd;
 
     if ( PL_is_atomic(term) )
     { while( PC < end )
@@ -2575,7 +2622,7 @@ pl_xr_member(term_t ref, term_t term, word h)
       { code op = decode(*PC);
 
 	if ( codeTable[op].argtype == CA1_FUNC )
-	{ FunctorDef fa = (FunctorDef)PC[1];
+	{ functor_t fa = (functor_t)PC[1];
 
 	  if ( fa == fd )
 	    succeed;
@@ -2604,7 +2651,7 @@ wamListClause(Clause clause)
 
   while( bp < ep )
   { code op = decode(*bp);
-    CodeInfo ci;
+    const code_info *ci;
     int n = 0;
     int isbreak;
 
@@ -2661,9 +2708,10 @@ wamListClause(Clause clause)
 	    break;
 	  }
 	  case CA1_FUNC:
-	  { FunctorDef f = (FunctorDef) *bp++;
+	  { functor_t f = (functor_t) *bp++;
+	    FunctorDef fd = valueFunctor(f);
 	    n++;
-	    Putf(" %s/%d", stringAtom(f->name), f->arity);
+	    Putf(" %s/%d", stringAtom(fd->name), fd->arity);
 	    break;
 	  }
 	  case CA1_DATA:
@@ -2700,6 +2748,13 @@ wamListClause(Clause clause)
 	    Putf(" %g", v.f);
 	    break;
 	  }
+	  case CA1_STRING:
+	  { word m = *bp++;
+	    int  n = wsizeofInd(m);
+	    Putf(" \"%s\"", (char *)bp);
+	    bp += n;
+	    break;
+	  }
 	}
         for(; n < codeTable[op].arguments; n++ )
 	  Putf("%s%d", n == 0 ? " " : ", ", *bp++);
@@ -2742,7 +2797,7 @@ pl_fetch_vm(term_t ref, term_t offset, term_t noffset, term_t instruction)
   int pcoffset;
   Code PC;
   code op;
-  CodeInfo ci;
+  const code_info *ci;
 
   if ( !PL_get_pointer(ref, (void **)&clause) ||
        !inCore(clause) || !isClause(clause) ||
@@ -2860,7 +2915,7 @@ pl_clause_term_position(term_t ref, term_t pc, term_t locterm)
 
   while( PC < loc )
   { code op = decode(*PC++);
-    CodeInfo ci;
+    const code_info *ci;
 
     if ( op == D_BREAK )
       op = decode(replacedBreak(PC-1));
@@ -3089,7 +3144,7 @@ pl_break_pc(term_t ref, term_t pc, term_t nextpc, control_t h)
 		 *         BREAK-POINTS		*
 		 *******************************/
 
-static Table breakTable;
+#define breakTable (GD->comp.breakpoints)
 
 typedef struct
 { Clause	clause;			/* Associated clause */
@@ -3214,14 +3269,14 @@ pl_current_break(term_t ref, term_t pc, control_t h)
       succeed;
   }
 
-  for( ; symb; symb = nextHTable(moduleTable, symb) )
+  for( ; symb; symb = nextHTable(breakTable, symb) )
   { BreakPoint bp = (BreakPoint) symb->value;
 
     { fid_t cid = PL_open_foreign_frame();
 
       if ( PL_unify_pointer(ref, bp->clause) &&
 	   PL_unify_integer(pc,  bp->offset) )
-      { if ( !(symb = nextHTable(moduleTable, symb)) )
+      { if ( !(symb = nextHTable(breakTable, symb)) )
 	  succeed;
 
 	ForeignRedoPtr(symb);
