@@ -106,9 +106,9 @@ Indirect data
 
 #define STG_MASK	(0x3<<3)
 #define STG_STATIC	(0x0<<3)	/* storage masks */
-#define STG_HEAP	(0x1<<3)
-#define STG_GLOBAL	(0x2<<3)
-#define STG_LOCAL	(0x3<<3)
+#define STG_GLOBAL	(0x1<<3)
+#define STG_LOCAL	(0x2<<3)
+#define STG_RESERVED	(0x3<<3)
 
 #define STG_INLINE	STG_STATIC
 #define STG_TRAIL	STG_STATIC
@@ -116,9 +116,6 @@ Indirect data
 #define MARK_MASK	(0x1<<5)	/* GC mark */
 #define FIRST_MASK	(0x2<<5)	/* GC first mark */
 
-GLOBAL unsigned long	base_addresses[STG_MASK+1];
-GLOBAL buffer		functor_array;
-
 #define tag(w)		((w) & TAG_MASK)
 #define storage(w)	((w) & STG_MASK)
 #define valPtr2(w, s)	((Word)(((w) >> 5) + base_addresses[s]))
@@ -162,10 +159,10 @@ extern const unsigned int tagtypeex[];
 		 *	COMPOUNDS AND LISTS	*
 		 *******************************/
 
-#define functorTerm(w)	valueFunctor(valueTerm(w)->definition)
+#define functorTerm(w)	valueTerm(w)->definition
 #define arityTerm(w)	arityFunctor(valueTerm(w)->definition)
 #define valueTerm(w)	((Functor)valPtr(w))
-#define hasFunctor(w,f) (isTerm(w) && valueTerm(w)->definition == (f)->functor)
+#define hasFunctor(w,f) (isTerm(w) && valueTerm(w)->definition == (f))
 #define argTerm(w, n)	(((Functor)valPtr(w))->arguments[n])
 #define argTermP(w, n)	(&argTerm(w, n))
 
@@ -189,15 +186,15 @@ extern const unsigned int tagtypeex[];
 #define valString(w)	((char *)valIndirectP(w))
 #define valBignum(w)	(*(long *)valIndirectP(w))
 #define isBignum(w)	(isInteger(w) && storage(w) != STG_INLINE)
-#define isTaggedInt(w)	(isInteger(w) && storage(w) == STG_INLINE)
+#define isTaggedInt(w)	(tagex(w) == (TAG_INTEGER|STG_INLINE))
+			/* == (isInteger(w) && storage(w) == STG_INLINE) */
 
 		 /*******************************
 		 *	       VALUES		*
 		 *******************************/
 
-#define atomValue(w)	(storage(w) == STG_HEAP ? \
-				(Atom)valPtr2(w, STG_HEAP) :  \
-				&atoms[valInt(w)])
+#define indexAtom(w)	((w)>>LMASK_BITS)
+#define atomValue(w)	fetchBuffer(&atom_array, indexAtom(w), Atom)
 #define stringAtom(w)	(atomValue(w)->name)
 #define valInteger(w)	(storage(w) == STG_INLINE ? valInt(w) : valBignum(w))
 
@@ -207,11 +204,15 @@ extern const unsigned int tagtypeex[];
 
 #define F_ARITY_BITS	5		/* upto 32 inlined arity */
 #define F_ARITY_MASK	((1<<F_ARITY_BITS)-1)
+#define MK_FUNCTOR(n, a) (((((n)<<F_ARITY_BITS)|(a))<<LMASK_BITS) | \
+			  TAG_ATOM|STG_GLOBAL)
+#define functorHashValue(f, n)	((f)>>(LMASK_BITS) & ((n)-1))
 #define indexFunctor(w)	((w)>>(LMASK_BITS+F_ARITY_BITS))
 #define valueFunctor(w) fetchBuffer(&functor_array,indexFunctor(w),FunctorDef)
-#define _arityFunc_(w)	((w) >> LMASK_BITS & F_ARITY_MASK)
+#define _arityFunc_(w)	(((w) >> LMASK_BITS) & F_ARITY_MASK)
 #define arityFunctor(w) (_arityFunc_(w)!=F_ARITY_MASK ? _arityFunc_(w) \
 						      : valueFunctor(w)->arity)
+#define isAtomFunctor(w) (arityFunctor(w) == 0)
 #define nameFunctor(w)	(valueFunctor(w)->name)
 
 		 /*******************************
@@ -230,7 +231,7 @@ extern const unsigned int tagtypeex[];
 #define MAXTAGGEDPTR	((1L<<((8*sizeof(word))-5-1)) - 1)
 
 #define consInt(n)	(((word)(n)<<7) | TAG_INTEGER)
-#if 0					/* O_DEBUG */
-#define consPtr(p,ts)	((word)((((word)(p)-base_addresses[s])<<5)|(ts)))
+#if !O_DEBUG
+#define consPtr(p,ts)	((word)((((word)(p)-base_addresses[(ts)&STG_MASK])<<5)|(ts)))
 #endif
 
