@@ -1,4 +1,4 @@
-/*  $Id: pl-dde.c,v 1.10 1996/10/30 17:32:33 jan Exp $
+/*  $Id: pl-dde.c,v 1.12 1997/08/26 13:34:52 jan Exp $
 
     Part of SWI-Prolog.
     Contributed by Dwig (dwig@markv.com)
@@ -62,13 +62,13 @@ static HCONV conv_handle[MAX_CONVERSATIONS];
 static HCONV server_handle[MAX_CONVERSATIONS];
 static DWORD ddeInst;			/* Instance of this process */
 
-static Module	  MODULE_dde;		/* win_dde */
-static FunctorDef FUNCTOR_dde_connect3;
-static FunctorDef FUNCTOR_dde_connect_confirm3;
-static FunctorDef FUNCTOR_dde_disconnect1;
-static FunctorDef FUNCTOR_dde_request4;
-static FunctorDef FUNCTOR_dde_execute3;
-static FunctorDef FUNCTOR_error1;
+static Module	 MODULE_dde;		/* win_dde */
+static functor_t FUNCTOR_dde_connect3;
+static functor_t FUNCTOR_dde_connect_confirm3;
+static functor_t FUNCTOR_dde_disconnect1;
+static functor_t FUNCTOR_dde_request4;
+static functor_t FUNCTOR_dde_execute3;
+static functor_t FUNCTOR_error1;
 
 static char *
 dde_error_message(int errn)
@@ -234,7 +234,6 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
      { fid_t cid = PL_open_foreign_frame();
        term_t argv = PL_new_term_refs(3);
        predicate_t pred = PL_pred(FUNCTOR_dde_connect_confirm3, MODULE_dde);
-       int rval;
        int plhandle;
 
        if ( (plhandle = allocServerHandle(hconv)) >= 0 )
@@ -246,30 +245,28 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
 	 PL_put_atom(   argv+1, hszToAtom(hsz1)); /* service */
 	 PL_put_integer(argv+2, plhandle);
 
-	 rval = PL_call_predicate(MODULE_dde, TRUE, pred, argv);
+	 PL_call_predicate(MODULE_dde, TRUE, pred, argv);
 	 PL_discard_foreign_frame(cid);
        } else
        { warning("No more DDE server handles");
-	 rval = FALSE;
        }
 
-       return (void *)rval; 
+       return NULL;
      }
      case XTYP_DISCONNECT:
      { fid_t cid = PL_open_foreign_frame();
        term_t argv = PL_new_term_refs(1);
        predicate_t pred = PL_pred(FUNCTOR_dde_disconnect1, MODULE_dde);
-       int rval;
        int plhandle = findServerHandle(hconv);
        
        if ( plhandle >= 0 && plhandle < MAX_CONVERSATIONS )
 	 server_handle[plhandle] = (HCONV)NULL;
 
        PL_put_integer(argv+0, plhandle);
-       rval = PL_call_predicate(MODULE_dde, TRUE, pred, argv);
+       PL_call_predicate(MODULE_dde, TRUE, pred, argv);
        PL_discard_foreign_frame(cid);
 
-       return (void *)rval;
+       return NULL;
      }
      case XTYP_EXECUTE:
      { int plhandle = findServerHandle(hconv);
