@@ -1,4 +1,4 @@
-/*  $Id: pl-ext.c,v 1.57 1997/03/10 09:36:30 jan Exp $
+/*  $Id: pl-ext.c,v 1.61 1997/10/28 13:40:52 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -45,8 +45,8 @@ Flags almost always is TRACE_ME.  Additional common flags:
 
 #define FRG(n, a, f, flags) { n, f, flags, a }
 
-static struct foreign {
-  char		*name;
+static const struct foreign {
+  const char *  name;
   Func		function;
   unsigned long flags;
   int		arity;
@@ -121,12 +121,12 @@ static struct foreign {
   FRG("$depth_limit_false",	4, pl_depth_limit_false,	TRACE_ME),
 #endif
 
-  FRG("display",		1, pl_display,			TRACE_ME),
-  FRG("displayq",		1, pl_displayq,			TRACE_ME),
+  FRG("write_canonical",	1, pl_write_canonical,		TRACE_ME),
+  FRG("write_term",		2, pl_write_term,		TRACE_ME),
+  FRG("write_term",		3, pl_write_term3,		TRACE_ME),
   FRG("write",			1, pl_write,			TRACE_ME),
   FRG("writeq",			1, pl_writeq,			TRACE_ME),
   FRG("print",			1, pl_print,			TRACE_ME),
-  FRG("$print",			2, pl_dprint,		   META|TRACE_ME),
 
   FRG("read_variables",		2, pl_read_variables,		TRACE_ME),
   FRG("read_term",		2, pl_read_term,		TRACE_ME),
@@ -136,6 +136,7 @@ static struct foreign {
   FRG("read_clause",		1, pl_read_clause,		TRACE_ME),
   FRG("read_clause",		2, pl_read_clause2,		TRACE_ME),
   FRG("$raw_read",		1, pl_raw_read,			TRACE_ME),
+  FRG("$raw_read",		2, pl_raw_read2,		TRACE_ME),
   FRG("current_op",		3, pl_current_op,	   NDET|TRACE_ME),
   FRG("current_atom",		1, pl_current_atom,	   NDET|TRACE_ME),
   FRG("current_functor",	2, pl_current_functor,	   NDET|TRACE_ME),
@@ -181,6 +182,7 @@ static struct foreign {
   FRG("$qlf_assert_clause",	1, pl_qlf_assert_clause,	TRACE_ME),
   FRG("$qlf_info",		4, pl_qlf_info,			TRACE_ME),
 
+  FRG("abolish",    		1, pl_abolish1,		   META|TRACE_ME),
   FRG("abolish",    		2, pl_abolish,		   META|TRACE_ME),
   FRG("$clause",    		3, pl_clause,	      NDET|META|TRACE_ME),
   FRG("$clause",	        4, pl_clause4,	      NDET|META|TRACE_ME),
@@ -301,7 +303,6 @@ static struct foreign {
   FRG("dwim_match",		3, pl_dwim_match,		TRACE_ME),
   FRG("$dwim_predicate",	2, pl_dwim_predicate,	   NDET|TRACE_ME),
 
-  FRG("$novice",		2, pl_novice,			TRACE_ME),
   FRG("protocol",		1, pl_protocol,			TRACE_ME),
   FRG("protocola",		1, pl_protocola,		TRACE_ME),
   FRG("noprotocol",		0, pl_noprotocol,		TRACE_ME),
@@ -360,10 +361,6 @@ static struct foreign {
   FRG("substring",		4, pl_substring,		TRACE_ME),
 #endif /* O_STRING */
 
-  FRG("save",			2, pl_save,			TRACE_ME),
-  FRG("restore",		1, pl_restore,			TRACE_ME),
-  FRG("$save_program",		2, pl_save_program,		TRACE_ME),
-
   FRG("is_list",		1, pl_is_list,			TRACE_ME),
   FRG("proper_list",		1, pl_proper_list,		TRACE_ME),
   FRG("$length",		2, pl_length,			TRACE_ME),
@@ -421,8 +418,7 @@ static struct foreign {
   FRG("write",			2, pl_write2,			TRACE_ME),
   FRG("writeq",			2, pl_writeq2,			TRACE_ME),
   FRG("print",			2, pl_print2,			TRACE_ME),
-  FRG("display",		2, pl_display2,			TRACE_ME),
-  FRG("displayq",		2, pl_displayq2,		TRACE_ME),
+  FRG("write_canonical",	2, pl_write_canonical2,		TRACE_ME),
   FRG("format",			3, pl_format3,			TRACE_ME),
 
   FRG("tty_get_capability",	3, pl_tty_get_capability,	TRACE_ME),
@@ -445,16 +441,25 @@ static struct foreign {
 };
 
 
-typedef struct extension_cell *ExtensionCell;
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+The extensions chain is used   to allow calling PL_register_extensions()
+*before* PL_initialise() to get foreign   extensions in embedded systems
+defined before the state is loaded, so executing directives in the state
+can use foreign extensions.
+
+If an extension is registered before the  system extension is loaded, it
+will be added to the chain. Right  after the system registers the system
+predicates, the extensions will be registered.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 struct extension_cell
 { PL_extension *extensions;
   ExtensionCell next;
 };
 
-static ExtensionCell extensions = NULL;
-static ExtensionCell ext_tail = NULL;
-static extensions_loaded = FALSE;
+#define ext_head		(GD->foreign._ext_head)
+#define ext_tail		(GD->foreign._ext_tail)
+#define extensions_loaded	(GD->foreign._loaded)
 
 static void
 bindExtensions(PL_extension *e)
@@ -491,7 +496,7 @@ PL_register_extensions(PL_extension *e)
     { ext_tail->next = cell;
       ext_tail = cell;
     } else
-    { extensions = ext_tail = cell;
+    { ext_head = ext_tail = cell;
     }
   }
 }
@@ -499,11 +504,12 @@ PL_register_extensions(PL_extension *e)
 
 void
 initBuildIns(void)
-{ struct foreign *f;
+{ const struct foreign *f;
   Definition def;
+  ExtensionCell ecell;
 
   for(f = &foreigns[0]; f->name; f++)
-  { FunctorDef fdef = lookupFunctorDef(lookupAtom(f->name), f->arity);
+  { functor_t fdef = lookupFunctorDef(lookupAtom(f->name), f->arity);
 
     def = lookupProcedure(fdef, MODULE_system)->definition;
     set(def, FOREIGN|SYSTEM|LOCKED);
@@ -514,16 +520,21 @@ initBuildIns(void)
     def->indexCardinality = 0;
     if ( false(def, NONDETERMINISTIC) && 
 	 f->arity <= 2 )
-      set(fdef, INLINE_F);
+      set(valueFunctor(fdef), INLINE_F);
   }
 
-  PROCEDURE_alt1 = lookupProcedure(FUNCTOR_alt1, MODULE_system);
+  PROCEDURE_alt1	     = lookupProcedure(FUNCTOR_alt1, MODULE_system);
   PROCEDURE_garbage_collect0 = lookupProcedure(FUNCTOR_dgarbage_collect1,
 					       MODULE_system);
-  PROCEDURE_block3 = lookupProcedure(FUNCTOR_block3, MODULE_system);
-  PROCEDURE_true0  = lookupProcedure(FUNCTOR_true0, MODULE_system);
-  PROCEDURE_fail0  = lookupProcedure(FUNCTOR_fail0, MODULE_system);
+  PROCEDURE_block3	     = lookupProcedure(FUNCTOR_block3, MODULE_system);
+  PROCEDURE_catch3           = lookupProcedure(FUNCTOR_catch3, MODULE_system);
+  PROCEDURE_true0            = lookupProcedure(FUNCTOR_true0, MODULE_system);
+  PROCEDURE_fail0            = lookupProcedure(FUNCTOR_fail0, MODULE_system);
+  PROCEDURE_print_message2   = lookupProcedure(FUNCTOR_print_message2,
+					       MODULE_system);
+
+  for( ecell = ext_head; ecell; ecell = ecell->next )
+    bindExtensions(ecell->extensions);
 
-  bindExtensions(PL_extensions);
   extensions_loaded = TRUE;
 }
