@@ -1,4 +1,4 @@
-/*  $Id: pl-fmt.c,v 1.21 1997/02/10 12:56:46 jan Exp $
+/*  $Id: pl-fmt.c,v 1.25 1997/08/07 07:57:54 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -47,7 +47,7 @@ struct rubber
   Char pad;				/* padding character */
 };
 
-static Table format_predicates;		/* Prolog defined fromatting */
+#define format_predicates (GD->format.predicates)
 
 forwards int	update_column(int, Char);
 forwards bool	do_format(const char *fmt, int argc, term_t argv);
@@ -60,15 +60,15 @@ forwards void	emit_rubber(char *buf, int, struct rubber *, int);
 
 word
 pl_format_predicate(term_t chr, term_t descr)
-{ int c;
+{ long c;
   Procedure proc;
   Symbol s;
 
-  if ( !PL_get_integer(chr, &c) || c < 0 || c > 255 )
+  if ( !PL_get_long(chr, &c) || c < 0 || c > 255 )
   { char *s;
     
     if ( PL_get_atom_chars(chr, &s) && s[0] && !s[1] )
-      c = s[0];
+      c = s[0] & 0xff;
     else
       return warning("format_predicate/2: illegal character");
   }
@@ -78,13 +78,13 @@ pl_format_predicate(term_t chr, term_t descr)
   if ( proc->definition->functor->arity == 0 )
     return warning("format_predicate/2: predicate must have at least 1 argument");
 
-  if ( format_predicates == NULL )
+  if ( !format_predicates )
     format_predicates = newHTable(8);
   
-  if ( (s = lookupHTable(format_predicates, (Void)c)) )
+  if ( (s = lookupHTable(format_predicates, (void *)c)) )
     s->value = (word) proc;
   else
-    addHTable(format_predicates, (Void)c, proc);
+    addHTable(format_predicates, (void *)c, proc);
 
   succeed;
 }
@@ -224,7 +224,8 @@ do_format(const char *fmt, int argc, term_t argv)
 	    }
 
 	    tellString(&str, BUFSIZE);
-	    qid = PL_open_query(proc->definition->module, FALSE, proc, av);
+	    qid = PL_open_query(proc->definition->module, PL_Q_NODEBUG,
+				proc, av);
 	    PL_next_solution(qid);
 	    PL_close_query(qid);
 	    toldString();
@@ -286,7 +287,7 @@ do_format(const char *fmt, int argc, term_t argv)
 	      case 'r':			/* radix number */
 	      case 'R':			/* Radix number */
 		{ int i;
-		  char *s;
+		  char tmp[50];
 
 		  NEED_ARG;
 		  if ( !PL_get_integer(argv, &i) )
@@ -294,11 +295,11 @@ do_format(const char *fmt, int argc, term_t argv)
 		  SHIFT;
 		  if ( arg == DEFAULT )
 		    arg = 0;
-		  s = ( (*fmt == 'd' || *fmt == 'D')
-			? formatInteger(*fmt == 'D', arg, 10, TRUE, i)
-			: formatInteger(FALSE, 0, arg, *fmt == 'r', i)
-		      );
-		  OUTSTRING(s);			
+		  if ( *fmt == 'd' || *fmt == 'D' )
+		    formatInteger(*fmt == 'D', arg, 10, TRUE, i, tmp);
+		  else
+		    formatInteger(FALSE, 0, arg, *fmt == 'r', i, tmp);
+		  OUTSTRING(tmp);			
 		  fmt++;
 		  break;
 		}
@@ -306,7 +307,7 @@ do_format(const char *fmt, int argc, term_t argv)
 		{ char *s;
 
 		  NEED_ARG;
-		  if ( !PL_get_list_chars(argv, &s, 0) )
+		  if ( !PL_get_chars(argv, &s, CVT_LIST|CVT_STRING) )
 		    ERROR("illegal argument to ~s");
 		  OUTSTRING(s);
 		  SHIFT;
@@ -323,12 +324,15 @@ do_format(const char *fmt, int argc, term_t argv)
 		  char buf[BUFSIZE];
 		  char *str;
 
-	      case 'k':			/* displayq */
-		  f = pl_displayq;	goto pl_common;
+	      case 'k':			/* write_canonical */
+		  f = pl_write_canonical; 
+	          goto pl_common;
 	      case 'p':			/* print */
-		  f = pl_print;		goto pl_common;
+		  f = pl_print;
+	          goto pl_common;
 	      case 'q':			/* writeq */
-		  f = pl_writeq;	goto pl_common;
+		  f = pl_writeq;
+	          goto pl_common;
 	      case 'w':			/* write */
 		  f = pl_write;
 		  pl_common:
