@@ -1,4 +1,4 @@
-/*  $Id: pl-funcs.h,v 1.75 1997/03/10 09:36:34 jan Exp $
+/*  $Id: pl-funcs.h,v 1.86 1997/11/14 16:25:24 jan Exp $
 
     Part of SWI-Prolog
     Designed and implemented by Jan Wielemaker
@@ -12,25 +12,21 @@
 /* pl-alloc.c */
 void		free_heap(Void mem, size_t n);
 volatile void	outOf(Stack s);
+volatile void	outOfCore(void);
 Void		alloc_global(int n);
 Void		alloc_heap(size_t n);
-word		globalFunctor(FunctorDef def);
+void		initMemAlloc(void);
+word		globalFunctor(functor_t def);
 int		sizeString(word w);
 word		globalString(const char *s);
 word		globalNString(long len, const char *s);
-word		heapString(const char *s);
 Word		newTerm(void);
 word		globalReal(real f);
 word		globalLong(long i);
-word		heapLong(long i);
-word		heapReal(real f);
 double		valReal(word w);
 word		globalIndirect(word in);
-word		heapIndirect(word in);
 int		equalIndirect(word r1, word r2);
 word		globalIndirectFromCode(Code *PC);
-int		equalIndirectFromCode(word a, Code *PC);
-void		freeHeapIndirect(word w);
 word		makeRef(Word p);
 #ifndef consPtr
 word		consPtr(void *p, int ts);
@@ -67,7 +63,6 @@ word		pl_between(term_t l, term_t h, term_t n, word b);
 word		pl_succ(term_t n1, term_t n2);
 word		pl_plus(term_t a, term_t b, term_t c);
 int		ar_compare(Number n1, Number n2, int what);
-word		compareNumbers(term_t n1, term_t n2, int what);
 word		pl_lessNumbers(term_t n1, term_t n2);
 word		pl_greaterNumbers(term_t n1, term_t n2);
 word		pl_lessEqualNumbers(term_t n1, term_t n2);
@@ -78,8 +73,8 @@ word		pl_is(term_t v, term_t e);
 word		pl_arithmetic_function(term_t descr);
 word		pl_current_arithmetic_function(term_t f, word h);
 void		initArith(void);
-int		indexArithFunction(FunctorDef fdef, Module m);
-FunctorDef	functorArithFunction(int n);
+int		indexArithFunction(functor_t fdef, Module m);
+functor_t	functorArithFunction(int n);
 bool		ar_func_n(code n, int argc, Number *stack);
 int		valueExpression(term_t p, Number n);
 int		toIntegerNumber(Number n);
@@ -113,7 +108,7 @@ word		pl_wam_list(term_t ref);
 word		pl_fetch_vm(term_t ref, term_t offset, term_t noffset,
 			    term_t instruction);
 int		unify_definition(term_t head, Definition def,
-				 term_t thehead);
+				 term_t thehead, int flags);
 word		pl_clause_term_position(term_t ref, term_t pc, term_t locterm);
 word		pl_break_pc(term_t ref, term_t pc, term_t nextpc, control_t h);
 word		pl_break_at(term_t ref, term_t pc, term_t set);
@@ -154,6 +149,12 @@ word		pl_dwim_predicate(term_t term, term_t dwim, word h);
 /* pl-ext.c */
 void		initBuildIns(void);
 
+/* pl-error.c */
+
+int		PL_error(const char *pred, int arity, const char *msg,
+			 int id, ...);
+char *		tostr(char *buf, const char *fmt, ...);
+
 /* pl-file.c */
 void		initIO(void);
 void		dieIO(void);
@@ -167,12 +168,11 @@ bool		readLine(char *buffer);
 bool		LockStream(void);
 bool		UnlockStream(void);
 bool		Put(int c);
+bool		PutOpenToken(int c);
 IOSTREAM *	PL_current_input(void);
 IOSTREAM *	PL_current_output(void);
 word		Putf(char *fm, ...);
 bool		Puts(const char *str);
-word		vPutf(char *fm, va_list args);
-int		currentInputLine(void);
 word		pl_told(void);
 word		pl_flush(void);
 word		pl_see(term_t f);
@@ -262,6 +262,10 @@ word		makeNum(long i);
 void		finish_foreign_frame();
 void		_PL_put_number(term_t t, Number n);
 int		_PL_unify_number(term_t t, Number n);
+predicate_t	_PL_predicate(const char *name, int arity, const char *module,
+			      predicate_t *bin);
+void		initialiseForeign(int argc, char **argv);
+char *		buffer_string(const char *s, int flags);
 
 /* pl-fmt.c */
 word		pl_format_predicate(term_t chr, term_t descr);
@@ -269,8 +273,8 @@ word		pl_format(term_t fmt, term_t args);
 word		pl_format3(term_t s, term_t fmt, term_t args);
 
 /* pl-funct.c */
-FunctorDef	lookupFunctorDef(atom_t atom, int arity);
-FunctorDef	isCurrentFunctor(atom_t atom, int arity);
+functor_t	lookupFunctorDef(atom_t atom, int arity);
+functor_t	isCurrentFunctor(atom_t atom, int arity);
 void		initFunctors(void);
 int 		checkFunctors(void);
 word		pl_current_functor(term_t name, term_t arity, word h);
@@ -283,6 +287,8 @@ void		resetGC(void);
 Word		findGRef(int n);
 int		growStacks(LocalFrame fr, Code PC, int l, int g, int t);
 void		clearUninitialisedVarsFrame(LocalFrame, Code);
+word		check_foreign(void);	/* O_SECURE stuff */
+
 
 /* pl-glob.c */
 word		pl_wildcard_match(term_t pattern, term_t string);
@@ -290,7 +296,6 @@ word		pl_expand_file_name(term_t f, term_t l);
 
 /* pl-itf.c */
 void		resetForeign(void);
-void		initialiseForeign(int argc, char **argv);
 
 /* pl-list.c */
 word		pl_is_list(term_t list);
@@ -351,28 +356,29 @@ real		CpuTime(void);
 Void		Allocate(long int n);
 long		Random(void);
 char *		canonisePath(char *path);
-char *		OsPath(const char *unixpath);
-char *		PrologPath(char *ospath, char *plpath);
+char *		OsPath(const char *plpath, char *ospath);
+char *		PrologPath(const char *ospath, char *plpath);
 long		LastModifiedFile(char *f);
-bool		ExistsFile(char *path);
-bool		AccessFile(char *path, int mode);
-bool		ExistsDirectory(char *path);
-long		SizeFile(char *path);
+bool		ExistsFile(const char *path);
+bool		AccessFile(const char *path, int mode);
+bool		ExistsDirectory(const char *path);
+long		SizeFile(const char *path);
 int		RemoveFile(const char *path);
-bool		RenameFile(char *old, char *new);
-bool		SameFile(char *f1, char *f2);
+bool		RenameFile(const char *old, const char *new);
+bool		SameFile(const char *f1, const char *f2);
 bool		OpenStream(int fd);
-bool		MarkExecutable(char *name);
-bool		expandVars(char *pattern, char *expanded);
-char *		ExpandOneFile(char *spec);
+bool		MarkExecutable(const char *name);
+bool		expandVars(const char *pattern, char *expanded);
+char *		ExpandOneFile(const char *spec, char *file);
 char *		getwd(char *buf);
-char *		AbsoluteFile(char *spec);
+char *		AbsoluteFile(const char *spec, char *path);
 int		IsAbsolutePath(const char *spec);
 char *		BaseName(char *f);
-char *		DirName(char *f);
-char *		ReadLink(char *f);
-char *		DeRefLink(char *link);
-bool		ChDir(char *path);
+char *		DirName(const char *f, char *buf);
+char *		ReadLink(const char *f, char *buf);
+char *		DeRefLink(const char *link, char *buf);
+bool		ChDir(const char *path);
+atom_t		TemporaryFile(const char *id);
 struct tm *	LocalTime(long int *t);
 Char		GetChar(void);
 void		ResetTty(void);
@@ -381,7 +387,7 @@ bool		PopTty(ttybuf *buf);
 char *		Setenv(char *name, char *value);
 char *		Unsetenv(char *name);
 int		System(char *cmd);
-char *		Symbols(void);
+char *		Symbols(char *buf);
 void		Pause(real time);
 #if __WIN32__
 int		iswin32s(void);
@@ -418,7 +424,7 @@ word		pl_arg(term_t n, term_t t, term_t a, word b);
 word		pl_setarg(term_t n, term_t term, term_t arg);
 int		lengthList(term_t list);
 word		pl_univ(term_t t, term_t l);
-int		numberVars(term_t t, FunctorDef functor, int n);
+int		numberVars(term_t t, functor_t functor, int n);
 word		pl_numbervars(term_t t, term_t atom,
 			      term_t start, term_t end);
 word		pl_free_variables(term_t t, term_t l);
@@ -429,7 +435,7 @@ word		pl_atom_length(term_t w, term_t n);
 word		pl_int_to_atom(term_t number, term_t base,
 			       term_t atom);
 char *		formatInteger(bool split, int div, int radix,
-			      bool small, long int n);
+			      bool small, long n, char *out);
 word		pl_format_number(term_t format, term_t number,
 				 term_t string);
 word		pl_name(term_t atom, term_t string);
@@ -457,9 +463,8 @@ word		pl_fail(void);
 word		pl_true(void);
 word		pl_halt(term_t code);
 word		pl_statistics(term_t k, term_t value);
-int		setFeature(atom_t name, word value);
-int		CSetFeature(char *name, char *value);
-word		getFeature(atom_t name);
+int		setFeature(atom_t name, int type, ...);
+void		CSetFeature(char *name, char *value);
 word		pl_feature(term_t key, term_t value, word h);
 word		pl_set_feature(term_t key, term_t value);
 word		pl_option(term_t key, term_t old, term_t new);
@@ -485,12 +490,12 @@ word		pl_abort(void);
 bool		prolog(atom_t toplevel);
 word		pl_metacut(void);
 int 		trap_gdb(void);
-word		checkData(Word p, int on_heap);
+word		checkData(Word p);
 
 /* pl-proc.c */
-Procedure	lookupProcedure(FunctorDef f, Module m);
-Procedure	isCurrentProcedure(FunctorDef f, Module m);
-Procedure	lookupProcedureToDefine(FunctorDef def, Module m);
+Procedure	lookupProcedure(functor_t f, Module m);
+Procedure	isCurrentProcedure(functor_t f, Module m);
+Procedure	lookupProcedureToDefine(functor_t def, Module m);
 bool		isDefinedProcedure(Procedure proc);
 int		get_procedure(term_t descr, Procedure *proc, term_t he, int f);
 word		pl_current_predicate(term_t name, term_t functor, word h);
@@ -502,11 +507,12 @@ void		freeClauseRef(ClauseRef c);
 ClauseRef	newClauseRef(Clause cl);
 void		gcClausesDefinition(Definition def);
 void		resetReferences(void);
-Procedure	resolveProcedure(FunctorDef f, Module module);
+Procedure	resolveProcedure(functor_t f, Module module);
 Definition	trapUndefined(Definition def);
 word		pl_retract(term_t term, word h);
 word		pl_retractall(term_t head);
 word		pl_abolish(term_t atom, term_t arity);
+word		pl_abolish1(term_t pred);
 word		pl_get_clause_attribute(term_t ref, term_t att, term_t value);
 word		pl_get_predicate_attribute(term_t pred, term_t k, term_t v);
 word		pl_set_predicate_attribute(term_t pred, term_t k, term_t v);
@@ -520,7 +526,7 @@ word		pl_source_file(term_t descr, term_t file, control_t h);
 word		pl_time_source_file(term_t file, term_t t, control_t h);
 word		pl_start_consult(term_t file);
 word		pl_default_predicate(term_t d1, term_t d2);
-Definition	autoImport(FunctorDef f, Module m);
+Definition	autoImport(functor_t f, Module m);
 word		pl_require(term_t pred);
 word		pl_check_definition(term_t spec);
 word		pl_clause_from_source(term_t file, term_t line, term_t clause);
@@ -538,9 +544,11 @@ word		pl_reset_profiler(void);
 void		resetRead(void);
 int		syntaxerrors(int new);
 word		pl_syntaxerrors(term_t old, term_t new);
-int		scan_number(char **, int, Number);
-int		get_number(char *string, char **end, Number value);
+int		get_number(const unsigned char *string,
+			   unsigned char **end,
+			   Number value);
 word		pl_raw_read(term_t term);
+word		pl_raw_read2(term_t stream, term_t term);
 word		pl_read_variables(term_t term, term_t variables);
 word		pl_read_variables3(term_t stream, term_t term,
 				   term_t variables);
@@ -553,8 +561,9 @@ word		pl_read_term3(term_t stream, term_t term, term_t pos);
 
 /* pl-rec.c */
 void		initRecords(void);
-Record		copyTermToHeap(term_t term);
-void		copyTermToGlobal(term_t copy, Record term);
+Record		compileTermToHeap(term_t term);
+void		copyRecordToGlobal(term_t copy, Record term);
+int		structuralEqualArg1OfRecord(term_t t, Record r);
 bool		freeRecord(Record record);
 bool		unifyKey(term_t key, word val);
 word		getKey(term_t key);
@@ -571,11 +580,13 @@ void		install_rl(void);
 void		setupProlog(void);
 handler_t	pl_signal(int sig, handler_t func);
 void		resetSignals(void);
+void		handleSignals(void);
 void		deliverSignal(int sig, int tp, SignalContext scp, char *addr);
 void		deallocateStacks(void);
 bool		restoreStack(Stack s);
 void		trimStacks(void);
 void		resetStacks(void);
+void		emptyStacks(void);
 word		pl_trim_stacks(void);
 word		pl_limit_stack(term_t s, term_t l);
 word		pl_stack_parameter(term_t s, term_t k, term_t o, term_t n);
@@ -642,9 +653,10 @@ word		setInteger(int *val, const char *name, term_t old, term_t new);
 word		setLong(long *val, const char *name, term_t old, term_t new);
 bool		strprefix(char *string, char *prefix);
 bool		strpostfix(char *string, char *postfix);
-bool		stripostfix(char *string, char *postfix);
+bool		stripostfix(const char *string, const char *postfix);
 void		systemMode(bool accept);
-bool		scan_options(term_t list, int flags, OptSpec specs, ...);
+bool		scan_options(term_t list, int flags, atom_t name,
+			     const opt_spec *specs, ...);
 #ifndef HAVE_STRICMP
 int		stricmp(const char *s1, const char *s2);
 #endif
@@ -673,17 +685,17 @@ word		pl_qlf_assert_clause(term_t ref);
 word		pl_qlf_info(term_t file, term_t cvers, term_t fvers, term_t i);
 
 /* pl-write.c */
-char *		varName(term_t var);
+char *		varName(term_t var, char *buf);
 word		pl_nl(void);
 word		pl_nl1(term_t stream);
-word		pl_display(term_t term);
-word		pl_displayq(term_t term);
-word		pl_display2(term_t stream, term_t term);
-word		pl_displayq2(term_t stream, term_t term);
+word		pl_write_canonical(term_t term);
+word		pl_write_canonical2(term_t stream, term_t term);
+word		pl_write_term(term_t term, term_t options);
+word		pl_write_term3(term_t stream,
+			       term_t term, term_t options);
 word		pl_write(term_t term);
 word		pl_writeq(term_t term);
 word		pl_print(term_t term);
-word		pl_dprint(term_t term, term_t g);
 word		pl_write2(term_t stream, term_t term);
 word		pl_writeq2(term_t stream, term_t term);
 word		pl_print2(term_t stream, term_t term);
@@ -696,11 +708,10 @@ word		pl_tty_put(term_t a, term_t affcnt);
 word		pl_set_tty(term_t old, term_t new);
 
 /* pl-main.c */
-int		startProlog(int argc, char **argv, char **env);
+int		startProlog(int argc, char **argv);
 bool		sysError(const char *fm, ...);
 bool		fatalError(const char *fm, ...);
 bool		warning(const char *fm, ...);
-bool		vsysError(const char *fm, va_list args);
 bool		vfatalError(const char *fm, va_list args);
 bool		vwarning(const char *fm, va_list args);
 
