@@ -1,4 +1,4 @@
-/*  $Id: pl-gc.c,v 1.38 1997/03/10 09:36:39 jan Exp $
+/*  $Id: pl-gc.c,v 1.45 1997/11/14 16:25:27 jan Exp $
 
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
@@ -96,6 +96,7 @@ Marking, testing marks and extracting values from GC masked words.
 #define VALUE_MASK	(~GC_MASK)
 
 #if O_SECURE
+char tmp[256];				/* for calling print_val(), etc. */
 #define check_relocation(p) do_check_relocation(p, __FILE__, __LINE__)
 #define recordMark(p)   { if ( (p) < gTop ) *mark_top++ = (p); }
 #else
@@ -126,11 +127,11 @@ Marking, testing marks and extracting values from GC masked words.
 #define val_ptr(w)	val_ptr2((w), storage(w))
 
 #define inShiftedArea(area, shift, ptr) \
-	((char *)ptr >= (char *)stacks.area.base + shift && \
-	 (char *)ptr <  (char *)stacks.area.max + shift )
+	((char *)ptr >= (char *)LD->stacks.area.base + shift && \
+	 (char *)ptr <  (char *)LD->stacks.area.max + shift )
 #define topPointerOnStack(name, addr) \
-	((char *)(addr) >= (char *)stacks.name.base && \
-	 (char *)(addr) <  (char *)stacks.name.max)
+	((char *)(addr) >= (char *)LD->stacks.name.base && \
+	 (char *)(addr) <  (char *)LD->stacks.name.max)
 
 #define onGlobal(p)	onStackArea(global, p) /* onStack()? */
 #define onLocal(p)	onStackArea(local, p)
@@ -153,7 +154,7 @@ forwards void		compact_trail(void);
 forwards void		sweep_mark(mark *);
 forwards void		sweep_trail(void);
 forwards LocalFrame	sweep_environments(LocalFrame);
-forwards void		sweep_choicepoints(LocalFrame);
+forwards LocalFrame	sweep_choicepoints(LocalFrame);
 forwards void		sweep_stacks(LocalFrame);
 forwards void		sweep_local(LocalFrame);
 forwards bool		is_downward_ref(Word);
@@ -202,12 +203,36 @@ needsRelocation(Word addr)
   addHTable(check_table, addr, (Void) TRUE);
 }
 
-char *
-print_val(word val)
-{ static char buf[256];
-  char *tag_name[] = { "var", "int", "float", "atom",
+
+static char *
+print_adr(Word adr, char *buf)
+{ char *name;
+  Word base;
+
+  if ( onGlobal(adr) )
+  { name = "global";
+    base = gBase;
+  } else if ( onLocal(adr) )
+  { name = "local";
+    base = (Word) lBase;
+  } else if ( onTrail(adr) )
+  { name = "trail";
+    base = (Word) tBase;
+  } else
+  { Ssprintf(buf, "%p", adr);
+    return buf;
+  }
+
+  Ssprintf(buf, "%p=%s(%d)", adr, name, adr-base);
+  return buf;
+}
+
+
+static char *
+print_val(word val, char *buf)
+{ char *tag_name[] = { "var", "int", "float", "atom",
 		       "string", "list", "term", "ref" };
-  char *stg_name[] = { "trail", "heap", "global", "local" };
+  char *stg_name[] = { "static/inline/trail", "global", "local", "reserved" };
 
   Ssprintf(buf, "%s at %s(%ld)",
 	   tag_name[tag(val)],
@@ -226,8 +251,10 @@ do_check_relocation(Word addr, char *file, int line)
 { Symbol s;
 
   if ( !(s=lookupHTable(check_table, addr)) )
-  { sysError("%s:%d: Address 0x%lx (%s) was not supposed to be relocated",
-	     file, line, addr, print_val(*addr));
+  { char buf1[256];
+    char buf2[256];
+    sysError("%s:%d: Address %s (%s) was not supposed to be relocated",
+	     file, line, print_adr(addr, buf1), print_val(*addr, buf2));
     return;
   }
 
@@ -276,14 +303,14 @@ static inline word
 makePtr(Word ptr, int tag)
 { int stg;
 
-  if ( onStackArea(trail, ptr) )
-    stg = STG_TRAIL;
+  if ( onStackArea(global, ptr) )
+    stg = STG_GLOBAL;
   else if ( onStackArea(local, ptr) )
     stg = STG_LOCAL;
-  else if ( onStackArea(global, ptr) )
-    stg = STG_GLOBAL;
   else
-    stg = STG_HEAP;
+  { assert(onStackArea(trail, ptr));
+    stg = STG_TRAIL;
+  }
 
   return consPtr(ptr, tag|stg);
 }
@@ -490,6 +517,14 @@ clearUninitialisedVarsFrame(LocalFrame fr, Code PC)
 	case D_BREAK:
 	  c = decode(replacedBreak(PC));
 	  goto again;
+#endif
+#if O_STRING
+	case H_INDIRECT:		/* only skip the size of the */
+	case B_INDIRECT:		/* string + header */
+	{ word m = PC[1];
+	  PC += wsizeofInd(m)+1;
+	  break;
+	}
 #endif
 	case I_EXIT:
 	case I_EXITFACT:
@@ -515,8 +550,6 @@ clearUninitialisedVarsFrame(LocalFrame fr, Code PC)
 	  }
 	  break;
       }
-      if ( c > I_HIGHEST )
-	sysError("GC: Illegal WAM instruction: %d", decode(*PC));
     }
   }
 }
@@ -949,27 +982,37 @@ sweep_environments(LocalFrame fr)
     if ( fr->parent != NULL )
       fr = fr->parent;
     else
-      return parentFrame(fr);	/* Prolog --> C --> Prolog calls */
+      return fr;			/* Prolog --> C --> Prolog calls */
   }
 }
 
 
-static void
+static LocalFrame
 sweep_choicepoints(LocalFrame bfr)
-{ for( ; bfr != (LocalFrame)NULL; bfr = bfr->backtrackFrame )
+{ for( ; ; )
   { sweep_environments(bfr);
     sweep_mark(&bfr->mark);
+    if ( !bfr->backtrackFrame )
+      return bfr;
+    else
+      bfr = bfr->backtrackFrame;
   }
 }
 
+
 static void
 sweep_stacks(LocalFrame fr)
-{ LocalFrame pfr;
+{ LocalFrame tfr, tbfr;
 
-  if ( (pfr = sweep_environments(fr)) != NULL )
-    sweep_stacks(pfr);
+  while(fr)
+  { tfr  = sweep_environments(fr);
+    tbfr = sweep_choicepoints(fr);
 
-  sweep_choicepoints(fr);
+    if ( tfr != tbfr )
+      sweep_mark(&tfr->mark);
+
+    fr = parentFrame(tfr);
+  }
 }
 
 
@@ -1190,7 +1233,7 @@ scan_global(int marked)
       if ( isTerm(*current) )
 	warning("0x%p is term %s/%d",
 		current,
-		stringAtom(functorTerm(*current)->name),
+		stringAtom(nameFunctor(functorTerm(*current))),
 		arityTerm(*current));
       if ( ++errors > 10 )
       { Sdprintf("...\n");
@@ -1263,7 +1306,7 @@ LocalFrame fr;
     slots   = (PC == NULL ? fr->predicate->functor->arity : slotsFrame(fr));
     sp = argFrameP(fr, 0);
     for( n=0; n < slots; n++ )
-    { key += checkData(&sp[n], FALSE);
+    { key += checkData(&sp[n]);
     }
     checked += slots;
     DEBUG(3, Sdprintf(" 0x%lx\n", key));
@@ -1317,6 +1360,25 @@ LocalFrame bfr;
 }
 
 
+word
+check_foreign()
+{ FliFrame ff;
+  word key = 0L;
+
+  for(ff = fli_context; ff; ff = ff->parent )
+  { Word sp = refFliP(ff, 0);
+    int n = ff->size;
+
+    for(n=0 ; n < ff->size; n++ )
+      key += checkData(&sp[n]);
+
+    check_mark(&ff->mark);
+  }
+
+  return key;
+}
+
+
 word
 checkStacks(frame)
 LocalFrame frame;
@@ -1346,11 +1408,7 @@ LocalFrame frame;
 
   assert(local_frames == 0);
 
-  { FliFrame ff;
-
-    for(ff = fli_context; ff; ff = ff->parent )
-      check_mark(&ff->mark);
-  }
+  key += check_foreign();
 
   return key;
 }
@@ -1454,17 +1512,17 @@ pl_garbage_collect(term_t d)
 { LocalFrame fr = environment_frame;
 
 #if O_DEBUG
-  int ol = status.debugLevel;
+  int ol = GD->debug_level;
   int nl;
 
   if ( !PL_get_integer(d, &nl) )
     return warning("garbage_collect/1: instantiation fault");
-  status.debugLevel = nl;
+  GD->debug_level = nl;
 #endif
   finish_foreign_frame();
   garbageCollect(fr);
 #if O_DEBUG
-  status.debugLevel = ol;
+  GD->debug_level = ol;
 #endif
   succeed;
 }
@@ -1660,10 +1718,10 @@ way to reach at dynamic stacks.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define updateStackHeader(name, offset) \
-	{ stacks.name.base  = addPointer(stacks.name.base,  offset); \
-	  stacks.name.top   = addPointer(stacks.name.top,   offset); \
-	  stacks.name.max   = addPointer(stacks.name.max,   offset); \
-	  stacks.name.limit = addPointer(stacks.name.limit, offset); \
+	{ LD->stacks.name.base  = addPointer(LD->stacks.name.base,  offset); \
+	  LD->stacks.name.top   = addPointer(LD->stacks.name.top,   offset); \
+	  LD->stacks.name.max   = addPointer(LD->stacks.name.max,   offset); \
+	  LD->stacks.name.limit = addPointer(LD->stacks.name.limit, offset); \
 	}
 
 
@@ -1780,7 +1838,7 @@ growStacks(LocalFrame fr, Code PC, int l, int g, int t)
     SECURE(key = checkStacks(fr));
 
     if ( t )
-    { tsize = nextStackSize((Stack) &stacks.trail);
+    { tsize = nextStackSize((Stack) &LD->stacks.trail);
       tb = xrealloc(tb, tsize);
       shift_status.trail_shifts++;
     }
@@ -1790,11 +1848,11 @@ growStacks(LocalFrame fr, Code PC, int l, int g, int t)
       assert(lb == addPointer(gb, loffset));	
 
       if ( g )
-      { gsize = nextStackSize((Stack) &stacks.global);
+      { gsize = nextStackSize((Stack) &LD->stacks.global);
 	shift_status.global_shifts++;
       }
       if ( l )
-      { lsize = nextStackSize((Stack) &stacks.local);
+      { lsize = nextStackSize((Stack) &LD->stacks.local);
 	shift_status.local_shifts++;
       }
 
@@ -1816,8 +1874,8 @@ growStacks(LocalFrame fr, Code PC, int l, int g, int t)
 #define PrintStackParms(stack, name, newbase, newsize) \
 	{ Sdprintf("%6s: 0x%08lx ... 0x%08lx --> 0x%08lx ... 0x%08lx\n", \
 		 name, \
-		 (unsigned long) stacks.stack.base, \
-		 (unsigned long) stacks.stack.max, \
+		 (unsigned long) LD->stacks.stack.base, \
+		 (unsigned long) LD->stacks.stack.max, \
 		 (unsigned long) newbase, \
 		 (unsigned long) addPointer(newbase, newsize)); \
 	}
@@ -1832,12 +1890,12 @@ growStacks(LocalFrame fr, Code PC, int l, int g, int t)
     DEBUG(1, Sdprintf("Updating stacks ..."));
     fr = updateStacks(fr, PC, lb, gb, tb);
 
-    stacks.local.max  = addPointer(stacks.local.base,  lsize);
-    stacks.global.max = addPointer(stacks.global.base, gsize);
-    stacks.trail.max  = addPointer(stacks.trail.base,  tsize);
+    LD->stacks.local.max  = addPointer(LD->stacks.local.base,  lsize);
+    LD->stacks.global.max = addPointer(LD->stacks.global.base, gsize);
+    LD->stacks.trail.max  = addPointer(LD->stacks.trail.base,  tsize);
 
-    SetHTop(stacks.local.max);
-    SetHTop(stacks.trail.max);
+    SetHTop(LD->stacks.local.max);
+    SetHTop(LD->stacks.trail.max);
 
     time = CpuTime() - time;
     shift_status.time += time;
