@@ -1,4 +1,4 @@
-/*  $Id: pl-incl.h,v 1.93 1997/03/10 09:36:43 jan Exp $
+/*  $Id: pl-incl.h,v 1.102 1997/10/17 16:35:39 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -15,9 +15,9 @@
 #define PLHOME       "c:/pl"
 #define DEFSTARTUP   ".plrc"
 #define ARCH	     "i386-win32"
-#define C_LIBS	     "-lreadline -lconsole -luxnt"
+#define C_LIBS	     ""
 #define C_STATICLIBS ""
-#define C_CC	     "msvc++ 4.2"
+#define C_CC	     "cl"
 #define C_LDFLAGS    ""
 #else
 #include <parms.h>			/* pick from the working dir */
@@ -81,6 +81,7 @@ handy for it someone wants to add a data type to the system.
 #define O_STRING		1
 #define O_PROLOG_FUNCTIONS	1
 #define O_BLOCK			1
+#define O_CATCHTHROW		1
 #define O_INLINE_FOREIGNS	1
 #define O_PCE			1	/* doesn't depend any longer */
 #define O_DEBUGGER		1
@@ -246,10 +247,6 @@ A common basis for C keywords.
 
 #define forwards static		/* forwards function declarations */
 
-#ifndef GLOBAL			/* global variables */
-#define GLOBAL extern
-#endif
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Booleans,  addresses,  strings  and other   goodies.   Note that  ANSI
 compilers have  `Void'.   This should  be  made part  of  the  general
@@ -274,11 +271,11 @@ typedef void *			Void;
 typedef RETSIGTYPE (*handler_t)(int);
 typedef void *SignalContext;		/* struct sigcontext on sun */
 
-GLOBAL struct
+typedef struct
 { handler_t os;				/* Os signal handler */
   handler_t user;			/* User signal handler */
   bool catched;				/* Prolog catches this one */
-} signalHandlers[MAXSIGNAL];
+} sig_handler;
 #endif /* HAVE_SIGNAL */
 
 #ifndef TRUE
@@ -287,7 +284,7 @@ GLOBAL struct
 #endif
 #define succeed			return TRUE
 #define fail			return FALSE
-#define TRY(goal)		{ if ((goal) == FALSE) fail; }
+#define TRY(goal)		if ((goal) == FALSE) fail
 
 #define CL_START		0	/* asserta */
 #define CL_END			1	/* assertz */
@@ -331,6 +328,7 @@ them.  Descriptions:
 	version (just costs virtual memory).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define BUFFER_RING_SIZE 	4	/* foreign buffer ring (pl-fli.c) */
 #define LINESIZ			1024	/* size of a data line */
 #define MAXARITY		1024	/* arity of predicate */
 #define MAXVARIABLES		65536	/* number of variables/clause */
@@ -338,6 +336,8 @@ them.  Descriptions:
 #define SMALLSTACK		200 * 1024 /* GC policy */
 
 #define WORDBITSIZE		(8 * sizeof(word))
+#define LONGBITSIZE		(8 * sizeof(long))
+#define INTBITSIZE		(8 * sizeof(int))
 
 				/* Prolog's integer range */
 #define PLMINTAGGEDINT		(-(long)(1L<<(WORDBITSIZE - LMASK_BITS - 1)))
@@ -378,7 +378,9 @@ sizes  of  the  hash  tables are defined.  Note that these should all be
 #define ARITHHASHSIZE		16	/* arithmetic function table */
 
 #define stringHashValue(s, n) (unboundStringHashValue(s) & ((n)-1))
-#define pointerHashValue(p, size) ((((long)(p) >> LMASK_BITS)^(long)(p)) & \
+#define pointerHashValue(p, size) ((((long)(p) >> LMASK_BITS) ^ \
+				    ((long)(p) >> (LMASK_BITS+5)) ^ \
+				    ((long)(p))) & \
 				   ((size)-1))
 
 #define TABLE_REF_MASK		0x1L
@@ -396,7 +398,9 @@ sizes  of  the  hash  tables are defined.  Note that these should all be
 	  ForeignRedoPtr(v); \
 	}
 
-#define NEED_REINDEX (1L << (8*sizeof(long)-1)) /* defenition reindexing */
+/* Definition->indexPattern is set to NEED_REINDEX if the definition's index
+   pattern needs to be recomputed */
+#define NEED_REINDEX ((unsigned long)1L << (LONGBITSIZE-1))
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Foreign language interface definitions.  Note that these macros MUST  be
@@ -538,7 +542,11 @@ codes.
 #define I_EXITFACT	((code)75)		/* exit from a fact */
 #define D_BREAK		((code)76)		/* Debugger break-point */
 
-#define I_HIGHEST	((code)76)		/* largest WAM code !!! */
+#if O_CATCHTHROW
+#define B_THROW		((code)77)		/* throw(Exception) */
+#endif
+
+#define I_HIGHEST	((code)77)		/* largest WAM code !!! */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Arithmetic comparison
@@ -702,6 +710,7 @@ REFERENCES
 Common Prolog objects typedefs.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+typedef unsigned long		term_t;		/* external term-reference */
 typedef unsigned long		word;		/* Anonimous 4 byte object */
 typedef word *			Word;		/* a pointer to anything */
 typedef word			atom_t;		/* encoded atom */
@@ -718,8 +727,7 @@ typedef struct definition *	Definition;	/* predicate definition */
 typedef struct clause *		Clause;		/* compiled clause */
 typedef struct clause_ref *	ClauseRef;      /* reference to a clause */
 typedef struct clause_index *	ClauseIndex;    /* Clause indexing table */
-typedef struct clause_chain *	ClauseChain;    /* Chian of clauses in table */
-typedef struct code_info *	CodeInfo;	/* WAM op-code info */
+typedef struct clause_chain *	ClauseChain;    /* Chain of clauses in table */
 typedef struct operator *	Operator;	/* see pl-op.c, pl-read.c */
 typedef struct record *		Record;		/* recorda/3, etc. */
 typedef struct recordList *	RecordList;	/* list of these */
@@ -735,6 +743,17 @@ typedef struct trail_entry *	TrailEntry;	/* Entry of train stack */
 typedef struct data_mark	mark;		/* backtrack mark */
 typedef struct index *		Index;		/* clause indexing */
 typedef struct stack *		Stack;		/* machine stack */
+typedef struct arithFunction * 	ArithFunction;  /* arithmetic function */
+typedef struct assoc *		Assoc;		/* pl-bags.c */
+typedef struct _varDef *	VarDef;		/* pl-comp.c */
+typedef struct extension_cell *	ExtensionCell;  /* pl-ext.c */
+typedef struct abort_handle *	AbortHandle;	/* PL_abort_hook() */
+typedef struct initialise_handle * InitialiseHandle;
+typedef struct tempfile *	TempFile; 	/* pl-os.c */
+typedef struct canonical_dir *	CanonicalDir;	/* pl-os.c */
+typedef struct on_halt *	OnHalt;		/* pl-os.c */
+typedef struct find_data_tag *	FindData; 	/* pl-trace.c */
+typedef struct feature *	Feature; 	/* pl-prims.c */
 
 
 		 /*******************************
@@ -766,6 +785,9 @@ typedef struct
 #define GP_DEFINE	4		/* define a procedure */
 #define GP_RESOLVE	5		/* find defenition */
 
+#define GP_HOW_MASK	0x0ff
+#define GP_NAMEARITY	0x100		/* or'ed mask */
+
 		 /*******************************
 		 *	      FLAGS		*
 		 *******************************/
@@ -818,6 +840,8 @@ with one operation, it turns out to be faster as well.
 
 #define INLINE_F		(0x0001) /* functor */
 
+#define R_DIRTY			(0x0001) /* recordlist */
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Handling environment (or local stack) frames.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -859,8 +883,6 @@ Handling environment (or local stack) frames.
 					     predicateName(def)); \
 				    def->references = 0; \
 				  } \
-				  if ( true(fr, FR_WATCHED) ) \
-				    frameFinished(fr); \
 				}
 #else /*O_DEBUG*/
 #define enterDefinition(def)	{ def->references++; }
@@ -910,8 +932,9 @@ programmer  should  call  startCritical  to start such a code region and
 endCritical to end it.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define startCritical { critical++; }
-#define endCritical   { if (--critical == 0 && aborted == TRUE) pl_abort(); }
+#define startCritical (void)(GD->critical++)
+#define endCritical   if ( --(GD->critical) == 0 && LD->aborted ) \
+			pl_abort()
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 LIST processing macros.
@@ -931,6 +954,7 @@ Structure declarations that must be shared across multiple files.
 
 struct atom
 { Atom		next;		/* next in chain */
+  word		atom;		/* as appearing on the global stack */
 #ifdef O_HASHTERM
   int		hash_value;	/* hash-key value */
 #endif
@@ -976,12 +1000,12 @@ struct clause_ref
 #define CA1_FLOAT	5	/* next 2 are double */
 #define CA1_STRING	6	/* inlined string */
 
-struct code_info
+typedef struct
 { char		*name;		/* name of the code */
   code		code;		/* number of the code */
   char		arguments;	/* # arguments code takes */
   char		argtype;	/* # `external' arguments code takes */
-};
+} code_info;
 
 struct data_mark
 { word		trailtop;	/* top of the trail stack */
@@ -1076,6 +1100,10 @@ struct localFrame
 };  
 
 
+#define QF_NODEBUG		0x0001	/* debug-able query */
+#define QF_DETERMINISTIC	0x0002	/* deterministic success */
+#define	QF_INTERACTIVE		0x0004	/* interactive goal (prolog()) */
+
 struct queryFrame
 { long		magic;			/* Magic code for security */
 #if O_SHIFT_STACKS
@@ -1088,11 +1116,13 @@ struct queryFrame
   long		saved_depth_limit;	/* saved values of these */
   long		saved_depth_reached;
 #endif
-  int		debug;			/* FALSE: nondebugging call */
+#if O_CATCHTHROW
+  term_t	exception;		/* Exception term */
+#endif
+  unsigned long	flags;
   int		debugSave;		/* saved debugstatus.debugging */
   Word	       *aSave;			/* saved argument-stack */
   int		solutions;		/* # of solutions produced */
-  int		deterministic;		/* Last success was deterministic */
   LocalFrame	bfr;			/* BacktrackFrame */
   LocalFrame	saved_environment;	/* Parent local-frame */
   struct localFrame frame;		/* The local frame */
@@ -1105,21 +1135,24 @@ struct fliFrame
   mark		mark;			/* data-stack mark */
 };
 
-
 struct record
-{ RecordList	list;		/* list I belong to */
-  Record	next;		/* next of chain */
-  word		term;		/* term associated */
-  int		n_vars;		/* number of variables */
-  Word		variables;	/* array of variables */
+{ RecordList	list;			/* list I belong to */
+  Record	next;			/* next of chain */
+  int		nvars;			/* # variables in the term */
+  int		gsize;			/* Stack space required (words) */
+  int		size;			/* # bytes of the record */
+  int		erased;			/* record has been erased */
+  char 		buffer[1];		/* array holding codes */
 };
 
 struct recordList
 { RecordList	next;		/* next record chain with same key */
   int		type;		/* RECORD_TYPE */
+  int		references;	/* choicepoints reference count */
   word		key;		/* key of record */
   Record	firstRecord;	/* first record associated with key */
   Record	lastRecord;	/* last record associated with key */
+  unsigned int  flags;		/* R_DIRTY */
 };
 
 struct sourceFile
@@ -1201,7 +1234,6 @@ struct symbol
 		 *	   FLI INTERNALS	*
 		 *******************************/
 
-typedef unsigned long term_t;		/* external term-reference */
 typedef unsigned long qid_t;		/* external query-id */
 typedef unsigned long PL_fid_t;		/* external foreign context-id */
 
@@ -1239,14 +1271,6 @@ typedef unsigned long PL_fid_t;		/* external foreign context-id */
 typedef struct
 { word		name;			/* Name of option */
   int		type;			/* Type of option */
-  union
-  { bool *b;				/* boolean value */
-    long *i;				/* integer value */
-    char **s;				/* string value */
-    word *a;				/* atom value */
-    term_t *t;				/* term-reference */
-    void *ptr;				/* anonymous pointer */
-  } value;
 } opt_spec, *OptSpec;
 
 		 /*******************************
@@ -1316,59 +1340,60 @@ struct stack STACK(caddress);		/* Anonymous stack */
 
 #define N_STACKS (4)
 
-GLOBAL struct
+typedef struct
 { struct STACK(LocalFrame) local;	/* local (environment) stack */
   struct STACK(Word)	   global;	/* local (environment) stack */
   struct STACK(TrailEntry) trail;	/* trail stack */
   struct STACK(Word *)	   argument;	/* argument stack */
-} stacks;
+} pl_stacks_t;
 
-#define tBase	(stacks.trail.base)
-#define tTop	(stacks.trail.top)
-#define tMax	(stacks.trail.max)
+#define tBase	(LD->stacks.trail.base)
+#define tTop	(LD->stacks.trail.top)
+#define tMax	(LD->stacks.trail.max)
 
-#define lBase	(stacks.local.base)
-#define lTop	(stacks.local.top)
-#define lMax	(stacks.local.max)
-#define lLimit	(stacks.local.limit)
+#define lBase	(LD->stacks.local.base)
+#define lTop	(LD->stacks.local.top)
+#define lMax	(LD->stacks.local.max)
+#define lLimit	(LD->stacks.local.limit)
 
-#define gBase	(stacks.global.base)
-#define gTop	(stacks.global.top)
-#define gMax	(stacks.global.max)
-#define gLimit	(stacks.global.limit)
+#define gBase	(LD->stacks.global.base)
+#define gTop	(LD->stacks.global.top)
+#define gMax	(LD->stacks.global.max)
+#define gLimit	(LD->stacks.global.limit)
 
-#define aBase	(stacks.argument.base)
-#define aTop	(stacks.argument.top)
-#define aMax	(stacks.argument.max)
-
-GLOBAL char *	hTop;			/* highest allocated heap address */
-GLOBAL char *	hBase;			/* lowest allocated heap address */
+#define aBase	(LD->stacks.argument.base)
+#define aTop	(LD->stacks.argument.top)
+#define aMax	(LD->stacks.argument.max)
+#define aLimit	(LD->stacks.argument.limit)
 
 #define SetHTop(val)	{ if ( (char *)(val) > hTop  ) hTop  = (char *)(val); }
 #define SetHBase(val)	{ if ( (char *)(val) < hBase ) hBase = (char *)(val); }
 
 #define onStack(name, addr) \
-	((char *)(addr) >= (char *)stacks.name.base && \
-	 (char *)(addr) <  (char *)stacks.name.top)
+	((char *)(addr) >= (char *)LD->stacks.name.base && \
+	 (char *)(addr) <  (char *)LD->stacks.name.top)
 #ifdef O_SHIFT_STACKS
 #define onStackArea(name, addr) \
-	((char *)(addr) >= (char *)stacks.name.base && \
-	 (char *)(addr) <  (char *)stacks.name.max)
+	((char *)(addr) >= (char *)LD->stacks.name.base && \
+	 (char *)(addr) <  (char *)LD->stacks.name.max)
 #else
 #define onStackArea(name, addr) \
-	((char *)(addr) >= (char *)stacks.name.base && \
-	 (char *)(addr) <  (char *)stacks.name.limit)
+	((char *)(addr) >= (char *)LD->stacks.name.base && \
+	 (char *)(addr) <  (char *)LD->stacks.name.limit)
 #endif
-#define usedStack(name) ((char *)stacks.name.top - (char *)stacks.name.base)
-#define sizeStack(name) ((char *)stacks.name.max - (char *)stacks.name.base)
-#define roomStack(name) ((char *)stacks.name.max - (char *)stacks.name.top)
-#define narrowStack(name) (roomStack(name) < stacks.name.minfree)
+#define usedStack(name) ((char *)LD->stacks.name.top - \
+			 (char *)LD->stacks.name.base)
+#define sizeStack(name) ((char *)LD->stacks.name.max - \
+			 (char *)LD->stacks.name.base)
+#define roomStack(name) ((char *)LD->stacks.name.max - \
+			 (char *)LD->stacks.name.top)
+#define narrowStack(name) (roomStack(name) < LD->stacks.name.minfree)
 
 #if O_DYNAMIC_STACKS
 #ifdef NO_SEGV_HANDLING
 #define requireStack(s, n) \
 	{ if ( roomStack(s) < (int)(n) ) \
- 	    ensureRoomStack((Stack)&stacks.s, n); \
+ 	    ensureRoomStack((Stack)&LD->stacks.s, n); \
 	}
 #define verifyStack(s) requireStack(s, 64)
 #else /*NO_SEGV_HANDLING*/
@@ -1378,46 +1403,25 @@ GLOBAL char *	hBase;			/* lowest allocated heap address */
 #else
 #define requireStack(s, n) \
 	{ if ( roomStack(s) < (n) ) \
- 	    outOf((Stack)&stacks.s); \
+ 	    outOf((Stack)&LD->stacks.s); \
 	}
 #define verifyStack(s) requireStack(s, 64)
 #endif
 
 
 		/********************************
-		*       GLOBAL VARIABLES        *
+		*       READ WARNINGS           *
 		*********************************/
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-General global variables  to  indicate  status  or  communicate  between
-modules.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-GLOBAL int 	  critical;		/* in critical code for abort? */
-GLOBAL bool	  aborted;		/* have we been aborted */
-GLOBAL char 	 *cannot_save_program;	/* Program cannot be saved */
-GLOBAL LocalFrame environment_frame;	/* current context frame */
-GLOBAL FliFrame   fli_context;		/* current FLI frame */
-GLOBAL bool	  novice;		/* novice user */
-GLOBAL atom_t	  source_file_name;	/* Current source file_name */
-GLOBAL int	  source_line_no;	/* Current source line_no */
-GLOBAL long	  source_char_no;	/* Current source charno */
-GLOBAL bool	  fileerrors;		/* Report file errors? */
-GLOBAL atom_t	  float_format;		/* Default floating point format */
-
 #define ReadingSource (source_line_no > 0 && \
 		       source_file_name != NULL_ATOM)
 
+
 		/********************************
 		*        FAST DISPATCHING	*
 		********************************/
 
 #if VMCODE_IS_ADDRESS
-GLOBAL char  *dewam_table;			/* decoding table */
-GLOBAL long  dewam_table_offset;		/* offset of 1st */
-GLOBAL code  wam_table[I_HIGHEST+1];		/* code --> address */
-GLOBAL void **interpreter_jmp_table;		/* interpreters table */
-
 #define encode(wam) (wam_table[wam])		/* WAM --> internal */
 						/* internal --> WAM */
 #define decode(c)   ((code) (dewam_table[(unsigned long)(c) - \
@@ -1431,7 +1435,7 @@ GLOBAL void **interpreter_jmp_table;		/* interpreters table */
 		*            STATUS             *
 		*********************************/
 
-GLOBAL struct
+typedef struct
 { bool		requested;		/* GC is requested by stack expander */
   int		blocked;		/* GC is blocked now */
   bool		active;			/* Currently running? */
@@ -1439,75 +1443,50 @@ GLOBAL struct
   long		global_gained;		/* global stack bytes collected */
   long		trail_gained;		/* trail stack bytes collected */
   real		time;			/* time spent in collections */
-} gc_status;
+} pl_gc_status_t;
 
-#ifdef O_SHIFT_STACKS
-GLOBAL struct
+
+typedef struct
 { int		blocked;		/* No shifts allowed */
   real		time;			/* time spent in stack shifts */
   int		local_shifts;		/* Shifts of the local stack */
   int		global_shifts;		/* Shifts of the global stack */
   int		trail_shifts;		/* Shifts of the trail stack */
-} shift_status;
-#endif
+} pl_shift_status_t;
 
-GLOBAL struct
+
+typedef struct
 { atom_t	symbolfile;		/* current symbol file */
   atom_t	orgsymbolfile;		/* symbol file we started with */
   atom_t	restored_state;		/* -r/restore state restored */
-} loaderstatus;
+} pl_loaderstatus_t;
 
 #define NO_PROFILING		0
 #define CUMULATIVE_PROFILING	1
 #define PLAIN_PROFILING		2
 
-/*  statistics information */
-
-GLOBAL struct
-{ long		inferences;		/* logical inferences made */
-  long		heap;			/* heap in use */
-  int		atoms;			/* No. of atoms defined */
-  int		functors;		/* No. of functors defined */
-  int		predicates;		/* No. of predicates defined */
-  int		modules;		/* No. of modules in the system */
-  long		codes;			/* No. of byte codes generated */
-  long		collections;		/* No. of garbage collections */
-  long		global_gained;		/* No. of cells global stack gained */
-  long		trail_gained;		/* No. of cells trail stack gained */
-#ifdef O_PROFILE
-  int		profiling;		/* profiler is on? */
-  long		profile_ticks;		/* profile ticks total */
-#endif /* O_PROFILE */
-} statistics;
 
 		/********************************
 		*            MODULES            *
 		*********************************/
 
-#define MODULE_user	(modules.user)
-#define MODULE_system	(modules.system)
-
-GLOBAL struct
-{ Module	typein;			/* module for type in goals */
-  Module	source;			/* module we are reading clauses in */
-  Module	user;			/* user module */
-  Module	system;			/* system predicate module */
-} modules;
-
-GLOBAL Table	moduleTable;		/* hash table of available modules */
+#define MODULE_user	(GD->modules.user)
+#define MODULE_system	(GD->modules.system)
 
 		/********************************
 		*         PREDICATES            *
 		*********************************/
 
-GLOBAL Procedure	PROCEDURE_alt1;	/* $alt/1, see C_OR */
-GLOBAL Procedure	PROCEDURE_garbage_collect0;
-GLOBAL Procedure	PROCEDURE_block3;
-GLOBAL Procedure	PROCEDURE_true0;
-GLOBAL Procedure	PROCEDURE_fail0;
-GLOBAL Procedure	PROCEDURE_event_hook1;
+#define PROCEDURE_alt1			(GD->procedures.alt1)
+#define PROCEDURE_garbage_collect0	(GD->procedures.garbage_collect0)
+#define PROCEDURE_block3		(GD->procedures.block3)
+#define PROCEDURE_catch3		(GD->procedures.catch3)
+#define PROCEDURE_true0			(GD->procedures.true0)
+#define PROCEDURE_fail0			(GD->procedures.fail0)
+#define PROCEDURE_event_hook1		(GD->procedures.event_hook1)
+#define PROCEDURE_print_message2	(GD->procedures.print_message2)
 
-extern struct code_info	codeTable[];
+extern const code_info codeTable[]; /* Instruction info (read-only) */
 
 		/********************************
 		*            DEBUGGER           *
@@ -1523,16 +1502,17 @@ Tracer communication declarations.
 #define ACTION_IGNORE	3
 #define ACTION_AGAIN	4
 
-#define CALL_PORT	0x01		/* port masks */
-#define EXIT_PORT	0x02
-#define FAIL_PORT	0x04
-#define REDO_PORT	0x08
-#define UNIFY_PORT	0x10
-#define BREAK_PORT	0x20
-#define CUT_CALL_PORT   0x40
-#define CUT_EXIT_PORT   0x80
+#define CALL_PORT	0x001		/* port masks */
+#define EXIT_PORT	0x002
+#define FAIL_PORT	0x004
+#define REDO_PORT	0x008
+#define UNIFY_PORT	0x010
+#define BREAK_PORT	0x020
+#define CUT_CALL_PORT   0x040
+#define CUT_EXIT_PORT   0x080
+#define EXCEPTION_PORT	0x100
 #define CUT_PORT	(CUT_CALL_PORT|CUT_EXIT_PORT)
-#define VERY_DEEP	10000000L	/* deep skiplevel */
+#define VERY_DEEP	1000000000L	/* deep skiplevel */
 
 #define LONGATOM_CHECK	    0x1		/* read/1: error on long atoms */
 #define SINGLETON_CHECK	    0x2		/* read/1: check singleton vars */
@@ -1545,7 +1525,7 @@ Tracer communication declarations.
 #define MAXNEWLINES	    5		/* maximum number of newlines in atom */
 #define SYSTEM_MODE	    (debugstatus.styleCheck & DOLLAR_STYLE)
 
-GLOBAL struct debuginfo
+typedef struct debuginfo
 { unsigned long	skiplevel;		/* current skip level */
   bool		tracing;		/* are we tracing? */
   bool		debugging;		/* are we debugging? */
@@ -1556,26 +1536,31 @@ GLOBAL struct debuginfo
   int		styleCheck;		/* source style checking */
   int		suspendTrace;		/* tracing is suspended now */
   LocalFrame	retryFrame;		/* Frame to retry */
-} debugstatus;
+} pl_debugstatus_t;
 
-#define CHARESCAPE_FEATURE	0x01	/* handle \ in atoms */
-#define GC_FEATURE		0x02	/* do GC */
-#define TRACE_GC_FEATURE	0x04	/* verbose gc */
-#define TTY_CONTROL_FEATURE	0x08	/* allow for tty control */
-#define READLINE_FEATURE	0x10	/* readline is loaded */
-#define DEBUG_ON_ERROR_FEATURE	0x20	/* start tracer on error */
-#define REPORT_ERROR_FEATURE	0x40	/* print error message */
+#define FT_ATOM		0		/* atom feature */
+#define FT_INTEGER	1		/* integer feature */
 
-GLOBAL struct _feature
+#define CHARESCAPE_FEATURE	0x0001	/* handle \ in atoms */
+#define GC_FEATURE		0x0002	/* do GC */
+#define TRACE_GC_FEATURE	0x0004	/* verbose gc */
+#define TTY_CONTROL_FEATURE	0x0008	/* allow for tty control */
+#define READLINE_FEATURE	0x0010	/* readline is loaded */
+#define DEBUG_ON_ERROR_FEATURE	0x0020	/* start tracer on error */
+#define REPORT_ERROR_FEATURE	0x0040	/* print error message */
+#define FILE_CASE_FEATURE	0x0080	/* file names are case sensitive */
+#define FILE_CASE_PRESERVING_FEATURE 0x0100 /* case preserving file names */
+#define DOS_FILE_NAMES_FEATURE  0x0200	/* dos (8+3) file names */
+#define ALLOW_VARNAME_FUNCTOR	0x0400	/* Read Foo(x) as 'Foo'(x) */
+
+typedef struct
 { unsigned long flags;			/* the feature flags */
-} features;
+} pl_features_t;
 
 #define trueFeature(mask)	true(&features, mask)
 
 #ifdef O_LIMIT_DEPTH
 #define DEPTH_NO_LIMIT	(~0x0L)		/* Highest value */
-GLOBAL unsigned long depth_limit;	/* recursion depth limit */
-GLOBAL unsigned long depth_reached;	/* Deapest point reached */
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1590,8 +1575,6 @@ struct state
   State		next;			/* next state loaded */
 };
 
-GLOBAL State stateList;			/* list of loaded states */
-
 #define QLF_TOPLEVEL 0x1		/* toplevel wic file */
 #define QLF_OPTIONS  0x2		/* only load options */
 #define QLF_EXESTATE 0x4		/* probe qlf exe state */
@@ -1615,7 +1598,7 @@ decrease).
 #define REL(a)		((Word)(a) - (Word)(lBase))
 
 #if O_DEBUG
-#define DEBUG(n, g) { if (status.debugLevel >= n) { g; } }
+#define DEBUG(n, g) { if (GD->debug_level >= (n)) { g; } }
 #else
 #define DEBUG(a, b) 
 #endif
@@ -1634,12 +1617,11 @@ decrease).
 
 #include "pl-funcs.h"			/* global functions */
 #include "pl-main.h"			/* Declarations needed by pl-main.c */
-
-extern struct atom atoms[];
-extern struct functorDef functors[];
+#include "pl-error.h"			/* Exception generation */
+#include "pl-global.h"			/* global data */
 
 #define NULL_ATOM ((atom_t)0)
-#define MK_ATOM(n) ((n)<<7|TAG_ATOM|STG_STATIC)
+#define MK_ATOM(n)    		((n)<<7|TAG_ATOM|STG_STATIC)
 #include "pl-atom.ih"
 #include "pl-funct.ih"
 
