@@ -1,4 +1,4 @@
-/*  $Id: pl-index.c,v 1.18 1997/03/10 09:36:47 jan Exp $
+/*  $Id: pl-index.c,v 1.21 1997/08/07 07:58:08 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -33,8 +33,8 @@ indexing only on the first argument as this is default.
 
 /* 1 <= c <= 4 */
 
-#define SHIFT(c, a)	((32/(c)) * a)
-#define MASK(c)		(c == 1 ? ~0L : ((1L << (32/(c))) - 1))
+#define SHIFT(c, a)	((LONGBITSIZE/(c)) * a)
+#define MASK(c)		(c == 1 ? ~0L : ((1L << (LONGBITSIZE/(c))) - 1))
 #define VM(c, a)	((unsigned long)(~(MASK(c) << SHIFT(c, a))))
 
 #define Shift(c, a)	(mask_shift[c][a])
@@ -104,7 +104,7 @@ hashIndex(word key, int buckets)
 }
 
 
-word
+static word
 indexOfWord(word w)
 { for(;;)
   { switch(tag(w))
@@ -384,7 +384,7 @@ unify_index_pattern(Procedure proc, term_t value)
   if ( pattern == 0 )
     fail;
 
-  if ( PL_unify_functor(value, def->functor) )
+  if ( PL_unify_functor(value, def->functor->functor) )
   { term_t a = PL_new_term_ref();
 
     for(n=0; n<arity; n++, pattern >>= 1)
@@ -522,6 +522,8 @@ gcClauseChain(ClauseChain ch, int dirty)
 }
 
 
+#define INFINT (~(1<<(INTBITSIZE-1)))
+
 void
 gcClauseIndex(ClauseIndex ci)
 { ClauseChain ch = ci->entries;
@@ -529,7 +531,7 @@ gcClauseIndex(ClauseIndex ci)
     
   if ( ci->alldirty )
   { for(; n; n--, ch++)
-      ci->size -= gcClauseChain(ch, PLMAXINT);
+      ci->size -= gcClauseChain(ch, INFINT);
   } else
   { for(; n; n--, ch++)
     { if ( ch->dirty )
