@@ -1,4 +1,4 @@
-/*  $Id: pl-itf.h,v 1.83 1997/03/10 09:36:49 jan Exp $
+/*  $Id: pl-itf.h,v 1.93 1997/10/28 13:40:56 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -28,7 +28,7 @@ before loading this file.  See end of this file.
 /* PLVERSION: 10000 * <Major> + 100 * <Minor> + <Patch> */
 
 #ifndef PLVERSION
-#define PLVERSION 20800
+#define PLVERSION 20906
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -65,11 +65,9 @@ _declspec(dllexport) is used by MSVC++ 2.0 to declare exports from DLL's.
 #endif /*_PL_EXPORT_DONE*/
 
 #ifdef _PL_INCLUDE_H
-typedef FunctorDef	functor_t;	/* Name/arity pair */
 typedef Module		module_t;	/* a module */
 typedef Procedure	predicate_t;	/* a predicate handle */
 #else
-typedef void *		functor_t;	/* name/arity pair as Prolog */
 typedef	unsigned long	atom_t;		/* Prolog atom */
 typedef void *		module_t;	/* Prolog module */
 typedef void *		predicate_t;	/* Prolog procedure */
@@ -77,10 +75,15 @@ typedef unsigned long	term_t;		/* opaque term handle */
 typedef unsigned long	qid_t;		/* opaque query handle */
 typedef unsigned long	PL_fid_t;	/* opaque foreign context handle */
 #endif
+typedef unsigned long	functor_t;	/* Name/arity pair */
 typedef unsigned long	atomic_t;	/* same a word */
 typedef unsigned long	control_t;	/* non-deterministic control arg */
 typedef unsigned long	foreign_t;	/* return type of foreign functions */
+#ifdef __cplusplus
+typedef void * 		pl_function_t;	/* can only pass function as void * */
+#else
 typedef foreign_t	(*pl_function_t)(); /* foreign language functions */
+#endif
 
 #define fid_t PL_fid_t			/* avoid AIX name-clash */
 
@@ -173,6 +176,17 @@ __pl_export int		PL_strip_module(term_t in, module_t *m, term_t out);
 		 *	     CALL-BACK		*
 		 *******************************/
 
+#ifdef PL_KERNEL
+#define PL_Q_DEBUG		0x01	/* = TRUE for backward compatibility */
+#endif
+#define PL_Q_NORMAL		0x02	/* normal usage */
+#define PL_Q_NODEBUG		0x04	/* use this one */
+#define PL_Q_CATCH_EXCEPTION	0x08	/* handle exceptions in C */
+#define PL_Q_PASS_EXCEPTION	0x10	/* pass to parent environment */
+#ifdef PL_KERNEL
+#define PL_Q_DETERMINISTIC	0x20	/* call was deterministic */
+#endif
+
 			/* Foreign context frames */
 __pl_export fid_t	PL_open_foreign_frame(void);
 __pl_export void	PL_close_foreign_frame(fid_t cid);
@@ -187,7 +201,7 @@ __pl_export int		PL_predicate_info(predicate_t pred,
 					  module_t *module);
 
 			/* Call-back */
-__pl_export qid_t	PL_open_query(module_t m, int debug,
+__pl_export qid_t	PL_open_query(module_t m, int flags,
 				      predicate_t pred, term_t t0);
 __pl_export int		PL_next_solution(qid_t qid);
 __pl_export void	PL_close_query(qid_t qid);
@@ -197,6 +211,8 @@ __pl_export void	PL_cut_query(qid_t qid);
 __pl_export int		PL_call(term_t t, module_t m);
 __pl_export int		PL_call_predicate(module_t m, int debug,
 					  predicate_t pred, term_t t0);
+__pl_export term_t	PL_exception(qid_t qid);
+__pl_export int		PL_throw(term_t exception);
 
 		 /*******************************
 		 *        TERM-REFERENCES	*
@@ -368,7 +384,7 @@ __pl_export int PL_open_stream(term_t t, IOSTREAM *s);
 		 *	    EMBEDDING		*
 		 *******************************/
 
-__pl_export int		PL_initialise(int argc, char **argv, char **env);
+__pl_export int		PL_initialise(int argc, char **argv);
 install_t		PL_install_readline(void);
 __pl_export int		PL_toplevel(void);
 __pl_export void	PL_halt(int status);
@@ -431,6 +447,8 @@ __pl_export PL_getkey_hook_t   PL_getkey_hook(PL_getkey_hook_t);
 
 __pl_export void (*PL_signal(int sig, void (*func)(int)))(int);
 __pl_export void PL_interrupt(int sig);
+__pl_export void PL_raise(int sig);
+__pl_export void PL_handle_signals(void);
 
 		/********************************
 		*      PROLOG ACTION/QUERY      *
@@ -446,7 +464,7 @@ __pl_export void PL_interrupt(int sig);
 #define PL_ACTION_WRITE		8	/* write via Prolog i/o buffer */
 #define PL_ACTION_FLUSH		9	/* Flush Prolog i/o buffer */
 
-__pl_export int	 PL_action(int, void *); /* perform some action */
+__pl_export int	 PL_action(int, ...);	/* perform some action */
 __pl_export void PL_on_halt(void (*)(int, void *), void *);
 
 		/********************************
