@@ -1,4 +1,4 @@
-/*  $Id: pl-main.c,v 1.53 1997/01/20 16:23:29 jan Exp $
+/*  $Id: pl-main.c,v 1.63 1997/12/03 08:52:01 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -25,17 +25,49 @@ forwards void	usage(void);
 static void	version(void);
 static void	arch(void);
 static void	runtime_vars(void);
+static bool	vsysError(const char *fm, va_list args);
 
 #define	optionString(s) { if (argc > 1) \
 			  { s = argv[1]; argc--; argv++; \
 			  } else \
 			    usage(); \
 			}
+#define K * 1024L
+
+#define EXECVARMAGIC "$EXECVARS="
+static const char exec_vars[512] = EXECVARMAGIC;
+
+static const char *
+exec_var(const char *name)
+{ const char *s=exec_vars + strlen(EXECVARMAGIC);
+  int l = strlen(name);
+
+  while(s < exec_vars+sizeof(exec_vars))
+  { if ( strncmp(name, s, l) == 0 && s[l] == '=' )
+      return &s[l+1];
+    while(*s && s< exec_vars+sizeof(exec_vars))
+      s++;
+    while(*s == '\0' && s< exec_vars+sizeof(exec_vars))
+      s++;
+  }
+
+  return NULL;
+}
+
 
 static char *
-findHome(char *symbols, char *def)
-{ char *home;
+findHome(char *symbols)
+{ char *home = NULL;
   char plp[MAXPATHLEN];
+  const char *val = exec_var("homevar");
+  
+  if ( (val  = exec_var("homevar")) &&
+       (home = getenv(val)) &&
+       (home = PrologPath(home, plp)) )
+    return store_string(home);
+  if ( (val = exec_var("home")) &&
+       (home = PrologPath(home, plp)) )
+    return store_string(home);
 
   if ( !(home = getenv("SWI_HOME_DIR")) )
     home = getenv("SWIPL");
@@ -47,7 +79,7 @@ findHome(char *symbols, char *def)
     char parent[MAXPATHLEN];
     IOSTREAM *fd;
 
-    strcpy(parent, DirName(DirName(AbsoluteFile(home))));
+    strcpy(parent, DirName(DirName(AbsoluteFile(home, buf), buf), buf));
     Ssprintf(buf, "%s/swipl", parent);
 
     if ( (fd = Sopen_file(buf, "r")) )
@@ -69,21 +101,22 @@ findHome(char *symbols, char *def)
 	{ char buf2[MAXPATHLEN];
 
 	  Ssprintf(buf2, "%s/%s", parent, buf);
-	  home = AbsoluteFile(buf2);
+	  home = AbsoluteFile(buf2, plp);
 	} else
-	home = AbsoluteFile(buf);
+	  home = AbsoluteFile(buf, plp);
 
 	if ( ExistsDirectory(home) )
 	{ Sclose(fd);
 	  return store_string(home);
 	}
       }
-	Sclose(fd);
+      Sclose(fd);
     }
   }
 
-  if ( ExistsDirectory(def) )
-    return def;
+  if ( (home = PrologPath(PLHOME, plp)) &&
+       ExistsDirectory(home) )
+    return store_string(home);
 
 #if tos || __DOS__ || __WINDOWS__
 #if tos
@@ -118,10 +151,10 @@ findHome(char *symbols, char *def)
 static char *
 proposeStartupFile(char *symbols)
 { char state[MAXPATHLEN];
+  char buf[MAXPATHLEN];
 
-  if ( !symbols )
-    if ( (symbols = Symbols()) )
-      symbols = DeRefLink(symbols);
+  if ( !symbols && (symbols = Symbols(state)) )
+    symbols = DeRefLink(symbols, buf);
 
   if ( symbols )
   { char *s, *dot = NULL;
@@ -144,7 +177,7 @@ proposeStartupFile(char *symbols)
   if ( systemDefaults.home )
   { Ssprintf(state, "%s/startup/startup.%s",
 	     systemDefaults.home, systemDefaults.arch);
-    return store_string(AbsoluteFile(state));
+    return store_string(AbsoluteFile(state, buf));
   } else
     return store_string("pl.qlf");
 }
@@ -160,16 +193,16 @@ findState(char *symbols)
     return full;
 
   if ( systemDefaults.home )
-  { Ssprintf(state, "%s/startup/startup.%s",
+  { char tmp[MAXPATHLEN];
+
+    Ssprintf(state, "%s/startup/startup.%s",
 	     systemDefaults.home, systemDefaults.arch);
-    full = AbsoluteFile(state);
-    if ( AccessFile(full, ACCESS_READ) )
-      return store_string(full);
+    if ( AccessFile(state, ACCESS_READ) )
+      return store_string(AbsoluteFile(state, tmp));
 
     Ssprintf(state, "%s/startup/startup", systemDefaults.home);
-    full = AbsoluteFile(state);
-    if ( AccessFile(full, ACCESS_READ) )
-      return store_string(full);
+    if ( AccessFile(state, ACCESS_READ) )
+      return store_string(AbsoluteFile(state, tmp));
   }
 
   return NULL;
@@ -190,7 +223,7 @@ warnNoState()
 {
 #ifdef O_RUNTIME
   Sfprintf(Serror, "[FATAL ERROR: Runtime system: can not find a state to run\n");
-  Sfprintf(Serror, "\tUsage: %s -x state\n", mainArgv[0]);
+  Sfprintf(Serror, "\tUsage: %s -x state\n", GD->cmdline.argv[0]);
   Sfprintf(Serror, "\t\twhere <state> is created using qsave_program/[1,2]\n");
   Sfprintf(Serror, "\t\tin the development system]\n");
 #else
@@ -202,20 +235,19 @@ warnNoState()
   if ( full )
     warnNoFile(full);
   if ( systemDefaults.home )
-  { Ssprintf(state, "%s/startup/startup.%s",
+  { char tmp[MAXPATHLEN];
+    Ssprintf(state, "%s/startup/startup.%s",
 	     systemDefaults.home, systemDefaults.arch);
-    full = AbsoluteFile(state);
-    warnNoFile(full);
+    warnNoFile(AbsoluteFile(state, tmp));
 
     Ssprintf(state, "%s/startup/startup", systemDefaults.home);
-    full = AbsoluteFile(state);
-    warnNoFile(full);
+    warnNoFile(AbsoluteFile(state, tmp));
   } else
     Sfprintf(Serror, "    No home directory!\n");
 
   Sfprintf(Serror,
 	  "\nUse\n\t`%s -O -o startup-file -b boot/init.pl'\n",
-	  mainArgv[0]);
+	  GD->cmdline.argv[0]);
   Sfprintf(Serror, "\nto create one]\n");
 #endif
 
@@ -245,17 +277,43 @@ defaultSystemInitFile(char *a0)
 }
 
 
+#define MEMAREA_INVALID_SIZE (unsigned long)(~0L)
+
+static unsigned long
+memarea_limit(const char *s)
+{ number n;
+  unsigned char *q;
+
+  if ( get_number((unsigned char *)s, &q, &n) && intNumber(&n) )
+  { switch((int)*q)
+    { case 'k':
+      case 'K':
+      case EOS:
+	return n.value.i K;
+      case 'm':
+      case 'M':
+	return n.value.i K K;
+      case 'b':
+      case 'B':
+	return n.value.i;
+    }
+  }
+
+  return MEMAREA_INVALID_SIZE;
+}
+
+
 #if O_LINK_PCE
 foreign_t
 pl_pce_init()
-{ prolog_pce_init(mainArgc, mainArgv);
+{ prolog_pce_init(GD->cmdline.argc, GD->cmdline.argv);
 
   succeed;
 }
 #endif
 
 int
-startProlog(int argc, char **argv, char **env)
+startProlog(int argc, char **argv)
 { char *s;
   int n;
   char *state = NULL, *symbols = NULL;
@@ -264,9 +322,8 @@ startProlog(int argc, char **argv, char **env)
   bool explicit_compile_out = FALSE;
   int loadflags = QLF_TOPLEVEL;
 
-  mainArgc			= argc;
-  mainArgv			= argv;
-  mainEnv			= env;
+  GD->cmdline.argc = argc;
+  GD->cmdline.argv = argv;
 
   DEBUG(1, Sdprintf("System compiled at %s %s\n", __TIME__, __DATE__));
 
@@ -274,20 +331,19 @@ startProlog(int argc, char **argv, char **env)
   malloc_debug(O_MALLOC_DEBUG);
 #endif
 
-  status.debugLevel = 0;
+  GD->debug_level = 0;
   DEBUG(1, Sdprintf("OS ...\n"));
   initOs();				/* initialise OS bindings */
 
-  if ( status.dumped == FALSE )
+  if ( GD->dumped == FALSE )
   { char plp[MAXPATHLEN];
 
-    symbols = Symbols();
-    if ( symbols ) 
-      symbols = store_string(DeRefLink(symbols));
+    if ( (symbols = Symbols(plp)) )
+      symbols = store_string(DeRefLink(symbols, plp));
 
     systemDefaults.arch        = ARCH;
-    systemDefaults.home	       = findHome(symbols,
-					  store_string(PrologPath(PLHOME,plp)));
+    systemDefaults.home	       = findHome(symbols);
+
 #ifdef O_XOS
     if ( systemDefaults.home )
     { char buf[MAXPATHLEN];
@@ -301,6 +357,7 @@ startProlog(int argc, char **argv, char **env)
     systemDefaults.global      = DEFGLOBAL;
     systemDefaults.trail       = DEFTRAIL;
     systemDefaults.argument    = DEFARGUMENT;
+    systemDefaults.heap	       = DEFHEAP;
     systemDefaults.goal	       = "'$welcome'";
     systemDefaults.toplevel    = "prolog";
 #ifndef NOTTYCONTROL
@@ -312,12 +369,11 @@ startProlog(int argc, char **argv, char **env)
   }
 
   compile			= FALSE;
-  status.io_initialised		= FALSE;
-  status.initialised		= FALSE;
-  status.notty			= systemDefaults.notty;
-  status.boot			= FALSE;
-  status.extendMode		= TRUE;
-  status.autoload		= TRUE;
+  GD->io_initialised		= FALSE;
+  GD->initialised		= FALSE;
+  GD->cmdline.notty		= systemDefaults.notty;
+  GD->bootsession		= FALSE;
+  LD->autoload			= TRUE;
 
   argc--; argv++;
 
@@ -325,18 +381,19 @@ startProlog(int argc, char **argv, char **env)
 					/* PceEmacs inferior processes */
   if ( ((s = getenv("EMACS")) != NULL && streq(s, "t")) ||
        ((s = getenv("INFERIOR")) != NULL && streq(s, "yes")) )
-    status.notty = TRUE;
+    GD->cmdline.notty = TRUE;
 
   for(n=0; n<argc; n++)			/* need to check this first */
   { DEBUG(2, Sdprintf("argv[%d] = %s\n", n, argv[n]));
     if (streq(argv[n], "-b") )
-      status.boot = TRUE;
+      GD->bootsession = TRUE;
   }
 
-  DEBUG(1, {if (status.boot) Sdprintf("Boot session\n");});
+  DEBUG(1, {if (GD->bootsession) Sdprintf("Boot session\n");});
 
   if ( argc >= 2 && streq(argv[0], "-r") )
-  { loaderstatus.restored_state = lookupAtom(AbsoluteFile(argv[1]));
+  { char tmp[MAXPATHLEN];
+    loaderstatus.restored_state = lookupAtom(AbsoluteFile(argv[1], tmp));
     argc -= 2, argv += 2;		/* recover; we've done this! */
   }
 
@@ -365,11 +422,9 @@ startProlog(int argc, char **argv, char **env)
       runtime_vars();
   }
 
-#define K * 1024L
+  GD->options.systemInitFile = defaultSystemInitFile(GD->cmdline.argv[0]);
 
-  options.systemInitFile = defaultSystemInitFile(mainArgv[0]);
-
-  if ( status.boot == FALSE && status.dumped == FALSE )
+  if ( !GD->bootsession && GD->dumped == FALSE )
   { int state_loaded = FALSE;
 
     if ( !explicit_state )
@@ -387,30 +442,31 @@ startProlog(int argc, char **argv, char **env)
     if ( !state_loaded && loadWicFile(state, loadflags|QLF_OPTIONS) != TRUE )
       Halt(1);
 
-    DEBUG(2, Sdprintf("options.localSize    = %ld\n", options.localSize));
-    DEBUG(2, Sdprintf("options.globalSize   = %ld\n", options.globalSize));
-    DEBUG(2, Sdprintf("options.trailSize    = %ld\n", options.trailSize));
-    DEBUG(2, Sdprintf("options.argumentSize = %ld\n", options.argumentSize));
-    DEBUG(2, Sdprintf("options.goal         = %s\n",  options.goal));
-    DEBUG(2, Sdprintf("options.topLevel     = %s\n",  options.topLevel));
-    DEBUG(2, Sdprintf("options.initFile     = %s\n",  options.initFile));
+    DEBUG(2, Sdprintf("options.localSize    = %ld\n", GD->options.localSize));
+    DEBUG(2, Sdprintf("options.globalSize   = %ld\n", GD->options.globalSize));
+    DEBUG(2, Sdprintf("options.trailSize    = %ld\n", GD->options.trailSize));
+    DEBUG(2, Sdprintf("options.argumentSize = %ld\n", GD->options.argumentSize));
+    DEBUG(2, Sdprintf("options.goal         = %s\n",  GD->options.goal));
+    DEBUG(2, Sdprintf("options.topLevel     = %s\n",  GD->options.topLevel));
+    DEBUG(2, Sdprintf("options.initFile     = %s\n",  GD->options.initFile));
   } else
   { if ( !explicit_state )
       systemDefaults.state = state = findState(symbols);
 
-    options.compileOut	  = "a.out";
-    options.localSize	  = systemDefaults.local    K;
-    options.globalSize	  = systemDefaults.global   K;
-    options.trailSize	  = systemDefaults.trail    K;
-    options.argumentSize  = systemDefaults.argument K;
-    options.goal	  = systemDefaults.goal;
-    options.topLevel	  = systemDefaults.toplevel;
-    options.initFile      = systemDefaults.startup;
+    GD->options.compileOut	  = "a.out";
+    GD->options.localSize	  = systemDefaults.local    K;
+    GD->options.globalSize	  = systemDefaults.global   K;
+    GD->options.trailSize	  = systemDefaults.trail    K;
+    GD->options.argumentSize  = systemDefaults.argument K;
+    GD->options.heapSize	  = systemDefaults.heap	    K;
+    GD->options.goal	  = systemDefaults.goal;
+    GD->options.topLevel	  = systemDefaults.toplevel;
+    GD->options.initFile      = systemDefaults.startup;
   }
 
   for( ; argc > 0 && (argv[0][0] == '-' || argv[0][0] == '+'); argc--, argv++ )
   { if ( streq(&argv[0][1], "tty") )
-    { status.notty = (argv[0][0] == '-');
+    { GD->cmdline.notty = (argv[0][0] == '-');
       continue;
     }
     if ( streq(&argv[0][1], "-" ) )	/* pl <plargs> -- <app-args> */
@@ -420,7 +476,7 @@ startProlog(int argc, char **argv, char **env)
     while(*s)
     { switch(*s)
       { case 'd':	if (argc > 1)
-			{ status.debugLevel = atoi(argv[1]);
+			{ GD->debug_level = atoi(argv[1]);
 			  argc--, argv++;
 			} else
 			  usage();
@@ -429,35 +485,49 @@ startProlog(int argc, char **argv, char **env)
 			  usage();
 			argc--, argv++;
 			break;
-	case 'O':	status.optimise = TRUE;
+	case 'O':	GD->cmdline.optimise = TRUE;
 			break;
-  	case 'o':	optionString(options.compileOut);
+  	case 'o':	optionString(GD->options.compileOut);
 			explicit_compile_out = TRUE;
 			break;
-	case 'f':	optionString(options.initFile);
+	case 'f':	optionString(GD->options.initFile);
 			break;
-	case 'F':	optionString(options.systemInitFile);
+	case 'F':	optionString(GD->options.systemInitFile);
 			break;
-	case 'g':	optionString(options.goal);
+	case 'g':	optionString(GD->options.goal);
 			break;
-	case 't':	optionString(options.topLevel);
+	case 't':	optionString(GD->options.topLevel);
 			break;
 	case 'c':	compile = TRUE;
 			break;
-	case 'b':	status.boot = TRUE;
+	case 'b':	GD->bootsession = TRUE;
 			break;
 	case 'B':
 #if !O_DYNAMIC_STACKS
-			options.localSize    = 32 K;
-			options.globalSize   = 8 K;
-			options.trailSize    = 8 K;
-			options.argumentSize = 1 K;
+			GD->options.localSize    = 32 K;
+			GD->options.globalSize   = 8 K;
+			GD->options.trailSize    = 8 K;
+			GD->options.argumentSize = 1 K;
 #endif
 			goto next;
-	case 'L':	options.localSize    = atoi(++s) K; goto next;
-	case 'G':	options.globalSize   = atoi(++s) K; goto next;
-	case 'T':	options.trailSize    = atoi(++s) K; goto next;
-	case 'A':	options.argumentSize = atoi(++s) K; goto next;
+	case 'L':
+	case 'G':
+	case 'T':
+	case 'A':
+	case 'H':
+        { unsigned long size = memarea_limit(&s[1]);
+	  
+	  if ( size == MEMAREA_INVALID_SIZE )
+	    usage();
+
+	  switch(*s)
+	  { case 'L':	GD->options.localSize    = size; goto next;
+	    case 'G':	GD->options.globalSize   = size; goto next;
+	    case 'T':	GD->options.trailSize    = size; goto next;
+	    case 'A':	GD->options.argumentSize = size; goto next;
+	    case 'H':	GD->options.heapSize     = size; goto next;
+	  }
+	}
       }
       s++;
     }
@@ -468,20 +538,22 @@ startProlog(int argc, char **argv, char **env)
   DEBUG(1, Sdprintf("Command line options parsed\n"));
 
   setupProlog();
+  initialiseForeign(argc, argv);	/* PL_initialise_hook() functions */
+
   systemMode(TRUE);
 
-  if ( status.boot )
+  if ( GD->bootsession )
   { if ( !explicit_compile_out )
-      options.compileOut = proposeStartupFile(NULL);
+      GD->options.compileOut = proposeStartupFile(NULL);
 
-    status.autoload = FALSE;
-    if ( compileFileList(options.compileOut, argc, argv) == TRUE )
+    LD->autoload = FALSE;
+    if ( compileFileList(GD->options.compileOut, argc, argv) == TRUE )
     {
 #if defined(__WINDOWS__) || defined(__WIN32__)
-      PlMessage("Boot compilation has created %s", options.compileOut);
+      PlMessage("Boot compilation has created %s", GD->options.compileOut);
 #else
       if ( !explicit_compile_out )
-	Sfprintf(Serror, "Result stored in %s\n", options.compileOut);
+	Sfprintf(Serror, "Result stored in %s\n", GD->options.compileOut);
 #endif
       Halt(0);
     }
@@ -490,17 +562,19 @@ startProlog(int argc, char **argv, char **env)
   }
 
   if ( state != NULL )
-  { status.boot = TRUE;
+  { GD->bootsession = TRUE;
     if ( loadWicFile(state, loadflags) != TRUE )
       Halt(1);
-    status.boot = FALSE;
+    GD->bootsession = FALSE;
     CSetFeature("boot_file", state);
   }
 
-  initialiseForeign(mainArgc, mainArgv); /* PL_initialise_hook() functions */
+  debugstatus.styleCheck = (LONGATOM_CHECK|
+			    SINGLETON_CHECK|
+			    DISCONTIGUOUS_STYLE);
   systemMode(FALSE);
-  status.dumped = TRUE;
-  status.initialised = TRUE;
+  GD->dumped = TRUE;
+  GD->initialised = TRUE;
 
 #if O_LINK_PCE
   PL_register_foreign("$pce_init", 0, pl_pce_init, PL_FA_TRANSPARENT, 0);
@@ -515,35 +589,36 @@ startProlog(int argc, char **argv, char **env)
   return prolog(lookupAtom("$initialise"));
 }
 
+typedef const char *cline;
 
 static void
 usage()
-{ static char *lines[] = {
+{ static const cline lines[] = {
     "%s: Usage:\n",
-    "    1) %s -help     Display this message\n",
-    "    2) %s -v        Display version information\n",
-    "    3) %s -arch     Display architecture\n",
+    "    1) %s -help      Display this message\n",
+    "    2) %s -v         Display version information\n",
+    "    3) %s -arch      Display architecture\n",
     "    4) %s -dump-runtime-variables\n"
-    "                    Dump link info in sh(1) format\n",
+    "                     Dump link info in sh(1) format\n",
     "    5) %s [options]\n",
     "    6) %s [options] [-o output] -c file ...\n",
     "    7) %s [options] [-o output] -b bootfile -c file ...\n",
     "Options:\n",
-    "    -x state        Start from state (must be first)\n",
-   "    -[LGTA]kbytes   Specify {Local,Global,Trail,Argument} stack limits\n",
-    "    -B              Small stack sizes to prepare for boot\n",
-    "    -t toplevel     Toplevel goal\n",
-    "    -g goal         Initialisation goal\n",
-    "    -f file         Initialisation file\n",
-    "    -F file         System Initialisation file\n",
-    "    [+/-]tty        Allow tty control\n",
-    "    -O              Optimised compilation\n",
+    "    -x state         Start from state (must be first)\n",
+    "    -[LGTAH]size[KM] Specify {Local,Global,Trail,Argument,Heap} limits\n",
+    "    -B               Small stack sizes to prepare for boot\n",
+    "    -t toplevel      Toplevel goal\n",
+    "    -g goal          Initialisation goal\n",
+    "    -f file          Initialisation file\n",
+    "    -F file          System Initialisation file\n",
+    "    [+/-]tty         Allow tty control\n",
+    "    -O               Optimised compilation\n",
     NULL
   };
-  char **lp = lines;
+  const cline *lp = lines;
 
   for(lp = lines; *lp; lp++)
-    Sfprintf(Serror, *lp, BaseName(mainArgv[0]));
+    Sfprintf(Serror, *lp, BaseName(GD->cmdline.argv[0]));
 
   Halt(1);
 }
@@ -629,7 +704,7 @@ warning(const char *fm, ...)
 }
 
 
-bool
+static bool
 vsysError(const char *fm, va_list args)
 { Sfprintf(Serror, "[PROLOG INTERNAL ERROR:\n\t");
   Svfprintf(Serror, fm, args);
@@ -638,7 +713,7 @@ vsysError(const char *fm, va_list args)
 	    "\n[While in %ld-th garbage collection; skipping stacktrace]\n",
 	    gc_status.collections);
   }
-  if ( status.boot || !status.initialised )
+  if ( GD->bootsession || !GD->initialised )
   { Sfprintf(Serror,
 	     "\n[While initialising; quitting]\n");
     Halt(1);
@@ -704,9 +779,8 @@ vwarning(const char *fm, va_list args)
 
   if ( trueFeature(REPORT_ERROR_FEATURE) )
   { if ( ReadingSource &&
-	 !status.boot &&
-	 status.initialised &&
-	 !status.outofstack )		/* cannot call Prolog */
+	 !GD->bootsession && GD->initialised &&
+	 !LD->outofstack )		/* cannot call Prolog */
     { fid_t cid = PL_open_foreign_frame();
       term_t argv = PL_new_term_refs(3);
       predicate_t pred = PL_pred(FUNCTOR_exception3, MODULE_user);
@@ -723,7 +797,7 @@ vwarning(const char *fm, va_list args)
       PL_get_arg(2, argv+1, a); PL_unify_integer(a, source_line_no);
       PL_get_arg(3, argv+1, a); PL_unify_string_chars(a, message);
       
-      qid = PL_open_query(MODULE_user, FALSE, pred, argv);
+      qid = PL_open_query(MODULE_user, PL_Q_NODEBUG, pred, argv);
       rval = PL_next_solution(qid);
       PL_close_query(qid);
       PL_discard_foreign_frame(cid);
