@@ -1,4 +1,4 @@
-/*  $Id: pl-nt.c,v 1.16 1996/11/06 16:35:23 jan Exp $
+/*  $Id: pl-nt.c,v 1.19 1997/10/17 16:35:41 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -141,22 +141,24 @@ CpuTime()
 
 
 char *
-Symbols()
-{ char buf[1024];
-  int n;
+Symbols(char *syms)
+{ int n;
+  char buf[MAXPATHLEN];
 
   if ( (n = GetModuleFileName(NULL, buf, sizeof(buf))) > 0 )
-  { char buf2[1024];
+  { char buf2[MAXPATHLEN];
 
     buf[n] = EOS;
     _xos_long_file_name(buf, buf2);
 
-    return store_string(buf2);
+    strcpy(syms, buf2);
   } else
-  { PrologPath(mainArgv[0], buf);
+  { PrologPath(GD->cmdline.argv[0], buf);
 
-    return Which(buf);
+    strcpy(syms, buf);
   }
+
+  return syms;
 }
 
 		 /*******************************
