@@ -1,4 +1,4 @@
-/*  $Id: pl-ntcon.c,v 1.1 1996/04/04 09:58:39 jan Exp $
+/*  $Id: pl-ntcon.c,v 1.2 1997/07/02 15:22:46 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -11,8 +11,8 @@
 #include "pl-itf.h"
 
 int
-main(int argc, char **argv, char **env)
-{ if ( !PL_initialise(argc, argv, env) )
+main(int argc, char **argv)
+{ if ( !PL_initialise(argc, argv) )
     PL_halt(1);
   
   PL_halt(PL_toplevel() ? 0 : 1);
