@@ -1,4 +1,4 @@
-/*  $Id: pl-prims.c,v 1.52 1997/03/10 09:36:57 jan Exp $
+/*  $Id: pl-prims.c,v 1.63 1997/10/28 13:41:04 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -10,6 +10,8 @@
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "pl-ctype.h"
+#undef ulong
+#define ulong unsigned long
 
 forwards char 	*prependBase(int, char *);
 
@@ -71,7 +73,7 @@ _pl_ground(Word p)
     fail;
   if (!isTerm(*p) )
     succeed;
-  arity = functorTerm(*p)->arity;
+  arity = arityFunctor(functorTerm(*p));
   for(p = argTermP(*p, 0); arity > 0; arity--, p++)
     TRY( _pl_ground(p) );
 
@@ -120,11 +122,11 @@ termHashValue(word term, long *hval)
 	succeed;
       }
       case TAG_COMPOUND:
-      { FunctorDef fd = functorTerm(term);
-	int arity = fd->arity;
+      { functor_t fd = functorTerm(term);
+	int arity = arityFunctor(fd);
 	Word a, a2;
 
-	*hval = atomValue(fd->name)->hash_value + arity;
+	*hval = atomValue(nameFunctor(fd))->hash_value + arity;
 	for(a = argTermP(term, 0); arity; arity--, a++)
 	{ long av;
 
@@ -217,7 +219,7 @@ _pl_equal(register Word t1, register Word t2)
        functorTerm(*t1) != functorTerm(*t2) )
     fail;
 
-  arity = functorTerm(*t1)->arity;
+  arity = arityFunctor(functorTerm(*t1));
   t1 = argTermP(*t1, 0);
   t2 = argTermP(*t2, 0);
   for(n=0; n<arity; n++, t1++, t2++)
@@ -347,8 +349,6 @@ tail_recursion:
 	}
         goto tail_recursion;
       }
-
-      return EQUAL;
     }
     default:
       assert(0);
@@ -437,7 +437,7 @@ struct uchoice
 };
 
 static bool
-structeql(Word t1, Word t2, Buffer buf)
+structeql(Word t1, Word t2, TmpBuffer buf)
 { int todo = 1;
   UChoice nextch = NULL, tailch = NULL;
 
@@ -492,13 +492,13 @@ structeql(Word t1, Word t2, Buffer buf)
 	  continue;
         fail;
       case TAG_COMPOUND:
-      { FunctorDef fd = functorTerm(w1);
+      { functor_t fd = functorTerm(w1);
 	int arity;
 
 	if ( !hasFunctor(w2, fd) )
 	  fail;
 
-	arity = fd->arity;
+	arity = arityFunctor(fd);
 	p1 = argTermP(w1, 0);
 	p2 = argTermP(w2, 0);
 	if ( todo == 0 )		/* right-most argument recursion */
@@ -528,7 +528,7 @@ structeql(Word t1, Word t2, Buffer buf)
 word
 pl_structural_equal(term_t t1, term_t t2)
 { bool rval;
-  buffer buf;
+  tmp_buffer buf;
   Reset r;
   Word p1 = valTermRef(t1);
   Word p2 = valTermRef(t2);
@@ -604,7 +604,11 @@ pl_arg(term_t n, term_t term, term_t arg, word b)
     { int idx;
 
       if ( !PL_get_name_arity(term, &name, &arity) )
-	return warning("arg/3: second argument in not a term");
+      { if ( PL_is_variable(term) )
+	  return PL_error("arg", 3, NULL, ERR_INSTANTIATION);
+	else
+	  return PL_error("arg", 3, NULL, ERR_TYPE, ATOM_compound, term);
+      }
   
       if ( PL_get_integer(n, &idx) )
       { if ( idx >= 0 && idx <= arity )
@@ -613,7 +617,9 @@ pl_arg(term_t n, term_t term, term_t arg, word b)
 	  PL_get_arg(idx, term, a);
 	  return PL_unify(arg, a);
 	}
-
+	if ( idx < 0 )
+	  return PL_error("arg", 3, NULL, ERR_DOMAIN,
+			  ATOM_not_less_than_zero, n);
 	fail;
       } 
       if ( PL_is_variable(n) )
@@ -631,14 +637,13 @@ pl_arg(term_t n, term_t term, term_t arg, word b)
 	}
 	fail;
       }
-      return warning("arg/3: first argument in not an integer or unbound");
+      return PL_error("arg", 3, NULL, ERR_TYPE, ATOM_integer, n);
     }
     case FRG_REDO:
     { int argn = ForeignContextInt(b) + 1;
       term_t a = PL_new_term_ref();
 
-      if ( !PL_get_name_arity(term, &name, &arity) )
-	return warning("arg/3: second argument in not a term");
+      PL_get_name_arity(term, &name, &arity);
 
       for(; argn <= arity; argn++)
       { PL_get_arg(argn, term, a);
@@ -784,7 +789,7 @@ pl_univ(term_t t, term_t list)
 
 
 static int
-do_number_vars(term_t t, FunctorDef functor, int n)
+do_number_vars(term_t t, functor_t functor, int n)
 { atom_t name;
   int arity;
 
@@ -797,7 +802,7 @@ start:
     PL_unify_arg(1, t, tmp);
 
     n++;
-  } else if ( _PL_get_name_arity(t, &name, &arity) && arity > 0 )
+  } else if ( _PL_get_name_arity(t, &name, &arity) )
   { if ( arity == 1 )
     { PL_get_arg(1, t, t);
       goto start;
@@ -823,7 +828,7 @@ start:
 
 
 int
-numberVars(term_t t, FunctorDef functor, int n)
+numberVars(term_t t, functor_t functor, int n)
 { term_t h2 = PL_copy_term_ref(t);
   int rval = do_number_vars(h2, functor, n);
 
@@ -837,7 +842,7 @@ word
 pl_numbervars(term_t t, term_t f,
 	      term_t start, term_t end)
 { int n;
-  FunctorDef functor;
+  functor_t functor;
   atom_t name;
   
   if ( !PL_get_integer(start, &n) ||
@@ -869,7 +874,7 @@ free_variables(Word t, term_t l, int n)
     return n+1;
   }
   if ( isTerm(*t) )
-  { int arity = functorTerm(*t)->arity;
+  { int arity = arityFunctor(functorTerm(*t));
 
     for(t = argTermP(*t, 0); arity > 0; arity--, t++)
       n = free_variables(t, l, n);
@@ -919,7 +924,7 @@ dobind_vars(Word t, atom_t constant)
     return;
   }
   if ( isTerm(*t) )
-  { int arity = functorTerm(*t)->arity;
+  { int arity = arityFunctor(functorTerm(*t));
 
     for(t = argTermP(*t, 0); arity > 0; arity--, t++)
       dobind_vars(t, constant);
@@ -936,7 +941,7 @@ bind_existential_vars(Word t)
     int arity;
     Word a;
 
-    if ( f->definition == FUNCTOR_hat2->functor )
+    if ( f->definition == FUNCTOR_hat2 )
     { dobind_vars(&f->arguments[0], ATOM_nil);
       return bind_existential_vars(&f->arguments[1]);
     }
@@ -1025,7 +1030,7 @@ pre_copy_analysis(Word t, int *index)
   }
 
   if ( isTerm(*t) )
-  { int arity = functorTerm(*t)->arity;
+  { int arity = arityFunctor(functorTerm(*t));
     int subvars = 0;
     int thisindex = (*index)++;
     term_t thisterm = PL_new_term_refs(0);
@@ -1132,8 +1137,8 @@ do_copy(term_t from, term_t to, CopyInfo info)
     { info->index++;
       PL_unify(to, from);
     } else
-    { FunctorDef fd = functorTerm(*p);
-      int n, arity = fd->arity;
+    { functor_t fd = functorTerm(*p);
+      int n, arity = arityFunctor(fd);
       term_t af = PL_new_term_ref();
       term_t at = PL_new_term_ref();
 
@@ -1281,9 +1286,10 @@ pl_int_to_atom(term_t number, term_t base, term_t atom)
  ** Fri Aug 19 22:26:41 1988  jan@swivax.UUCP (Jan Wielemaker)  */
 
 char *
-formatInteger(bool split, int div, int radix, bool small, long int n)
-{ static char tmp[100];
-  char *s = tmp + 99;
+formatInteger(bool split, int div, int radix, bool small, long int n,
+	      char *out)
+{ char tmp[100];
+  char *s = &tmp[sizeof(tmp)-1];	/* i.e. start at the end */
   int before = (div == 0);
   int digits = 0;
   bool negative = FALSE;
@@ -1294,8 +1300,9 @@ formatInteger(bool split, int div, int radix, bool small, long int n)
     negative = TRUE;
   }
   if ( n == 0 && div == 0 )
-  { *--s = '0';
-    return s;
+  { out[0] = '0';
+    out[1] = EOS;
+    return out;
   }
   while( n > 0 || div >= 0 )
   { if ( div-- == 0 && !before )
@@ -1310,7 +1317,7 @@ formatInteger(bool split, int div, int radix, bool small, long int n)
   if ( negative )
     *--s = '-';  
 
-  return s;
+  return strcpy(out, s);
 }	  
 
 
@@ -1333,14 +1340,14 @@ pl_format_number(term_t format, term_t number, term_t string)
     case 'r':
     case 'R':
       { long i;
-	char *result;
+	char result[50];
 
 	if ( !PL_get_long(number, &i) )
 	  return warning("format_number/3: 2nd argument is not an integer");
 	if (conv == 'd' || conv == 'D')
-	  result = formatInteger(conv == 'D', arg, 10, TRUE, i);
+	  formatInteger(conv == 'D', arg, 10, TRUE, i, result);
 	else
-	  result = formatInteger(FALSE, 0, arg, conv == 'r', i);
+	  formatInteger(FALSE, 0, arg, conv == 'r', i, result);
 
 	return PL_unify_list_chars(string, result);
       }
@@ -1390,9 +1397,9 @@ x_chars(const char *pred, term_t atom, term_t string, int how)
       case X_NUMBER:
       default:
       { number n;
-	char *q;
+	unsigned char *q;
 
-	if ( get_number(s, &q, &n) && *q == EOS )
+	if ( get_number((unsigned char *)s, &q, &n) && *q == EOS )
 	{ if ( intNumber(&n) )
 	    return PL_unify_integer(atom, n.value.i);
 	  else
@@ -1527,14 +1534,14 @@ pl_concat(term_t a1, term_t a2, term_t a3)
 
 word
 pl_concat_atom3(term_t list, term_t sep, term_t atom)
-{ buffer b;
-  char *s;
+{ char *s;
   term_t l = PL_copy_term_ref(list);
   term_t head = PL_new_term_ref();
   int first = TRUE;
   char *sp;
   int splen;
-
+  tmp_buffer b;
+  
   if ( sep )
   { if ( !PL_get_chars(sep, &sp, CVT_ATOMIC) )
       return warning("concat_atom/3: illegal separator");
@@ -1667,6 +1674,8 @@ pl_string_to_list(term_t str, term_t list)
 
   if ( PL_get_chars(str, &s, CVT_ALL) )
     return PL_unify_list_chars(list, s);
+  if ( PL_get_list_chars(list, &s, 0) )	/* string_to_list(S, []). */
+    return PL_unify_string_chars(str, s);
   if ( PL_get_chars(list, &s, CVT_ALL) )
     return PL_unify_string_chars(str, s);
 
@@ -1710,7 +1719,7 @@ write_on(term_t goal, int how, term_t target)
   char *string = buf;
   bool rval;
 
-  tellString(&string, 10240);
+  tellString(&string, 1024);
   rval = callProlog(MODULE_user, goal, FALSE);
   toldString();
   TRY(rval);
@@ -1975,7 +1984,7 @@ pl_depth_limit_false(term_t limit, term_t olimit, term_t oreached, term_t res)
  ** Sun Apr 17 15:38:46 1988  jan@swivax.UUCP (Jan Wielemaker)  */
 
 #define makeNum(n)	((n) < PLMAXTAGGEDINT ? consInt(n) : globalLong(n))
-#define limitStack(n)	diffPointers(stacks.n.limit, stacks.n.base)
+#define limitStack(n)	diffPointers(LD->stacks.n.limit, LD->stacks.n.base)
 
 word
 pl_statistics(term_t k, term_t value)
@@ -1988,15 +1997,31 @@ pl_statistics(term_t k, term_t value)
   if      (key == ATOM_cputime)				/* time */
     result = globalReal(CpuTime());
   else if (key == ATOM_inferences)			/* inferences */
-    result = makeNum(statistics.inferences);
+    result = makeNum(LD->statistics.inferences);
   else if (key == ATOM_local)				/* local stack */
     result = makeNum((long)lMax - (long)lBase);
   else if (key == ATOM_localused)
     result = makeNum((long)lTop - (long)lBase);
   else if (key == ATOM_locallimit)
     result = makeNum(limitStack(local));
-  else if (key == ATOM_heapused)			/* heap */
-    result = makeNum(statistics.heap);
+  else if (key == ATOM_heaplimit)			/* heap */
+#ifdef MMAP_STACK
+  { ulong heap = (ulong)gBase - heap_base;
+    result = makeNum(heap);
+  }
+#else
+    fail;
+#endif
+  else if (key == ATOM_heap)
+#ifdef MMAP_STACK
+  { ulong heap = (ulong)sbrk(0) - heap_base;
+    result = makeNum(heap);
+  }
+#else
+    fail;
+#endif
+  else if (key == ATOM_heapused)			/* heap usage */
+    result = makeNum(GD->statistics.heap);
   else if (key == ATOM_trail)				/* trail */
     result = makeNum((long)tMax - (long)tBase);
   else if (key == ATOM_trailused)
@@ -2010,15 +2035,15 @@ pl_statistics(term_t k, term_t value)
   else if (key == ATOM_globallimit)
     result = makeNum(limitStack(global));
   else if (key == ATOM_atoms)				/* atoms */
-    result = consInt(statistics.atoms);
+    result = consInt(GD->statistics.atoms);
   else if (key == ATOM_functors)			/* functors */
-    result = consInt(statistics.functors);
+    result = consInt(GD->statistics.functors);
   else if (key == ATOM_predicates)			/* predicates */
-    result = consInt(statistics.predicates);
+    result = consInt(GD->statistics.predicates);
   else if (key == ATOM_modules)				/* modules */
-    result = consInt(statistics.modules);
+    result = consInt(GD->statistics.modules);
   else if (key == ATOM_codes)				/* codes */
-    result = consInt(statistics.codes);
+    result = consInt(GD->statistics.codes);
   else if (key == ATOM_gctime)
     result = globalReal(gc_status.time);
   else if (key == ATOM_collections)
@@ -2054,14 +2079,18 @@ pl_statistics(term_t k, term_t value)
 		 *	 FEATURE HANDLING	*
 		 *******************************/
 
-typedef struct feature *Feature;
 struct feature
-{ atom_t name;
-  word value;
+{ atom_t	name;
+  int	 	type;
+  union
+  { atom_t	atom;			/* value as atom */
+    long	i;			/* value as integer */
+  } value;
+
   Feature next;
 };
 
-static Feature feature_list = NULL;
+#define feature_list (GD->_feature_list)
 
 typedef struct
 { atom_t	name;
@@ -2073,32 +2102,49 @@ typedef struct
   atom_t       *address;
 } builtin_named_feature;
 
-builtin_boolean_feature builtin_boolean_features[] = 
-{ { ATOM_character_escapes,	CHARESCAPE_FEATURE },
-  { ATOM_gc,			GC_FEATURE },
-  { ATOM_trace_gc,		TRACE_GC_FEATURE },
-  { ATOM_tty_control,		TTY_CONTROL_FEATURE },
-  { ATOM_readline,		READLINE_FEATURE },
-  { ATOM_debug_on_error,	DEBUG_ON_ERROR_FEATURE },
-  { ATOM_report_error,		REPORT_ERROR_FEATURE },
-  { NULL_ATOM,			0L }
+static const builtin_boolean_feature builtin_boolean_features[] = 
+{ { ATOM_character_escapes,	         CHARESCAPE_FEATURE },
+  { ATOM_gc,			         GC_FEATURE },
+  { ATOM_trace_gc,		         TRACE_GC_FEATURE },
+  { ATOM_tty_control,		         TTY_CONTROL_FEATURE },
+  { ATOM_readline,		         READLINE_FEATURE },
+  { ATOM_debug_on_error,	         DEBUG_ON_ERROR_FEATURE },
+  { ATOM_report_error,		         REPORT_ERROR_FEATURE },
+  { ATOM_case_sensitive_file_names,      FILE_CASE_FEATURE },
+  { ATOM_allow_variable_name_as_functor, ALLOW_VARNAME_FUNCTOR },
+  { NULL_ATOM,			         0L }
 };
 
-builtin_named_feature builtin_named_features[] =
-{ { ATOM_float_format,		&float_format },
-  { NULL_ATOM,			NULL }
+static const builtin_named_feature builtin_named_features[] =
+{ { ATOM_float_format,		    &float_format },
+  { NULL_ATOM,			    NULL }
 };
 
 int
-setFeature(atom_t name, word value)
+setFeature(atom_t name, int type, ...)
 { Feature f;
-  builtin_boolean_feature *bf;
-  builtin_named_feature   *nf;
+  const builtin_boolean_feature *bf;
+  const builtin_named_feature   *nf;
+  atom_t a = 0;
+  long i = 0;
+  va_list args;
+
+  va_start(args, type);
+  switch(type)
+  { case FT_ATOM:
+      a = va_arg(args, atom_t);
+      break;
+    case FT_INTEGER:
+      i = va_arg(args, long);
+      break;
+    default:
+      assert(0);
+  }
 
   for(nf = builtin_named_features; nf->name; nf++)
   { if ( name == nf->name )
-    { if ( isAtom(value) )
-      { *nf->address = value;
+    { if ( type == FT_ATOM )
+      { *nf->address = a;
       } else
       { warning("set_feature/2: %s feature is atom", stringAtom(name));
 	fail;
@@ -2108,16 +2154,20 @@ setFeature(atom_t name, word value)
   }
   for(bf = builtin_boolean_features; bf->name; bf++ )
   { if ( name == bf->name )
-    { if ( value == ATOM_true || value == ATOM_on )
-	set(&features, bf->mask);
-      else if ( value == ATOM_false || value == ATOM_off )
-	clear(&features, bf->mask);
-      else
-      { warning("set_feature/2: %s feature is boolean", stringAtom(name));
+    { if ( type == FT_ATOM )
+      { if ( a == ATOM_true || a == ATOM_on )
+	  set(&features, bf->mask);
+	else if ( a == ATOM_false || a == ATOM_off )
+	  clear(&features, bf->mask);
+	else
+	  goto nobool;
+      } else
+      { nobool:
+	warning("set_feature/2: %s feature is boolean", stringAtom(name));
 	fail;
       }
-      if ( name == ATOM_tty_control )	/* status.notty should be feature */
-	status.notty = (trueFeature(TTY_CONTROL_FEATURE) ? FALSE : TRUE);
+      if ( name == ATOM_tty_control )	/* GD->cmdline.notty should be feature */
+	GD->cmdline.notty = (trueFeature(TTY_CONTROL_FEATURE) ? FALSE : TRUE);
       goto doset;
     }
   }
@@ -2125,28 +2175,36 @@ setFeature(atom_t name, word value)
 doset:
   for(f=feature_list; f; f = f->next)
   { if ( f->name == name )
-    { f->value = value;
-      succeed;
-    }
+      goto setf;
   }
 
   f = allocHeap(sizeof(struct feature));
   f->next = feature_list;
   f->name = name;
-  f->value = value;
   feature_list = f;
 
+setf:
+  f->type = type;
+  switch(f->type)
+  { case FT_ATOM:
+      f->value.atom = a;
+      break;
+    case FT_INTEGER:
+      f->value.i = i;
+      break;
+  }
+
   succeed;
 }
 
 
-word
-getFeature(atom_t name)
+static Feature
+findFeature(atom_t name)
 { Feature f;
 
   for(f=feature_list; f; f = f->next)
   { if ( f->name == name )
-      return f->value;
+      return f;
   }
 
   fail;
@@ -2156,14 +2214,32 @@ getFeature(atom_t name)
 word
 pl_set_feature(term_t key, term_t value)
 { atom_t k;
-  word v;
+  atom_t a;
+  long i;
   
-  if ( !PL_get_atom(key, &k) ||
-       !(PL_is_atom(value) || PL_is_integer(value)) )
-    return warning("set_feature/2; instantiation fault");
+  if ( !PL_get_atom(key, &k) )
+    return warning("set_feature/2: key is not an atom");
 
-  v = _PL_get_atomic(value);
-  return setFeature(k, v);
+  if ( PL_get_atom(value, &a) )
+    return setFeature(k, FT_ATOM, a);
+  if ( PL_get_long(value, &i) )
+    return setFeature(k, FT_INTEGER, i);
+
+  return warning("set_feature/2: illegal value");
+}
+
+
+static int
+unify_feature_value(Feature f, term_t val)
+{ switch(f->type)
+  { case FT_ATOM:
+      return PL_unify_atom(val, f->value.atom);
+    case FT_INTEGER:
+      return PL_unify_integer(val, f->value.i);
+    default:
+      assert(0);
+      fail;
+  }
 }
 
 
@@ -2176,10 +2252,10 @@ pl_feature(term_t key, term_t value, word h)
     { atom_t k;
 
       if ( PL_get_atom(key, &k) )
-      { word val;
+      { Feature f;
 
-	if ( (val=getFeature(k)) )
-	  return _PL_unify_atomic(value, val);
+	if ( (f = findFeature(k)) )
+	  return unify_feature_value(f, value);
 	fail;
       } else if ( PL_is_variable(key) )
       { here = feature_list;
@@ -2196,7 +2272,7 @@ pl_feature(term_t key, term_t value, word h)
 
   for(; here; here = here->next)
   { if ( PL_unify_atom(key, here->name) &&
-	 _PL_unify_atomic(value, here->value) )
+	 unify_feature_value(here, value) )
     { if ( !here->next )
 	succeed;
       ForeignRedoPtr(here->next);
@@ -2223,20 +2299,20 @@ pl_option(term_t key, term_t old, term_t new)
   if ( !PL_get_atom(key, &k) )
     fail;
 
-  if (     k == ATOM_goal)	       result = options.goal;
-  else if (k == ATOM_top_level)	       result = options.topLevel;
-  else if (k == ATOM_init_file)        result = options.initFile;
-  else if (k == ATOM_system_init_file) result = options.systemInitFile;
+  if (     k == ATOM_goal)	       result = GD->options.goal;
+  else if (k == ATOM_top_level)	       result = GD->options.topLevel;
+  else if (k == ATOM_init_file)        result = GD->options.initFile;
+  else if (k == ATOM_system_init_file) result = GD->options.systemInitFile;
   else fail;
 
   if ( !PL_unify_atom_chars(old, result) ||
        !PL_get_atom_chars(new, &n) )
     fail;
 
-  if (     k == ATOM_goal)		   options.goal     = n;
-  else if (k == ATOM_top_level)		   options.topLevel = n;
-  else if (k == ATOM_init_file) 	   options.initFile = n;
-  else /*if (k == ATOM_system_init_file)*/ options.systemInitFile = n;
+  if (     k == ATOM_goal)		   GD->options.goal     = n;
+  else if (k == ATOM_top_level)		   GD->options.topLevel = n;
+  else if (k == ATOM_init_file) 	   GD->options.initFile = n;
+  else /*if (k == ATOM_system_init_file)*/ GD->options.systemInitFile = n;
 
   succeed;
 }
@@ -2250,7 +2326,7 @@ pl_please(term_t key, term_t old, term_t new)
     fail;
 
   if   ( k == ATOM_optimise )
-    return setBoolean(&status.optimise, "please", old, new);
+    return setBoolean(&GD->cmdline.optimise, "please", old, new);
   else
     return warning("please/3: unknown key: %s", stringAtom(k));
 }
@@ -2272,13 +2348,4 @@ pl_style_check(term_t old, term_t new)
   }
 
   fail;
-}
-
-		/********************************
-		*        USER MODELLING?        *
-		*********************************/
-
-word
-pl_novice(term_t old, term_t new)
-{ return setBoolean(&novice, "$novice", old, new);
 }
