@@ -1,4 +1,4 @@
-/*  $Id: pl-proc.c,v 1.49 1997/03/10 09:37:00 jan Exp $
+/*  $Id: pl-proc.c,v 1.54 1997/12/03 08:52:03 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -24,22 +24,22 @@ SourceFile 	tailSourceFileTable = (SourceFile) NULL;
 static void	removeClausesProcedure(Procedure proc, int sfindex);
 
 Procedure
-lookupProcedure(FunctorDef f, Module m)
+lookupProcedure(functor_t f, Module m)
 { Procedure proc;
   register Definition def;
   Symbol s;
   
-  if ((s = lookupHTable(m->procedures, f)) != (Symbol) NULL)
+  if ( (s = lookupHTable(m->procedures, (void *)f)) )
     return (Procedure) s->value;
 
   proc = (Procedure)  allocHeap(sizeof(struct procedure));
   def  = (Definition) allocHeap(sizeof(struct definition));
   proc->type = PROCEDURE_TYPE;
   proc->definition = def;
-  def->functor = f;
+  def->functor = valueFunctor(f);
   def->module  = m;
-  addHTable(m->procedures, f, proc);
-  statistics.predicates++;
+  addHTable(m->procedures, (void *)f, proc);
+  GD->statistics.predicates++;
 
   def->definition.clauses = NULL;
   def->lastClause = NULL;
@@ -57,11 +57,19 @@ lookupProcedure(FunctorDef f, Module m)
   return proc;
 }
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+resetProcedure() is called  by  lookupProcedure()   for  new  ones,  and
+abolishProcedure() by abolish/2. In the latter   case, abolish may leave
+dirty clauses when called on a   running predicate. Hence, NEEDSCLAUSEGC
+should be retained. Bug found by Paulo Moura, LogTalk developer.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static void
 resetProcedure(Procedure proc)
 { register Definition def = proc->definition;
 
-  def->flags ^= def->flags & ~SPY_ME;	/* Preserve the spy flag */
+  def->flags ^= def->flags & ~(SPY_ME|NEEDSCLAUSEGC);
   set(def, TRACE_ME);
   def->indexCardinality = 0;
   def->number_of_clauses = 0;
@@ -79,21 +87,36 @@ resetProcedure(Procedure proc)
 }
 
 Procedure
-isCurrentProcedure(FunctorDef f, Module m)
+isCurrentProcedure(functor_t f, Module m)
 { Symbol s;
 
-  if ((s = lookupHTable(m->procedures, f)) != (Symbol) NULL)
+  if ( (s = lookupHTable(m->procedures, (void *)f)) )
     return (Procedure) s->value;
 
-  return (Procedure) NULL;
+  return NULL;
 }
 
 bool
 isDefinedProcedure(Procedure proc)
-{ if ( /* true(proc->definition, FOREIGN) || not needed; union */
-       proc->definition->definition.clauses ||
-       true(proc->definition, DYNAMIC) )
+{ Definition def = proc->definition;
+
+  if ( true(def, DYNAMIC|FOREIGN) )
     succeed;
+
+  if ( def->definition.clauses && false(def, FOREIGN) )
+  { ClauseRef c;
+
+    if ( false(def, NEEDSCLAUSEGC) )
+      succeed;
+    
+    for(c = def->definition.clauses; c; c = c->next)
+    { Clause cl = c->clause;
+
+      if ( false(cl, ERASED) )
+	succeed;
+    }
+  }
+
   fail;
 }
 
@@ -104,7 +127,7 @@ to be defined is a system predicate.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Procedure
-lookupProcedureToDefine(FunctorDef def, Module m)
+lookupProcedureToDefine(functor_t def, Module m)
 { Procedure proc;
 
   if ( (proc = isCurrentProcedure(def, m)) && false(proc->definition, SYSTEM) )
@@ -126,6 +149,80 @@ lookupProcedureToDefine(FunctorDef def, Module m)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+get_functor() translates term  of  the   format  +Name/+Arity  into  the
+internal functor represenation. It fails and  raises an exception on the
+various possible format or represenation errors.  ISO compliant.
+
+The return value is 1 normally, -1  if no functor exists and GF_EXISTING
+is defined, and 0 if an error was raised.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#define GF_EXISTING	1
+#define GF_PROCEDURE	2		/* check for max arity */
+
+static int
+get_functor(term_t descr, functor_t *fdef, Module *m, term_t h, int how)
+{ term_t head = PL_new_term_ref();
+
+  if ( !PL_strip_module(descr, m, head) )
+    fail;
+
+  if ( PL_is_functor(head, FUNCTOR_divide2) )
+  { term_t a = PL_new_term_ref();
+    atom_t name;
+    int arity;
+
+    PL_get_arg(1, head, a);
+    if ( PL_get_atom(a, &name) )
+    { PL_get_arg(2, head, a);
+      if ( PL_get_integer(a, &arity) )
+      { if ( arity < 0 )
+	{ return PL_error(NULL, 0, NULL, ERR_DOMAIN,
+			  ATOM_not_less_than_zero, a);
+	} else if ( (how&GF_PROCEDURE) && arity > MAXARITY )
+	{ char buf[100];
+
+	  return PL_error(NULL, 0,
+			  tostr(buf, "limit is %d, request = %d",
+				MAXARITY, arity),
+			  ERR_REPRESENTATION, ATOM_max_arity);
+	} else
+	{ *fdef = PL_new_functor(name, arity);
+	  
+	  if ( h )
+	    PL_put_term(h, head);
+	  
+	  succeed;
+	}
+      } else
+      { if ( PL_is_variable(a) )
+	  goto ierror;
+
+	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, a);
+      }
+    } else
+    { if ( PL_is_variable(a) )
+	goto ierror;
+
+      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, a);
+    }
+  } else if ( PL_get_functor(head, fdef) )
+  { if ( h )
+      PL_put_term(h, head);
+	  
+    succeed;
+  } else
+  { if ( PL_is_variable(head) )
+    { ierror:
+      return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
+    } else
+      return PL_error(NULL, 0, NULL, ERR_TYPE,
+		      ATOM_predicate_indicator, head);
+  }
+}
+
+      
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Get the specified procedure from a   Prolog  argument.  This argument is
 either a head or a term of the form module:head.  If `create' is TRUE, a
@@ -136,16 +233,34 @@ module-inheritance chain to find the existing procedure.
 int
 get_procedure(term_t descr, Procedure *proc, term_t h, int how)
 { Module m = (Module) NULL;
-  term_t head = PL_new_term_ref();
-  FunctorDef fdef;
+  functor_t fdef;
   Procedure p;
 
-  if ( !PL_strip_module(descr, &m, head) )
-    fail;
-  if ( !PL_get_functor(head, &fdef) )
-    return warning("Illegal predicate specification");
+  if ( (how&GP_NAMEARITY) )
+  { if ( !get_functor(descr, &fdef, &m, h, GF_PROCEDURE) )
+      fail;
+  } else
+  { term_t head = PL_new_term_ref();
+    int arity;
+
+    if ( !PL_strip_module(descr, &m, head) )
+      fail;
+
+    if ( h )
+      PL_put_term(h, head);
+
+    if ( !PL_get_functor(head, &fdef) )
+      return warning("Illegal predicate specification");
+    if ( (arity=arityFunctor(fdef)) > MAXARITY )
+    { char buf[100];
+      return PL_error(NULL, 0,
+			  tostr(buf, "limit is %d, request = %d",
+				MAXARITY, arity),
+			  ERR_REPRESENTATION, ATOM_max_arity);
+    }
+  }
   
-  switch( how )
+  switch( how & GP_HOW_MASK )
   { case GP_CREATE:
       *proc = lookupProcedure(fdef, m);
       break;
@@ -180,8 +295,6 @@ get_procedure(term_t descr, Procedure *proc, term_t h, int how)
   }
 
 out:
-  if ( h )
-    PL_put_term(h, head);
 
   succeed;
 }
@@ -201,7 +314,7 @@ to C.
 word
 pl_current_predicate(term_t name, term_t spec, word h)
 { atom_t n;
-  FunctorDef f;
+  functor_t f;
   Module m = (Module) NULL;
   Procedure proc;
   Symbol symb;
@@ -221,7 +334,7 @@ pl_current_predicate(term_t name, term_t spec, word h)
   }
   if ( !PL_get_functor(functor, &f) )
   { if ( PL_is_variable(functor) )
-      f = NULL;
+      f = 0;
     else
       fail;
   }
@@ -229,7 +342,7 @@ pl_current_predicate(term_t name, term_t spec, word h)
   if ( ForeignControl(h) == FRG_FIRST_CALL)
   { if ( f ) 
     { if ( (proc = isCurrentProcedure(f, m)) )
-	return PL_unify_atom(name, f->name);
+	return PL_unify_atom(name, nameFunctor(f));
       fail;
     }
     symb = firstHTable(m->procedures);
@@ -242,9 +355,9 @@ pl_current_predicate(term_t name, term_t spec, word h)
     proc = (Procedure) symb->value;
     fdef = proc->definition->functor;
 
-    if ( ( n && n != fdef->name) ||
+    if ( (n && n != fdef->name) ||
 	 !PL_unify_atom(name, fdef->name) ||
-	 !PL_unify_functor(functor, fdef) )
+	 !PL_unify_functor(functor, fdef->functor) )
       continue;
 
     if ( (symb = nextHTable(m->procedures, symb)) )
@@ -460,7 +573,7 @@ freeClause(Clause c)
     clearBreakPointsClause(c);
 #endif
 
-  statistics.codes -= c->code_size;
+  GD->statistics.codes -= c->code_size;
   freeHeap(c->codes, sizeof(code) * c->code_size);
   freeHeap(c, sizeof(struct clause));
 }
@@ -552,7 +665,7 @@ void
 resetReferences(void)
 { Symbol s;
 
-  for_table(s, moduleTable)
+  for_table(s, GD->tables.modules)
     resetReferencesModule((Module) s->value);
 }
 
@@ -617,13 +730,12 @@ error message (or link the procedure from the library via autoload).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Procedure
-resolveProcedure(FunctorDef f, Module module)
+resolveProcedure(functor_t f, Module module)
 { Procedure proc;
   Module m;
 
-  for( m = module; m != (Module) NULL; m = m->super )
-  { if ( (proc = isCurrentProcedure(f, m)) != (Procedure) NULL &&
-       isDefinedProcedure(proc) )
+  for( m = module; m; m = m->super )
+  { if ( (proc = isCurrentProcedure(f, m)) && isDefinedProcedure(proc) )
       return proc;
   }
 
@@ -645,15 +757,14 @@ decided its not save, but can't recall why ...)
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Definition
-autoImport(FunctorDef f, Module m)
+autoImport(functor_t f, Module m)
 { Procedure proc;
   Definition def;
 					/* Defined: no problem */
-  if ( (proc = isCurrentProcedure(f, m)) != NULL &&
-       isDefinedProcedure(proc) )
+  if ( (proc = isCurrentProcedure(f, m)) && isDefinedProcedure(proc) )
     return proc->definition;
   
-  if ( m->super == (Module) NULL )	/* No super: can't import */
+  if ( !m->super )			/* No super: can't import */
     return NULL;
 
   if ( !(def = autoImport(f, m->super)) )
@@ -684,7 +795,7 @@ trapUndefined(Definition def)
 
   retry:
 					/* Auto import */
-  if ( (newdef = autoImport(functor, module)) )
+  if ( (newdef = autoImport(functor->functor, module)) )
     return newdef;
 					/* Pred/Module does not want to trap */
   if ( true(def, DYNAMIC|MULTIFILE|DISCONTIGUOUS) || false(module, UNKNOWN) )
@@ -693,7 +804,7 @@ trapUndefined(Definition def)
   DEBUG(5, Sdprintf("trapUndefined(%s)\n", predicateName(def)));
 
 					/* Trap via exception/3 */
-  if ( status.autoload )
+  if ( LD->autoload )
   { if ( undefined_nesting > 100 )
     { undefined_nesting = 1;
       sysError("trapUndefined(): undefined: %s", predicateName(def));
@@ -717,7 +828,7 @@ trapUndefined(Definition def)
       PL_put_variable(argv+3);
 
       undefined_nesting++;
-      qid = PL_open_query(MODULE_system, FALSE, pred, argv);
+      qid = PL_open_query(MODULE_system, PL_Q_NODEBUG, pred, argv);
       if ( PL_next_solution(qid) )
 	PL_get_atom(argv+3, &answer);
       PL_close_query(qid);
@@ -726,7 +837,7 @@ trapUndefined(Definition def)
       source_line_no   = sln;
       PL_discard_foreign_frame(cid);
 
-      def = lookupProcedure(functor, module)->definition;
+      def = lookupProcedure(functor->functor, module)->definition;
 
       if ( answer == ATOM_fail )
 	return def;
@@ -754,7 +865,10 @@ word
 pl_require(term_t pred)
 { Procedure proc;
 
-  return get_procedure(pred, &proc, 0, GP_DEFINE);
+  if ( !get_procedure(pred, &proc, 0, GP_RESOLVE) )
+    return get_procedure(pred, &proc, 0, GP_DEFINE);
+
+  succeed;
 }
 
 
@@ -785,7 +899,7 @@ pl_retract(term_t term, word h)
       return warning("retract/1: illegal clause");
 
     if ( ForeignControl(h) == FRG_FIRST_CALL )
-    { FunctorDef fd;
+    { functor_t fd;
 
       if ( !PL_get_functor(head, &fd) )
 	return warning("retract/1: illegal head");
@@ -910,7 +1024,7 @@ pl_retractall(term_t head)
 
 word
 pl_abolish(term_t atom, term_t arity)
-{ FunctorDef f;
+{ functor_t f;
   Procedure proc;
   Module m = (Module) NULL;
   term_t tmp = PL_new_term_ref();
@@ -927,7 +1041,34 @@ pl_abolish(term_t atom, term_t arity)
     succeed;
 
   if ( true(proc->definition, LOCKED) && !SYSTEM_MODE && m == MODULE_system )
-    return warning("abolish/2: attempt to abolish a system predicate");
+    return PL_error("abolish", 2, NULL, ERR_MODIFY_STATIC_PROC, proc);
+
+  return abolishProcedure(proc, m);
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+abolish(Name/Arity)
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+word
+pl_abolish1(term_t spec)
+{ Procedure proc;
+  functor_t f;
+  Module m = NULL;
+
+  switch( get_functor(spec, &f, &m, 0, GF_PROCEDURE|GF_EXISTING) )
+  { case FALSE:				/* exception */
+      fail;
+    case -1:				/* no functor */
+      succeed;
+  }
+
+  if ( !(proc = isCurrentProcedure(f, m)) )
+    succeed;
+
+  if ( true(proc->definition, LOCKED) && !SYSTEM_MODE && m == MODULE_system )
+    return PL_error("abolish", 1, NULL, ERR_MODIFY_STATIC_PROC, proc);
 
   return abolishProcedure(proc, m);
 }
@@ -963,7 +1104,7 @@ pl_get_predicate_attribute(term_t pred,
 			   term_t what, term_t value)
 { Procedure proc;
   Definition def;
-  FunctorDef fd;
+  functor_t fd;
   atom_t key;
   Module module = (Module) NULL;
   unsigned long att;
@@ -992,8 +1133,7 @@ pl_get_predicate_attribute(term_t pred,
   } else if ( key == ATOM_defined )
   { int d;
 
-    if ( def->definition.clauses ||
-	 true(def, DYNAMIC|MULTIFILE|DISCONTIGUOUS) )
+    if ( isDefinedProcedure(proc) )
       d = 1;
     else
       d = 0;
@@ -1036,7 +1176,6 @@ pl_set_predicate_attribute(term_t pred,
   atom_t key;
   int val;
   unsigned long att;
-  int nodef;				/* does not define pred */
 
   if ( !PL_get_atom(what, &key) ||
        !PL_get_integer(value, &val) || val & ~1 )
@@ -1044,10 +1183,13 @@ pl_set_predicate_attribute(term_t pred,
   if ( !(att = attribute_mask(key)) )
     return warning("$set_predicate_attribute/4: unknown key: %s",
 		   stringAtom(key));
-
-  nodef = (att & (TRACE_ANY|SPY_ME));
-  if ( !get_procedure(pred, &proc, 0, nodef ? GP_RESOLVE : GP_DEFINE) )
-    fail;
+  if ( att & (TRACE_ANY|SPY_ME) )
+  { if ( !get_procedure(pred, &proc, 0, GP_RESOLVE) )
+      fail;
+  } else
+  { if ( !get_procedure(pred, &proc, 0, GP_DEFINE|GP_NAMEARITY) )
+      fail;
+  }
   def = proc->definition;
 
   if ( !val )
@@ -1231,7 +1373,7 @@ lookupSourceFile(atom_t name)
   file->count = 0;
   file->time = 0L;
   file->index = ++source_index;
-  file->system = status.boot;
+  file->system = GD->bootsession;
   file->procedures = NULL;
   file->next = NULL;
 
@@ -1337,7 +1479,7 @@ pl_source_file(term_t descr, term_t file, control_t h)
     Definition def = proc->definition;
     fid_t cid = PL_open_foreign_frame();
 
-    if ( unify_definition(descr, def, 0) )
+    if ( unify_definition(descr, def, 0, 0) )
     { PL_close_foreign_frame(cid);
 
       if ( cell->next )
