@@ -1,4 +1,4 @@
-/*  $Id: pl-read.c,v 1.44 1997/03/10 09:37:03 jan Exp $
+/*  $Id: pl-read.c,v 1.52 1997/10/28 13:41:07 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -59,10 +59,10 @@ struct token
 
 
 struct variable
-{ struct atom	atom;		/* Normal atom ($VAR/0) */
-  char *	name;		/* Name of the variable */
+{ char *	name;		/* Name of the variable */
   term_t	variable;	/* Term-reference to the variable */
   int		times;		/* Number of occurences */
+  word		signature;	/* Pseudo atom */
   Variable 	next;		/* Next of chain */
 };
 
@@ -77,10 +77,10 @@ struct variable
 #define T_FULLSTOP	7	/* Prolog end of clause */
 
 extern int Input;		/* current input stream (from pl-file.c) */
-static char *here;		/* current character */
-static char *base;		/* base of clause */
-static char *token_start;	/* start of most recent read token */
-static char *last_syntax_error;	/* last syntax error */
+static unsigned char *here;		/* current character */
+static unsigned char *base;		/* base of clause */
+static unsigned char *token_start;	/* start of most recent read token */
+static char *last_syntax_error; /* last syntax error */
 static struct token token;	/* current token */
 static bool unget = FALSE;	/* unget_token() */
 
@@ -100,8 +100,8 @@ static
 struct read_buffer
 { int	size;			/* current size of read buffer */
   int	left;			/* left space in read buffer */
-  char *base;			/* base of read buffer */
-  char *here;			/* current position in read buffer */
+  unsigned char *base;		/* base of read buffer */
+  unsigned char *here;		/* current position in read buffer */
   int   stream;			/* stream we are reading from */
 } rb;
 
@@ -191,7 +191,7 @@ errorWarning(char *what)
   } else
   { predicate_t pred = PL_pred(FUNCTOR_exception3, MODULE_user);
     int rval;
-    char *s;
+    unsigned char *s;
 
     for(s = base; s < token_start; s++ )
     { if ( *s == '\n' )
@@ -211,7 +211,7 @@ errorWarning(char *what)
       PL_get_arg(2, argv+1, a); PL_unify_integer(a, source_line_no);
       PL_get_arg(3, argv+1, a); PL_unify_atom_chars(a, what);
 	       
-      qid = PL_open_query(MODULE_user, FALSE, pred, argv);
+      qid = PL_open_query(MODULE_user, PL_Q_NODEBUG, pred, argv);
       rval = PL_next_solution(qid);
       PL_close_query(qid);
       PL_discard_foreign_frame(cid);
@@ -248,7 +248,7 @@ singletonWarning(atom_t *vars, int nvars)
   }
   PL_unify_nil(a);
 
-  qid = PL_open_query(MODULE_user, FALSE, pred, argv);
+  qid = PL_open_query(MODULE_user, PL_Q_NODEBUG, pred, argv);
   rval = PL_next_solution(qid);
   PL_close_query(qid);
   PL_discard_foreign_frame(cid);
@@ -289,7 +289,7 @@ reading.
 static void
 clearBuffer()
 { if (rb.size == 0)
-  { if ((rb.base = Malloc(RBSIZE)) == (char *) NULL)
+  { if ( !(rb.base = (unsigned char *)malloc(RBSIZE)) )
       fatalError("%s", OsError());
     rb.size = RBSIZE;
   }
@@ -305,7 +305,7 @@ clearBuffer()
 static inline void
 addToBuffer(int c)
 { if (rb.left-- == 0)
-  { if ((rb.base = Realloc(rb.base, rb.size * 2)) == (char *)NULL)
+  { if ( !(rb.base = (unsigned char *)realloc(rb.base, rb.size * 2)) )
       fatalError("%s", OsError());
     DEBUG(8, Sdprintf("Reallocated read buffer at %ld\n", (long) rb.base));
     base = rb.base;
@@ -415,8 +415,8 @@ raw_read2()
 		  }
 		} else
 		{ addToBuffer('/');
-		  if ( char_type[c] == SY )
-		  { while( c != EOF && char_type[c] == SY )
+		  if ( isSymbol(c) )
+		  { while( c != EOF && isSymbol(c) )
 		    { addToBuffer(c);
 		      c = getchr();
 		    }
@@ -492,15 +492,15 @@ raw_read2()
 		set_start_line;
 		dotseen++;
 		c = getchr();
-		if ( char_type[c] == SY )
-		{ while( c != EOF && char_type[c] == SY )
+		if ( isSymbol(c) )
+		{ while( c != EOF && isSymbol(c) )
 		  { addToBuffer(c);
 		    c = getchr();
 		  }
 		  dotseen = FALSE;
 		}
 		goto handle_c;
-      default:	switch(char_type[c])
+      default:	switch(_PL_char_types[c])
 		{ case SP:
 		    if ( dotseen )
 		    { if ( rb.here - rb.base == 1 )
@@ -515,14 +515,14 @@ raw_read2()
 		      else
 			ensure_space(c);
 		      c = getchr();
-		    } while( c != EOF && char_type[c] == SP );
+		    } while( c != EOF && isBlank(c) );
 		    goto handle_c;
 		  case SY:
 		    set_start_line;
 		    do
 		    { addToBuffer(c);
 		      c = getchr();
-		    } while( c != EOF && char_type[c] == SY );
+		    } while( c != EOF && isSymbol(c) == SY );
 		    dotseen = FALSE;
 		    goto handle_c;
 		  case LC:
@@ -531,8 +531,7 @@ raw_read2()
 		    do
 		    { addToBuffer(c);
 		      c = getchr();
-		    } while( c != EOF &&
-			     (char_type[c] == LC || char_type[c] == UC) );
+		    } while( c != EOF && isAlpha(c) );
 		    dotseen = FALSE;
 		    goto handle_c;
 		  default:
@@ -572,7 +571,7 @@ raw_read(void)
 These functions manipulate the  variable-name  base   for  a  term. When
 building the term on the global stack, variables are represented using a
 reference to a `struct variable'. The first  part of this structure is a
-stuct atom, so if a garbage   collection  happens, the garbage collector
+struct atom, so if a garbage   collection happens, the garbage collector
 will simply assume an atom.
 
 The buffer `var_buffer' is a  list  if   pointers  to  a  stock of these
@@ -594,7 +593,15 @@ static Variable  var_tail;		/* tail of linked list */
 
 static void
 initVarTable()
-{ var_list = var_tail = NULL;
+{ static int done = 0;
+
+  if ( !done )
+  { initBuffer(&var_name_buffer);
+    initBuffer(&var_buffer);
+    done++;
+  }
+
+  var_list = var_tail = NULL;
   var_free = 0;
   emptyBuffer(&var_name_buffer);
 }
@@ -620,8 +627,14 @@ save_var_name(const char *name)
 
 					/* use hash-key? */
 
-static char *uniquename = "read/1 tmp var";
-#define isVarAtom(w) (isAtom(w) && stringAtom(w) == uniquename)
+static Variable
+isVarAtom(word w)
+{ if ( tagex(w) == (TAG_ATOM|STG_GLOBAL) )
+    return baseBuffer(&var_buffer, Variable)[w>>7];
+  
+  return NULL;
+}
+
 
 static Variable
 lookupVariable(const char *name)
@@ -636,10 +649,8 @@ lookupVariable(const char *name)
        
   while ( var_free >= var_allocated )
   { Variable v = allocHeap(sizeof(struct variable));
-    v->atom.next        = NULL;
-    v->atom.hash_value  = 0;
-    v->atom.name	= uniquename;
     addBuffer(&var_buffer, v, Variable);
+    v->signature = (var_allocated<<7)|TAG_ATOM|STG_GLOBAL;
     var_allocated++;
   }
 
@@ -734,11 +745,11 @@ forwards void	build_term(term_t term, atom_t name, int arity, term_t *args);
 forwards bool	complex_term(const char *end, term_t term, term_t positions);
 forwards bool	simple_term(bool, term_t term, bool *isname, term_t positions);
 
-int
+static int
 scan_number(char **s, int b, Number n)
 { int d;
-  unsigned int maxi = PLMAXINT/b;	/* cache? */
-  unsigned int t = 0;
+  unsigned long maxi = PLMAXINT/b;	/* cache? */
+  unsigned long t = 0;
 
   while((d = digitValue(b, **s)) >= 0)
   { (*s)++;
@@ -779,10 +790,10 @@ scan_number(char **s, int b, Number n)
 	{ if ( n >= bufsize ) \
 	  { if ( bufsize == 0 ) \
 	    { bufsize = 512; \
-	      buf = Malloc(bufsize); \
+	      buf = malloc(bufsize); \
 	    } else \
 	    { bufsize *= 2; \
-	      buf = Realloc(buf, bufsize); \
+	      buf = realloc(buf, bufsize); \
 	    } \
 	    if ( buf == NULL ) \
 	      fatalError("%s", OsError()); \
@@ -792,7 +803,7 @@ scan_number(char **s, int b, Number n)
 		
 
 static char *
-get_string(char *in, char **end)
+get_string(unsigned char *in, unsigned char **end)
 { static char *buf;
   static int bufsize = 0;
   int n;
@@ -886,9 +897,10 @@ get_string(char *in, char **end)
 
 
 int
-get_number(char *in, char **end, Number value)
+get_number(const unsigned char *cin, unsigned char **end, Number value)
 { int negative = FALSE;
   unsigned int c;
+  char *in = (char *)cin;		/* const hack */
 
   if ( *in == '-' )			/* skip optional sign */
   { negative = TRUE;
@@ -1002,9 +1014,6 @@ get_number(char *in, char **end, Number value)
 			               : (double)exponent.value.i);
   }
 
-  if ( isAlpha(c = *in) )
-    fail;				/* illegal number */
-
   if ( negative )
   { if ( intNumber(value) )
       value->value.i = -value->value.i;
@@ -1020,7 +1029,7 @@ get_number(char *in, char **end, Number value)
 static Token
 get_token(bool must_be_op)
 { unsigned int c;
-  char *start;
+  unsigned char *start;
   int end;
 
   if ( unget )
@@ -1031,7 +1040,7 @@ get_token(bool must_be_op)
   skipSpaces;
   token_start = here - 1;
   token.start = source_char_no + token_start - base;
-  switch(char_type[c])
+  switch(_PL_char_types[c])
   { case LC:	{ start = here-1;
 		  while(isAlpha(*here) )
 		    here++;
@@ -1050,6 +1059,12 @@ get_token(bool must_be_op)
 		    here++;
 		  c = *here;
 		  *here = EOS;
+		  if ( c == '(' && trueFeature(ALLOW_VARNAME_FUNCTOR) )
+		  { token.value.atom = lookupAtom(start);
+		    token.type = T_FUNCTOR;
+		    *here = c;
+		    break;
+		  }
 		  if (start[0] == '_' && here == start + 1)
 		  { DEBUG(9, Sdprintf("VOID\n"));
 		    token.type = T_VOID;
@@ -1066,7 +1081,8 @@ get_token(bool must_be_op)
     case_digit:
     case DI:	{ number value;
 
-		  if ( get_number(&here[-1], &here, &value) )
+		  if ( get_number(&here[-1], &here, &value) &&
+		       !isAlpha(here[0]) )
 		  { token.value.number = value;
 		    token.type = T_NUMBER;
 		    break;
@@ -1101,7 +1117,7 @@ get_token(bool must_be_op)
 		  }
 
 		  end = *here, *here = EOS;
-		  token.value.atom = lookupAtom(start);
+		  token.value.atom = lookupAtom((char *)start);
 		  *here = end;
 
 		  token.type = (end == '(' ? T_FUNCTOR : T_NAME);
@@ -1181,11 +1197,10 @@ statically allocated and thus unique.
 static inline void
 readValHandle(term_t term, Word argp)
 { word w = *valTermRef(term);
+  Variable var;
 
-  if ( isVarAtom(w) )
-  { Variable var = (Variable) atomValue(w);
-
-    DEBUG(2, Sdprintf("readValHandle(): var at 0x%x\n", var));
+  if ( (var = isVarAtom(w)) )
+  { DEBUG(9, Sdprintf("readValHandle(): var at 0x%x\n", var));
 
     if ( !var->variable )		/* new variable */
     { var->variable = PL_new_term_ref();
@@ -1201,12 +1216,12 @@ readValHandle(term_t term, Word argp)
 
 static void
 build_term(term_t term, atom_t atom, int arity, term_t *argv)
-{ FunctorDef functor = lookupFunctorDef(atom, arity);
+{ functor_t functor = lookupFunctorDef(atom, arity);
   Word argp = allocGlobal(arity+1);
 
   DEBUG(9, Sdprintf("Building term %s/%d ... ", stringAtom(atom), arity));
   setHandle(term, consPtr(argp, TAG_COMPOUND|STG_GLOBAL));
-  *argp++ = functor->functor;
+  *argp++ = functor;
 
   for( ; arity-- > 0; argv++, argp++)
     readValHandle(*argv, argp);
@@ -1242,7 +1257,7 @@ typedef struct
   short	right_pri;			/* priority at right hand */
   short	op_pri;				/* priority of operator */
   term_t tpos;				/* term-position */
-  char *token_start;			/* start of the token for message */
+  unsigned char *token_start;		/* start of the token for message */
 } op_entry;
 
 
@@ -1315,7 +1330,7 @@ the values.
 	if ( side_p >= 0 && cpri > side[side_p].right_pri ) \
 	{ term_t tmp; \
 	  if ( side[side_p].kind == OP_PREFIX && rmo == 0 ) \
-	  { DEBUG(1, Sdprintf("Prefix %s to atom\n", \
+	  { DEBUG(9, Sdprintf("Prefix %s to atom\n", \
 			      stringAtom(side[side_p].op))); \
 	    rmo++; \
 	    tmp = PL_new_term_ref(); \
@@ -1329,7 +1344,7 @@ the values.
 	    side_p = (side_n == 0 ? -1 : side_p-1); \
 	  } else if ( side[side_p].kind == OP_INFIX && out_n > 0 && rmo == 0 && \
 		      isOp(side[side_p].op, OP_POSTFIX, &side[side_p]) ) \
-	  { DEBUG(1, Sdprintf("Infix %s to postfix\n", \
+	  { DEBUG(9, Sdprintf("Infix %s to postfix\n", \
 			      stringAtom(side[side_p].op))); \
 	    rmo++; \
 	    tmp = PL_new_term_ref(); \
@@ -1457,8 +1472,6 @@ bad_operator(out_entry *out, op_entry *op)
   }
 
   syntaxError(buf);
-
-  fail;
 }
 
 #define Reduce(cpri) \
@@ -1471,7 +1484,7 @@ bad_operator(out_entry *out, op_entry *op)
               return bad_operator(&out[out_n-arity], &side[side_p]); \
 	    break; \
 	  } \
-	  DEBUG(1, Sdprintf("Reducing %s/%d\n", \
+	  DEBUG(9, Sdprintf("Reducing %s/%d\n", \
 			    stringAtom(side[side_p].op), arity));\
 	  tmp = PL_new_term_ref(); \
 	  out_n -= arity; \
@@ -1532,10 +1545,10 @@ complex_term(const char *stop, term_t term, term_t positions)
       in_op.tpos = pin;
       in_op.token_start = token_start;
 
-      DEBUG(1, Sdprintf("name %s, rmo = %d\n", stringAtom(name), rmo));
+      DEBUG(9, Sdprintf("name %s, rmo = %d\n", stringAtom(name), rmo));
 
       if ( isOp(name, OP_INFIX, &in_op) )
-      { DEBUG(1, Sdprintf("Infix op: %s\n", stringAtom(name)));
+      { DEBUG(9, Sdprintf("Infix op: %s\n", stringAtom(name)));
 
 	Modify(in_op.left_pri);
 	if ( rmo == 1 )
@@ -1547,7 +1560,7 @@ complex_term(const char *stop, term_t term, term_t positions)
 	}
       }
       if ( isOp(name, OP_POSTFIX, &in_op) )
-      { DEBUG(1, Sdprintf("Postfix op: %s\n", stringAtom(name)));
+      { DEBUG(9, Sdprintf("Postfix op: %s\n", stringAtom(name)));
 
 	Modify(in_op.left_pri);
 	if ( rmo == 1 )
@@ -1558,7 +1571,7 @@ complex_term(const char *stop, term_t term, term_t positions)
 	}
       }
       if ( rmo == 0 && isOp(name, OP_PREFIX, &in_op) )
-      { DEBUG(1, Sdprintf("Prefix op: %s\n", stringAtom(name)));
+      { DEBUG(9, Sdprintf("Prefix op: %s\n", stringAtom(name)));
 	
 	PushOp();
 
@@ -1613,8 +1626,8 @@ simple_term(bool must_be_op, term_t term, bool *name, term_t positions)
       setHandle(term, 0L);		/* variable */
       goto atomic_out;
     case T_VARIABLE:
-      setHandle(term, consPtr(token->value.variable, TAG_ATOM|STG_HEAP));
-      DEBUG(2, Sdprintf("Pushed var at 0x%x\n", token->value.variable));
+      setHandle(term, token->value.variable->signature);
+      DEBUG(9, Sdprintf("Pushed var at 0x%x\n", token->value.variable));
       goto atomic_out;
     case T_NAME:
       *name = TRUE;
@@ -1774,7 +1787,7 @@ term is to be written.
 		argp = allocGlobal(3);
 		*unRef(*valTermRef(tail)) = consPtr(argp,
 						    TAG_COMPOUND|STG_GLOBAL);
-		*argp++ = FUNCTOR_dot2->functor;
+		*argp++ = FUNCTOR_dot2;
 		readValHandle(tmp, argp++);
 		setVar(*argp);
 		setHandle(tail, makeRef(argp));
@@ -1897,6 +1910,13 @@ pl_raw_read(term_t term)
   return PL_unify_atom_chars(term, s);
 }
 
+
+word
+pl_raw_read2(term_t stream, term_t term)
+{ streamInput(stream, pl_raw_read(term));
+}
+
+
 word
 pl_read_variables(term_t term, term_t variables)
 { return read_term(term, variables, 0, FALSE);
@@ -1928,7 +1948,7 @@ pl_read_clause2(term_t stream, term_t term)
 { streamInput(stream, pl_read_clause(term));
 }
 
-static opt_spec read_term_options[] = 
+static const opt_spec read_term_options[] = 
 { { ATOM_syntax_errors,	    OPT_TERM },	/* quiet, fail, X */
   { ATOM_variable_names,    OPT_TERM },
   { ATOM_singletons,        OPT_TERM },
@@ -1948,7 +1968,7 @@ pl_read_term(term_t term, term_t options)
   int rval;
   atom_t w;
 
-  if ( !scan_options(options, 0, read_term_options,
+  if ( !scan_options(options, 0, ATOM_read_option, read_term_options,
 		     &syntax_errors, &varnames, &singles, &tpos, &subtpos) )
     return warning("read_term/2: illegal option list");
 
