@@ -1,4 +1,4 @@
-/*  $Id: pl-save.c,v 1.34 1996/04/17 14:02:07 jan Exp $
+/*  $Id: pl-save.c,v 1.36 1997/08/07 07:58:38 jan Exp $
 
     Copyright (c) 1991 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -97,7 +97,7 @@ PORTABILITY/OPTIONS
 #undef DEBUG
 #define DEBUG(l, g) {g;}
 #define O_SAVE 1
-#define OsPath(x) x
+#define OsPath(x, y) x
 #define ResetTty()
 #define Sdprintf printf
 #define Ssprintf sprintf
@@ -276,7 +276,7 @@ topOfCStack(void)
 }
 
 static caddr
-baseOfCStack(int *argc, char **argv, char **env)
+baseOfCStack(int *argc, char **argv)
 { 
 #ifdef BASE_OF_C_STACK
   return (caddr) BASE_OF_C_STACK;
@@ -295,8 +295,6 @@ baseOfCStack(int *argc, char **argv, char **env)
   for(p = argv; *p; p++)
     base = BoundStack(base, (unsigned long)*p);
   base = BoundStack(base, (unsigned long)env);
-  for(p = env; *p; p++)
-    base = BoundStack(base, (unsigned long)*p);
 
 #if STACK_DIRECTION > 0
   return (caddr) (base & ~(STACK_BASE_ALIGN-1));
@@ -516,6 +514,7 @@ save(char *file, char *interpreter, int kind,
   int csects, nsects, sects_size;
   int n;
   SaveSection sect;
+  int tmp[MAXPATHLEN];
 
   fill_c_data_sections(interpreter, &csects, sects);
   nsects = csects + nsections;
@@ -527,9 +526,11 @@ save(char *file, char *interpreter, int kind,
     return warning("save/1: cannot write %s: %s\n", file, OsError());
   
 #if OS2
-  Ssprintf(buf, "/* Self-starting SWI-Prolog state */\r\n'@ECHO OFF'\r\nparse source . . name\r\n\"%s -r \" name arg(1)\r\nexit\r\n\032", OsPath(interpreter));
+  Ssprintf(buf, "/* Self-starting SWI-Prolog state */\r\n'@ECHO OFF'\r\nparse source . . name\r\n\"%s -r \" name arg(1)\r\nexit\r\n\032",
+	   OsPath(interpreter, tmp));
 #else
-  Ssprintf(buf, "#!/bin/sh\nexec %s -r $0 \"$@\"\n", OsPath(interpreter));
+  Ssprintf(buf, "#!/bin/sh\nexec %s -r $0 \"$@\"\n",
+	   OsPath(interpreter, tmp));
 #endif
   header_offset = strlen(buf) + 1; /* +1 to write the EOS too */
   DEBUG(1, Sdprintf("header_offset = %d\n", header_offset));
@@ -598,7 +599,7 @@ main() stub
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-PL_initialise(int argc, char **argv, char **env)
+PL_initialise(int argc, char **argv)
 { int rval;
 
 #ifdef FORCED_MALLOC_BASE
@@ -606,19 +607,18 @@ PL_initialise(int argc, char **argv, char **env)
   fprintf(stderr, "FORCED_MALLOC_BASE at 0x%08x\n", FORCED_MALLOC_BASE);
 #endif
 
-  c_stack_base = baseOfCStack(&argc, argv, env);
+  c_stack_base = baseOfCStack(&argc, argv);
 
   if ( setjmp(ret_main_ctx) )
   { DEBUG(1, Sdprintf("Restarting startProlog()\n"));
-    environ = env;
-    rval = startProlog(argc, argv, env);
+    rval = startProlog(argc, argv);
   } else
   { char **av;
 
 #ifndef TEST
     for(av=argv; *av; av++)
       if ( streq(*av, "-d") && av[1] )
-	status.debugLevel = atoi(av[1]);
+	GD->debug_level = atoi(av[1]);
 #endif
 
     if ( argc >= 3 && streq(argv[1], "-r") )
@@ -634,7 +634,7 @@ PL_initialise(int argc, char **argv, char **env)
     }
 #endif
 
-    rval = startProlog(argc, argv, env);
+    rval = startProlog(argc, argv);
   }
 
   return(rval);
@@ -647,14 +647,14 @@ PL_initialise(int argc, char **argv, char **env)
 #endif
 
 int
-PL_initialise(int argc, char **argv, char **env)
+PL_initialise(int argc, char **argv)
 { 
 #ifdef FORCED_MALLOC_BASE
   start_memory((void *)FORCED_MALLOC_BASE);
   Sdprintf("FORCED_MALLOC_BASE at 0x%08x\n", FORCED_MALLOC_BASE);
 #endif
 
-  return startProlog(argc, argv, env);
+  return startProlog(argc, argv);
 }
 
 #endif /*O_SAVE*/
@@ -699,7 +699,7 @@ OsError()
 
 
 int
-startProlog(int argc, char **argv, char **env)
+startProlog(int argc, char **argv)
 { char *interpreter = argv[0];
 
   argc--; argv++;
