@@ -18,6 +18,8 @@
 
 #undef ulong
 #define ulong unsigned long
+#undef max
+#define max(a,b) ((a) > (b) ? (a) : (b))
 
 #define K * 1024
 
@@ -50,39 +52,30 @@ void
 setupProlog(void)
 { DEBUG(1, Sdprintf("Starting Heap Initialisation\n"));
 
-  critical = 0;
-  aborted = FALSE;
+  GD->critical = 0;
+  LD->aborted = FALSE;
+  signalled = 0;
 
   startCritical;
+  initMemAlloc();
 #if HAVE_SIGNAL
   DEBUG(1, Sdprintf("Prolog Signal Handling ...\n"));
   initSignals();
 #endif
-  if ( status.dumped == FALSE )
-  { hBase = (char *)(~0L);
-    hTop  = (char *)NULL;
-  }
   DEBUG(1, Sdprintf("Stacks ...\n"));
-  initStacks(options.localSize, 
-	     options.globalSize, 
-	     options.trailSize, 
-	     options.argumentSize);
+  initStacks(GD->options.localSize, 
+	     GD->options.globalSize, 
+	     GD->options.trailSize, 
+	     GD->options.argumentSize);
 
   lTop = lBase;
   tTop = tBase;
   gTop = gBase;
   aTop = aBase;
 
-  { void *hbase = allocHeap(sizeof(word));
-
-    base_addresses[STG_HEAP]   = (ulong)hbase & ~0x007fffffL;
-    freeHeap(hbase, sizeof(word));
-  }
   base_addresses[STG_LOCAL]  = (unsigned long)lBase;
   base_addresses[STG_GLOBAL] = (unsigned long)gBase;
   base_addresses[STG_TRAIL]  = (unsigned long)tBase;
-  DEBUG(1, Sdprintf("base_addresses[STG_HEAP] = %p\n",
-		    base_addresses[STG_HEAP]));
   DEBUG(1, Sdprintf("base_addresses[STG_LOCAL] = %p\n",
 		    base_addresses[STG_LOCAL]));
   DEBUG(1, Sdprintf("base_addresses[STG_GLOBAL] = %p\n",
@@ -95,9 +88,9 @@ setupProlog(void)
   depth_reached = 0;
 #endif
 
-  PL_open_foreign_frame();
+  emptyStacks();
 
-  if ( status.dumped == FALSE )
+  if ( !GD->dumped )
   { DEBUG(1, Sdprintf("Atoms ...\n"));
     initAtoms();
     DEBUG(1, Sdprintf("Features ...\n"));
@@ -124,8 +117,8 @@ setupProlog(void)
     initWamTable();
   } else
   { resetReferences();
-    resetGC();			/* reset garbage collector */
-    stateList = (State) NULL;	/* all states are already in core */
+    resetGC();				/* reset garbage collector */
+    GD->stateList = (State) NULL;	/* all states are already in core */
   }
   DEBUG(1, Sdprintf("IO ...\n"));
   initIO();
@@ -135,16 +128,16 @@ setupProlog(void)
   getSymbols();
   DEBUG(1, Sdprintf("Term ...\n"));
   resetTerm();
-  status.io_initialised = TRUE;
+  GD->io_initialised = TRUE;
 
   endCritical;
 
   environment_frame = (LocalFrame) NULL;
-  statistics.inferences = 0;
+  LD->statistics.inferences = 0;
 #if O_STORE_PROGRAM || O_SAVE
-  cannot_save_program = NULL;
+  GD->cannot_save_program = NULL;
 #else
-  cannot_save_program = "Not supported on this machine";
+  GD->cannot_save_program = "Not supported on this machine";
 #endif
 
 #if O_XWINDOWS
@@ -159,14 +152,14 @@ setupProlog(void)
 Feature interface
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-int
+void
 CSetFeature(char *name, char *value)
-{ return setFeature(lookupAtom(name), lookupAtom(value));
+{ setFeature(lookupAtom(name), FT_ATOM, lookupAtom(value));
 }
 
 static void
 CSetIntFeature(char *name, long value)
-{ setFeature(lookupAtom(name), consInt(value));
+{ setFeature(lookupAtom(name), FT_INTEGER, value);
 }
 
 static void
@@ -218,12 +211,13 @@ initFeatures()
   CSetFeature("runtime",	"true");
   CSetFeature("debug_on_error",	"false");
   CSetFeature("report_error",	"false");
-#endif
+#else
   CSetFeature("debug_on_error",	"true");
   CSetFeature("report_error",	"true");
+#endif
 					/* ISO features */
-  setFeature(lookupAtom("max_integer"), heapLong(PLMAXINT));
-  setFeature(lookupAtom("min_integer"), heapLong(PLMININT));
+  CSetIntFeature("max_integer", PLMAXINT);
+  CSetIntFeature("min_integer", PLMININT);
   CSetIntFeature("max_tagged_integer", PLMAXTAGGEDINT);
   CSetIntFeature("min_tagged_integer", PLMINTAGGEDINT);
   CSetFeature("bounded",	"true");
@@ -234,7 +228,8 @@ initFeatures()
   CSetFeature("max_arity", "unbounded");
   CSetFeature("float_format", "%g");
   CSetFeature("character_escapes", "true");
-  CSetFeature("tty_control", status.notty ? "false" : "true");
+  CSetFeature("tty_control", GD->cmdline.notty ? "false" : "true");
+  CSetFeature("allow_variable_name_as_functor", "false");
 #if defined(__unix__) || defined(unix)
   CSetFeature("unix", "true");
 #endif
@@ -276,12 +271,14 @@ some day.
 #define HAVE_SIGNALS 1
 #endif
 
+#if !O_DEBUG
 static void
 fatal_signal_handler(int sig, int type, SignalContext scp, char *addr)
 { DEBUG(1, Sdprintf("Fatal signal %d\n", sig));
 
   deliverSignal(sig, type, scp, addr);
 }
+#endif
 
 
 #ifdef HAVE_SIGSETMASK
@@ -292,16 +289,16 @@ static void
 initSignals(void)
 { int n;
 
-  if ( status.dumped == FALSE )
+  if ( !GD->dumped )
   { for( n = 0; n < MAXSIGNAL; n++ )
-    { signalHandlers[n].os = signalHandlers[n].user = SIG_DFL;
-      signalHandlers[n].catched = FALSE;
+    { LD_sig_handler(n).os = LD_sig_handler(n).user = SIG_DFL;
+      LD_sig_handler(n).catched = FALSE;
     }
 
 #ifdef SIGTTOU
     pl_signal(SIGTTOU, SIG_IGN);
 #endif
-#if !O_DEBUG				/* just crash when debugging */
+#if !O_DEBUG && !defined(_DEBUG)	/* just crash when debugging */
     pl_signal(SIGSEGV, (handler_t)fatal_signal_handler);
     pl_signal(SIGILL,  (handler_t)fatal_signal_handler);
 #ifdef SIGBUS
@@ -310,8 +307,8 @@ initSignals(void)
 #endif
   } else
   { for( n = 0; n < MAXSIGNAL; n++ )
-      if ( signalHandlers[n].os != SIG_DFL )
-        signal(n, signalHandlers[n].os);
+      if ( LD_sig_handler(n).os != SIG_DFL )
+        signal(n, LD_sig_handler(n).os);
   }
 
 #ifdef HAVE_SIGGETMASK
@@ -330,6 +327,8 @@ resetSignals()
 #ifdef HAVE_SIGSETMASK			/* fixes Linux repeated ^C */
   sigsetmask(defsigmask);
 #endif
+
+  signalled = 0L;
 }
 
 
@@ -346,8 +345,8 @@ pl_signal(int sig, handler_t func)
 	      sig, (unsigned long)func, OsError());
 #endif
 
-    signalHandlers[sig].os = func;
-    signalHandlers[sig].catched = (func == SIG_DFL ? FALSE : TRUE);
+    LD_sig_handler(sig).os = func;
+    LD_sig_handler(sig).catched = (func == SIG_DFL ? FALSE : TRUE);
 
     return old;
   } else
@@ -360,11 +359,11 @@ deliverSignal(int sig, int type, SignalContext scp, char *addr)
 { typedef RETSIGTYPE (*uhandler_t)(int, int, void *, char *);
     
 #ifndef BSD_SIGNALS
-  signal(sig, signalHandlers[sig].os);	/* ??? */
+  signal(sig, LD_sig_handler(sig).os);	/* ??? */
 #endif
 
-  if ( signalHandlers[sig].user != SIG_DFL )
-  { uhandler_t uh = (uhandler_t)signalHandlers[sig].user;
+  if ( LD_sig_handler(sig).user != SIG_DFL )
+  { uhandler_t uh = (uhandler_t)LD_sig_handler(sig).user;
 
     (*uh)(sig, type, scp, addr);
     return;
@@ -373,8 +372,60 @@ deliverSignal(int sig, int type, SignalContext scp, char *addr)
   sysError("Unexpected signal: %d\n", sig);
 }
 
+
+void
+PL_handle_signals()
+{ typedef RETSIGTYPE (*uhandler_t)(int);
+
+  while(signalled)
+  { ulong mask = 1L;
+    int sig = 1;
+
+    for( ; ; mask <<= 1, sig++ )
+    { if ( signalled & mask )
+      { signalled &= ~mask;
+
+	if ( LD_sig_handler(sig).os == SIG_DFL )
+	{ fatalError("Unhandled signal: %d\n", sig);
+	} else if ( LD_sig_handler(sig).os != SIG_IGN )
+	{ uhandler_t uh = (uhandler_t)LD_sig_handler(sig).os;
+
+	  (*uh)(sig);
+	}				/* SIG_IGN: ignored */
+
+	break;
+      }
+    }
+  }
+}
+
 #endif /*HAVE_SIGNAL*/
 
+		 /*******************************
+		 *	       STACKS		*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Create nice empty stacks. exception_bin   and  exception_printed are two
+term-references that must be low on  the   stack  to  ensure they remain
+valid while the stack is unrolled after an exception.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+void
+emptyStacks()
+{ environment_frame = NULL;
+  fli_context       = NULL;
+  lTop = lBase;
+  tTop = tBase;
+  gTop = gBase;
+  aTop = aBase;
+
+  PL_open_foreign_frame();
+  exception_bin     = PL_new_term_ref();
+  exception_printed = PL_new_term_ref();
+}
+
+
 #if O_DYNAMIC_STACKS
 
 static void init_stack(Stack s, char *name,
@@ -423,11 +474,13 @@ interesting  to  do  this  automatically  at  certain points to minimise
 memory requirements.  How?
 
 Currently this mechanism can use mmap() and munmap() of SunOs 4.0 or the
-system-V shared memory primitives (if they meet certain criteria.
+system-V shared memory primitives (if they meet certain criteria).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #include <errno.h>
+#ifndef WIN32
 extern int errno;
+#endif /*WIN32*/
 
 static int size_alignment;	/* Stack sizes must be aligned to this */
 static int base_alignment;	/* Stack bases must be aligned to this */
@@ -445,6 +498,11 @@ align_base(long int x)
 { return x % base_alignment ? (x / base_alignment + 1) * base_alignment : x;
 }
 
+static long
+align_base_down(long int x)
+{ return (x / base_alignment) * base_alignment;
+}
+
 #ifdef MMAP_STACK
 #include <sys/mman.h>
 #include <fcntl.h>
@@ -495,8 +553,9 @@ get_map_fd()
       for(n=1024, s = buf; n > 0; n--)
         *s++ = EOS;
       for(n=size_alignment/1024; n > 0; n--)
-        if ( write(fd, buf, 1024) != 1024 )
+      { if ( write(fd, buf, 1024) != 1024 )
           fatalError("Failed to create map file %s: %s\n", map, OsError());
+      }
 
       return fd;
     }
@@ -508,6 +567,58 @@ get_map_fd()
 }
 #endif /*HAVE_MAP_ANON*/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Estimate the top if the heap. The default is to get the size of the heap
+using getrlimit(), add this to the estimated  base and use the result as
+top address.
+
+This does not always appewar  to  work.   If  you  know the top, #define
+TOPOFHEAP in config.h. Othewise #define  it  to   0,  in  which case the
+system will allocate a default heap of 64 MB and the stacks above that.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#ifdef HAVE_GETRLIMIT
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
+
+#ifdef RLIMIT_DATA
+static ulong
+dataLimit()
+{ struct rlimit limit;
+
+  if ( getrlimit(RLIMIT_DATA, &limit) == 0 )
+    return limit.rlim_cur;
+
+  return 0L;
+}
+#else
+#define dataLimit() (0L)
+#endif /*RLIMIT_DATA*/
+#else
+#define dataLimit() (0L)
+#endif /*HAVE_GETRLIMIT*/
+
+
+#ifdef TOPOFHEAP
+#define topOfHeap() TOPOFHEAP
+#else /*TOPOFHEAP*/
+ulong
+topOfHeap()
+{ ulong data = dataLimit();
+
+  if ( data )
+  { ulong top = heap_base + data;
+
+    DEBUG(1, Sdprintf("Heap: %p ... %p\n", (void *)heap_base, (void *)top));
+    return top;
+  }
+    
+  return 0L;
+}
+#endif /*TOPOFHEAP*/
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Expand stack `s' by one page.  This might not be  enough,  but  in  this
 (very  rare) case another segmentation fault will follow to get the next
@@ -543,12 +654,13 @@ mapOrOutOf(Stack s)
 }
 
 
+#ifdef NO_SEGV_HANDLING
 void
 ensureRoomStack(Stack s, int bytes)
 { while((char *)s->max - (char *)s->top < (int)bytes)
     mapOrOutOf(s);
 }
-
+#endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 unmap() returns all memory resources of a stack that are  no  longer  in
@@ -568,6 +680,8 @@ unmap(Stack s)
 }
 
 
+#ifdef O_SAVE
+
 static void
 deallocateStack(Stack s)
 { long len = (unsigned long)s->max - (unsigned long)s->base;
@@ -579,15 +693,13 @@ deallocateStack(Stack s)
 
 void
 deallocateStacks(void)
-{ deallocateStack((Stack) &stacks.local);
-  deallocateStack((Stack) &stacks.global);
-  deallocateStack((Stack) &stacks.trail);
-  deallocateStack((Stack) &stacks.argument);
+{ deallocateStack((Stack) &LD->stacks.local);
+  deallocateStack((Stack) &LD->stacks.global);
+  deallocateStack((Stack) &LD->stacks.trail);
+  deallocateStack((Stack) &LD->stacks.argument);
 }
 
 
-#ifdef O_SAVE
-
 bool
 restoreStack(Stack s)
 { caddress max;
@@ -840,12 +952,13 @@ mapOrOutOf(Stack s)
 }
 
 
+#ifdef NO_SEGV_HANDLING
 void
 ensureRoomStack(Stack s, int bytes)
 { while((char *)s->max - (char *)s->top < (int)bytes)
     mapOrOutOf(s);
 }
-
+#endif
 
 static void
 unmap(Stack s)
@@ -915,7 +1028,7 @@ initStacks(long local, long global, long trail, long argument)
 
 #define INIT_STACK(name, print, limit, minsize) \
   DEBUG(1, Sdprintf("%s stack at 0x%x; size = %ld\n", print, base, limit)); \
-  init_stack((Stack) &stacks.name, print, (caddress) base, limit, minsize); \
+  init_stack((Stack) &LD->stacks.name, print, (caddress) base, limit, minsize); \
   base += limit + STACK_SEPARATION; \
   base = align_base(base);
 #define K * 1024
@@ -936,18 +1049,12 @@ Reset the stacks after an abort
 
 void
 resetStacks()
-{ environment_frame = NULL;
-  fli_context       = NULL;
-  lTop = lBase;
-  tTop = tBase;
-  gTop = gBase;
-  aTop = aBase;
+{ emptyStacks();
 
 #ifndef NO_SEGV_HANDLING
   pl_signal(SIGSEGV, (handler_t) segv_handler);
 #endif
   trimStacks();
-  PL_open_foreign_frame();
 }
 
 #endif /*HAVE_VIRTUAL_ALLOC*/
@@ -967,7 +1074,7 @@ segv_handler(int sig, int type, SignalContext scp, char *addr)
 #else
 segv_handler(int sig)
 #endif
-{ Stack stacka = (Stack) &stacks;
+{ Stack stacka = (Stack) &LD->stacks;
   int i;
 
 #ifndef SIGNAL_HANDLER_PROVIDES_ADDRESS
@@ -1055,15 +1162,13 @@ initStacks() initialises the stacks structure,  thus  assigning  a  base
 address,  a limit and a name to each of the stacks.  Finally it installs
 a signal handler for handling  segmentation  faults.   The  segmentation
 fault handler will actually create and expand the stacks on segmentation
-faults.   Currently,  it is assumed memory can be mapped from sbrk(0) to
-MMAP_MAX_ADDRESS and the stack is outside this area.  This is true on
-SUN  and  GOULD.   On  other  machines  the  C-stack  might   start   at
-MMAP_MAX_ADDRESS   and   grow   downwards.    In   this   case  lower
-MMAP_MAX_ADDRESS a bit (if you have 100 MB virtual address  space  or
-more,  I  would  suggest  16 MB), so space is allocated for the C-stack.
-The stacks are allocated right below MMAP_MAX_ADDRESS, at a  distance
-STACK_SEPARATION  from  each other.  STACK_SEPARATION must be a multiple
-of the page size.
+faults.
+
+The big problem is finding a safe   area  for the stacks. Currently, the
+system tries to find an area as far   as possible from the heap, growing
+downwards  if  it  can  determine  the    top  of  the  heap-area  using
+topOfHeap(). If it cannot, it will work from the current top of the heap
+as returned by sbrk(0).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifdef FORCED_MALLOC_BASE
@@ -1077,9 +1182,8 @@ of the page size.
 
 static void
 initStacks(long local, long global, long trail, long argument)
-{ long heap = 0;			/* malloc() heap */
-  int large = 1;
-  unsigned long base, top, space, large_size;
+{ int large = 0;
+  ulong base, top, space, large_size, min_space;
 
   size_alignment = getpagesize();
 #ifdef MMAP_STACK
@@ -1092,48 +1196,82 @@ initStacks(long local, long global, long trail, long argument)
 		  base_alignment, base_alignment));
 #endif
 
-  local    = (long) align_size(local);	/* Round up to page boundary */
-  global   = (long) align_size(global);
-  trail    = (long) align_size(trail);
-  argument = (long) align_size(argument);
+  local    = (ulong) align_size(local);	/* Round up to page boundary */
+  global   = (ulong) align_size(global);
+  trail    = (ulong) align_size(trail);
+  argument = (ulong) align_size(argument);
+
+  min_space = align_base(1) +
+	      align_base(local + STACK_SEPARATION) +
+	      align_base(global + STACK_SEPARATION) +
+	      align_base(trail + STACK_SEPARATION) +
+	      align_base(argument);
 
   if ( local    == 0 ) large++;		/* find dynamic ones */
   if ( global   == 0 ) large++;
   if ( trail    == 0 ) large++;
   if ( argument == 0 ) large++;
 
+  if ( (top = topOfHeap()) > 0L && !GD->options.heapSize )
+  { if ( large > 0 )			/* we have dynamic stacks */
+    { base = heap_base;
+      space = top - base;
+      space -= min_space;
+      large++;				/* heap as well */
+      large_size = ((space / large+1) / base_alignment) * base_alignment;
+      if ( large_size > 64 MB )
+	large_size = 64 MB;
+      if ( large_size < STACK_MINIMUM )
+	fatalError("Can't fit requested stack sizes in address space");
+      DEBUG(1, Sdprintf("Large stacks are %ld\n", large_size));
+
+      if ( local    == 0 ) local    = large_size;
+      if ( global   == 0 ) global   = large_size;
+      if ( trail    == 0 ) trail    = large_size;
+      if ( argument == 0 ) argument = large_size;
+    }
+
+    base = top - (align_base(1) +
+		  align_base(local + STACK_SEPARATION) +
+		  align_base(global + STACK_SEPARATION) +
+		  align_base(trail + STACK_SEPARATION) +
+		  align_base(argument));
+    base = align_base_down(base);
+  } else				/* we don't know the top */
+  { ulong maxdata = dataLimit();
+
+    if ( !GD->options.heapSize )
+    { if ( maxdata )
+      { large_size = align_base_down((maxdata-min_space)/(large+1));
+	large_size = max(large_size, 64 MB);
+      } else
+	large_size = 64 MB;
+	
+      GD->options.heapSize = large_size;
+    } else
+    { large_size = align_base_down((maxdata-min_space)/(large+1));
+      large_size = max(large_size, 64 MB);
+    }
+
 #ifdef MMAP_MIN_ADDRESS
-  base	= MMAP_MIN_ADDRESS;
+    base = MMAP_MIN_ADDRESS;
 #else
-  base  = (long) align_base((long)sbrk(0));
+    base = (ulong) align_base((ulong)sbrk(0) + GD->options.heapSize);
 #endif
-  top   = (long) MMAP_MAX_ADDRESS;
-  DEBUG(1, Sdprintf("top = 0x%x, stack at 0x%x\n", top, (unsigned) &top));
-  space = top - base;
-  space -= align_base(heap) +
-           align_base(local + STACK_SEPARATION) +
-	   align_base(global + STACK_SEPARATION) +
-	   align_base(trail + STACK_SEPARATION) +
-	   align_base(argument);
+
+    if ( large > 0 )
+    { DEBUG(1, Sdprintf("Large stacks are %ld\n", large_size));
   
-  large_size = ((space / large) / base_alignment) * base_alignment;
-  if ( large_size > 64 MB )
-    large_size = 64 MB;
-  if ( large_size < STACK_MINIMUM )
-    fatalError("Can't fit requested stack sizes in address space");
-  DEBUG(1, Sdprintf("Large stacks are %ld\n", large_size));
-
-  heap                          = large_size;
-  if ( local    == 0 ) local    = large_size;
-  if ( global   == 0 ) global   = large_size;
-  if ( trail    == 0 ) trail    = large_size;
-  if ( argument == 0 ) argument = large_size;
-
-  base += heap;
+      if ( local    == 0 ) local    = large_size;
+      if ( global   == 0 ) global   = large_size;
+      if ( trail    == 0 ) trail    = large_size;
+      if ( argument == 0 ) argument = large_size;
+    }
+  }
 
 #define INIT_STACK(name, print, limit, minsize) \
   DEBUG(1, Sdprintf("%s stack at 0x%x; size = %ld\n", print, base, limit)); \
-  init_stack((Stack) &stacks.name, print, (caddress) base, limit, minsize); \
+  init_stack((Stack) &LD->stacks.name, print, (caddress) base, limit, minsize); \
   base += limit + STACK_SEPARATION; \
   base = align_base(base);
 #define K * 1024
@@ -1143,6 +1281,8 @@ initStacks(long local, long global, long trail, long argument)
   INIT_STACK(trail,    "trail",    trail,    8 K);
   INIT_STACK(argument, "argument", argument, 1 K);
 
+  assert(top == 0L || (ulong)aLimit <= top);
+
 #ifndef NO_SEGV_HANDLING
   pl_signal(SIGSEGV, (handler_t) segv_handler);
 #endif
@@ -1150,18 +1290,12 @@ initStacks(long local, long global, long trail, long argument)
 
 void
 resetStacks()
-{ environment_frame = NULL;
-  fli_context       = NULL;
-  lTop = lBase;
-  tTop = tBase;
-  gTop = gBase;
-  aTop = aBase;
+{ emptyStacks();
 
 #ifndef NO_SEGV_HANDLING
   pl_signal(SIGSEGV, (handler_t) segv_handler);
 #endif
   trimStacks();
-  PL_open_foreign_frame();
 }
 
 
@@ -1173,8 +1307,8 @@ resetStacks()
 
 static void
 gcPolicy(Stack s, int policy)
-{ s->gc = ((s == (Stack) &stacks.global ||
-	    s == (Stack) &stacks.trail) ? TRUE : FALSE);
+{ s->gc = ((s == (Stack) &LD->stacks.global ||
+	    s == (Stack) &LD->stacks.trail) ? TRUE : FALSE);
   if ( s->gc )
   { s->small  = SMALLSTACK;
     s->factor = 3;
@@ -1191,8 +1325,8 @@ word
 pl_trim_stacks()
 { trimStacks();
 
-  gcPolicy((Stack) &stacks.global, GC_FAST_POLICY);
-  gcPolicy((Stack) &stacks.trail,  GC_FAST_POLICY);
+  gcPolicy((Stack) &LD->stacks.global, GC_FAST_POLICY);
+  gcPolicy((Stack) &LD->stacks.trail,  GC_FAST_POLICY);
 
   succeed;
 }
@@ -1230,13 +1364,13 @@ pl_stack_parameter(term_t name, term_t key, term_t old, term_t new)
 
   if ( PL_get_atom(name, &a) )
   { if ( a == ATOM_local )
-      stack = (Stack) &stacks.local;
+      stack = (Stack) &LD->stacks.local;
     else if ( a == ATOM_global )
-      stack = (Stack) &stacks.global;
+      stack = (Stack) &LD->stacks.global;
     else if ( a == ATOM_trail )
-      stack = (Stack) &stacks.trail;
+      stack = (Stack) &LD->stacks.trail;
     else if ( a == ATOM_argument )
-      stack = (Stack) &stacks.argument;
+      stack = (Stack) &LD->stacks.argument;
   }
   if ( !stack )
     return warning("stack_parameter/4: unknown stack");
@@ -1265,8 +1399,8 @@ init_stack(Stack s, char *name, long size, long limit, long minfree)
   s->minfree	= minfree;
   s->max	= (char *)s->base + size;
   s->gced_size = 0L;			/* size after last gc */
-  s->gc	       = ((s == (Stack) &stacks.global ||
-		   s == (Stack) &stacks.trail) ? TRUE : FALSE);
+  s->gc	       = ((s == (Stack) &LD->stacks.global ||
+		   s == (Stack) &LD->stacks.trail) ? TRUE : FALSE);
   s->small     = (s->gc ? SMALLSTACK : 0);
 }
 
@@ -1279,12 +1413,12 @@ not exist.
 #if tos
 #define MALLOC(p, n)    Allocate(n)
 #else
-#define MALLOC(p, n)	(status.dumped == FALSE ? Malloc(n) : Realloc(p, n))
+#define MALLOC(p, n)	(!GD->dumped ? malloc(n) : realloc(p, n))
 #endif
 
 static void
-initStacks(long local, long global, long trail, long argument)
-{ long old_heap = statistics.heap;
+initStacks(long local, long global, long trail, long arg)
+{ long old_heap = GD->statistics.heap;
 #if O_SHIFT_STACKS
   long itrail  = 32 K;
   long iglobal = 200 K;
@@ -1300,26 +1434,19 @@ initStacks(long local, long global, long trail, long argument)
 			       MAXARITY * sizeof(word));
   lBase = (LocalFrame)	addPointer(gBase, iglobal+sizeof(word));
   tBase = (TrailEntry)	MALLOC(tBase, itrail);
-  aBase = (Word *)	MALLOC(aBase, argument);
+  aBase = (Word *)	MALLOC(aBase, arg);
 
-  init_stack((Stack)&stacks.global,   "global",	 iglobal,  global,   100 K);
-  init_stack((Stack)&stacks.local,    "local",	 ilocal,   local,    16 K);
-  init_stack((Stack)&stacks.trail,    "trail",	 itrail,   trail,    8 K);
-  init_stack((Stack)&stacks.argument, "argument",argument, argument, 0);
+  init_stack((Stack)&LD->stacks.global,	  "global",   iglobal, global, 100 K);
+  init_stack((Stack)&LD->stacks.local,    "local",    ilocal,  local,   16 K);
+  init_stack((Stack)&LD->stacks.trail,    "trail",    itrail,  trail,    8 K);
+  init_stack((Stack)&LD->stacks.argument, "argument", arg,     arg,      0 K);
 
-  statistics.heap = old_heap;
+  GD->statistics.heap = old_heap;
 }
 
 void
 resetStacks()
-{ environment_frame = NULL;
-  fli_context       = NULL;
-  lTop = lBase;
-  tTop = tBase;
-  gTop = gBase;
-  aTop = aBase;
-
-  PL_open_foreign_frame();
+{ emptyStacks();
 }
 
 #endif /* O_DYNAMIC_STACKS */
@@ -1328,12 +1455,12 @@ void
 trimStacks()
 {
 #ifdef O_DYNAMIC_STACKS
-  unmap((Stack) &stacks.local);
-  unmap((Stack) &stacks.global);
-  unmap((Stack) &stacks.trail);
-  unmap((Stack) &stacks.argument);
+  unmap((Stack) &LD->stacks.local);
+  unmap((Stack) &LD->stacks.global);
+  unmap((Stack) &LD->stacks.trail);
+  unmap((Stack) &LD->stacks.argument);
 #endif /*O_DYNAMIC_STACKS*/
 
-  stacks.global.gced_size = usedStack(global);
-  stacks.trail.gced_size  = usedStack(trail);
+  LD->stacks.global.gced_size = usedStack(global);
+  LD->stacks.trail.gced_size  = usedStack(trail);
 }
