@@ -1,4 +1,4 @@
-/*  $Id: pl-stream.c,v 1.18 1997/03/10 09:37:08 jan Exp $
+/*  $Id: pl-stream.c,v 1.22 1997/11/12 14:21:19 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -55,6 +55,8 @@ int Slinesize = SIO_LINESIZE;		/* Sgets() buffer size */
 
 static int	S__flushbuf(IOSTREAM *s);
 
+#define S__fupdatefilepos(s, c) S___fupdatefilepos(s, c)
+
 		 /*******************************
 		 *	      BUFFER		*
 		 *******************************/
@@ -226,8 +228,8 @@ S__fillbuf(IOSTREAM *s)
 		 *******************************/
 
 
-__inline int
-S__fupdatefilepos(IOPOS *p, int c)
+inline int
+S___fupdatefilepos(IOPOS *p, int c)
 { if ( p )
   { switch(c)
     { case '\n':
@@ -256,7 +258,6 @@ S__fupdatefilepos(IOPOS *p, int c)
   return c;
 }
 
-
 int
 Sputc(int c, IOSTREAM *s)
 { c &= 0xff;
@@ -265,8 +266,11 @@ Sputc(int c, IOSTREAM *s)
     *s->bufp++ = (char)c;
   else
   { if ( S__flushbufc(c, s) < 0 )
+    { s->lastc = EOF;
       return -1;
+    }
   }
+  s->lastc = c;
 
   if ( c == '\n' && (s->flags & SIO_LBUF) )
   { if ( S__flushbuf(s) < 0 )
@@ -1125,25 +1129,33 @@ out:
 
 static int
 Sread_file(void *handle, char *buf, int size)
-{ return read((int)handle, buf, size);
+{ long h = (long) handle;
+
+  return read((int)h, buf, size);
 }
 
 
 static int
 Swrite_file(void *handle, char *buf, int size)
-{ return write((int)handle, buf, size);
+{ long h = (long) handle;
+
+  return write((int)h, buf, size);
 }
 
 
 static long
 Sseek_file(void *handle, long pos, int whence)
-{ return lseek((int)handle, pos, whence);
+{ long h = (long) handle;
+
+  return lseek((int)h, pos, whence);
 }
 
 
 static int
 Sclose_file(void *handle)
-{ return close((int) handle);
+{ long h = (long) handle;
+
+  return close((int) h);
 }
 
 
@@ -1165,6 +1177,7 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
   }
   memset((char *)s, 0, sizeof(IOSTREAM));
   s->magic     = SIO_MAGIC;
+  s->lastc     = EOF;
   s->flags     = flags;
   s->handle    = handle;
   s->functions = functions;
@@ -1186,6 +1199,7 @@ Sopen_file(const char *path, char *how)
 { int fd;
   int oflags = 0, flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS;
   int op = *how++;
+  long lfd;
 
   for( ; *how; how++)
   { switch(*how)
@@ -1227,13 +1241,15 @@ Sopen_file(const char *path, char *how)
   if ( fd < 0 )
     return NULL;
 
-  return Snew((void *)fd, flags, &Sfilefunctions);
+  lfd = (long)fd;
+  return Snew((void *)lfd, flags, &Sfilefunctions);
 }
 
 
 IOSTREAM *
 Sfdopen(int fd, char *type)
 { int flags;
+  long lfd;
 
   if ( fd < 0 )
     return NULL;
@@ -1243,14 +1259,18 @@ Sfdopen(int fd, char *type)
   else
     flags = SIO_FILE|SIO_OUTPUT;
 
-  return Snew((void *)fd, flags, &Sfilefunctions);
+  lfd = (long)fd;
+
+  return Snew((void *)lfd, flags, &Sfilefunctions);
 }
 
 
 int
 Sfileno(IOSTREAM *s)
 { if ( s->flags & SIO_FILE )
-    return (int)s->handle;
+  { long h = (long)s->handle;
+    return (int)h;
+  }
   if ( s->flags & SIO_PIPE )
     return fileno((FILE *)s->handle);
 
@@ -1260,7 +1280,7 @@ Sfileno(IOSTREAM *s)
 
 
 #define STDIO(n, f) { NULL, NULL, NULL, NULL, \
-		      SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
+		      SIO_MAGIC, EOF, 0, f, {0, 0, 0}, NULL, \
 		      ((void *)(n)), &Sfilefunctions \
 		    }
 
@@ -1601,3 +1621,10 @@ Sopen_string(IOSTREAM *s, char *buf, int size, char *mode)
 
   return s;
 }
+
+#undef S__fupdatefilepos
+
+int
+S__fupdatefilepos(IOPOS *p, int c)
+{ return S___fupdatefilepos(p, c);
+}
