@@ -1,4 +1,4 @@
-/*  $Id: pl-stream.h,v 1.11 1997/03/10 09:37:09 jan Exp $
+/*  $Id: pl-stream.h,v 1.13 1997/11/12 14:21:21 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -83,6 +83,7 @@ typedef struct io_stream
   char		       *limitp;		/* read/write limit */
   char		       *buffer;		/* the buffer */
   char		       *unbuffer;	/* Sungetc buffer */
+  int			lastc;		/* last character written */
   int			magic;		/* magic number SIO_MAGIC */
   int  			bufsize;	/* size of the buffer */
   int			flags;		/* Status flags */
@@ -135,7 +136,7 @@ __pl_export_data IOSTREAM    S__iob[];		/* Libs standard streams */
 	((s)->position ? S__fupdatefilepos((s)->position, (c)) \
 		       : (c))
 
-#define Snpgetc(s) ((s)->bufp < (s)->limitp ? *(s)->bufp++&0xff \
+#define Snpgetc(s) ((s)->bufp < (s)->limitp ? (int)(*(s)->bufp++)&0xff \
 					    : S__fillbuf(s))
 #define Sgetc(s) S__updatefilepos((s), Snpgetc(s))
 
