@@ -1,4 +1,4 @@
-/*  $Id: pl-trace.c,v 1.35 1997/03/10 09:37:11 jan Exp $
+/*  $Id: pl-trace.c,v 1.39 1997/08/26 10:09:14 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -17,11 +17,23 @@ int trace_continuation;			/* how to continue? */
 #define W_WRITEQ	3		/* writeq/1 */
 #define W_DISPLAY	4		/* display/1 */
 
-static struct
+#define TRACE_FIND_NONE	0
+#define TRACE_FIND_ANY	1
+#define TRACE_FIND_NAME	2
+#define TRACE_FIND_TERM	3
+
+typedef struct find_data_tag
 { int	 port;				/* Port to find */
   bool	 searching;			/* Currently searching? */
-  Record goal;				/* Goal to find */
-} find;
+  int	 type;				/* TRACE_FIND_* */
+  union
+  { atom_t	name;			/* Name of goal to find */
+    struct
+    { functor_t	functor;		/* functor of the goal */
+      Record	term;			/* Goal to find */
+    } term;
+  } goal;
+} find_data;
 
 #define PrologRef(fr)	 ((Word)fr - (Word)lBase)
 #define FrameRef(w)	 ((LocalFrame)((Word)lBase + w))
@@ -46,7 +58,6 @@ forwards bool		hasAlternativesFrame(LocalFrame);
 forwards void		alternatives(LocalFrame);
 forwards void		listProcedure(Definition);
 forwards int		traceInterception(LocalFrame, LocalFrame, int, Code);
-forwards bool		canUnifyTermWithGoal(Word, LocalFrame);
 forwards void		writeFrameGoal(LocalFrame frame, int how);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -73,29 +84,47 @@ as the record is not in the proper format.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static bool
-canUnifyTermWithGoal(Word t, LocalFrame fr)
-{ deRef(t);
-  if ( isVar(*t) )
-    succeed;
-  if ( isAtom(*t) && fr->predicate->functor->name == *t )
-    succeed;
-  if ( hasFunctor(*t, fr->predicate->functor) )
-  { Word a, b;
-    int arity;
+canUnifyTermWithGoal(LocalFrame fr)
+{ find_data *find = LD->trace.find;
 
-    a = argTermP(*t, 0);
-    b = argFrameP(fr, 0);
-    arity = functorTerm(*t)->arity;
-    while( arity > 0 )
-    { if ( !can_unify(a, b) )
-        fail;
+  switch(find->type)
+  { case TRACE_FIND_ANY:
+      succeed;
+    case TRACE_FIND_NAME:
+      return find->goal.name == fr->predicate->functor->name;
+    case TRACE_FIND_TERM:
+    { if ( find->goal.term.functor == fr->predicate->functor->functor )
+      { fid_t cid = PL_open_foreign_frame();
+	term_t t = PL_new_term_ref();
+	Word a, b;
+	int arity = fr->predicate->functor->arity;
+	int rval = TRUE;
+
+	copyRecordToGlobal(t, find->goal.term.term);
+	a = valTermRef(t);
+	deRef(a);
+	a = argTermP(*a, 0);
+	b = argFrameP(fr, 0);
+	while( arity-- > 0 )
+	{ if ( !can_unify(a++, b++) )
+	  { rval = FALSE;
+	    break;
+	  }
+	}
+
+	PL_discard_foreign_frame(cid);
+	return rval;
+      }
+
+      fail;
     }
-    succeed;
+    default:
+      assert(0);
+      fail;
   }
-  
-  fail;
 }
 
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Toplevel  of  the  tracer.   This  function  is  called  from  the   WAM
 interpreter.   It  can  take  care of most of the tracer actions itself,
@@ -153,7 +182,8 @@ tracePort(LocalFrame frame, LocalFrame bfr, int port, Code PC)
     }
   }
 
-  if ( port != BREAK_PORT &&
+  if ( (port != BREAK_PORT) &&
+       /*(port != EXCEPTION_PORT) &&*/
        ((!debugstatus.tracing &&
 	 (false(def, SPY_ME) || (port & CUT_PORT)))	|| /* non-tracing */
 	debugstatus.skiplevel < levelFrame(frame)	|| /* skipped */
@@ -194,11 +224,11 @@ Give a trace on the skipped goal for a redo.
 We are in searching mode; should we actually give this port?
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-  if ( find.searching )
+  if ( LD->trace.find &&  LD->trace.find->searching )
   { DEBUG(2, Sdprintf("Searching\n"));
 
-    if ( (port & find.port) && canUnifyTermWithGoal(&find.goal->term, frame) )
-    { find.searching = FALSE;		/* Got you */
+    if ( (port & LD->trace.find->port) && canUnifyTermWithGoal(frame) )
+    { LD->trace.find->searching = FALSE; /* Got you */
     } else
     { return ACTION_CONTINUE;		/* Continue the search */
     }
@@ -225,14 +255,15 @@ again:
   Put( true(def, TRANSPARENT) ? '^' : ' ');
 
   switch(port)
-  { case CALL_PORT:	Putf(" Call:  ");	break;
-    case REDO_PORT:	Putf(" Redo:  ");	break;
-    case FAIL_PORT:	Putf(" Fail:  ");	break;
-    case EXIT_PORT:	Putf(" Exit:  ");	break;
-    case UNIFY_PORT:	Putf(" Unify: ");	break;
-    case BREAK_PORT:	Putf(" Break: ");	break;
-    case CUT_CALL_PORT:	Putf(" Cut call: ");	break;
-    case CUT_EXIT_PORT:	Putf(" Cut exit: ");	break;
+  { case CALL_PORT:	 Putf(" Call:  ");	break;
+    case REDO_PORT:	 Putf(" Redo:  ");	break;
+    case FAIL_PORT:	 Putf(" Fail:  ");	break;
+    case EXIT_PORT:	 Putf(" Exit:  ");	break;
+    case UNIFY_PORT:	 Putf(" Unify: ");	break;
+    case BREAK_PORT:	 Putf(" Break: ");	break;
+    case EXCEPTION_PORT: Putf(" Exception: ");	break;
+    case CUT_CALL_PORT:	 Putf(" Cut call: ");	break;
+    case CUT_EXIT_PORT:	 Putf(" Cut exit: ");	break;
   }
   Putf("(%3ld) ", levelFrame(frame));
   writeFrameGoal(frame, debugstatus.style);
@@ -245,7 +276,7 @@ again:
 
     Putf(" ? ");
     pl_flush();
-    if ( status.notty )
+    if ( GD->cmdline.notty )
     { buf[0] = EOS;
       readLine(buf);
     } else
@@ -256,7 +287,7 @@ again:
 	readLine(buf);
       }
     }
-    if ((action = traceAction(buf, port, frame, status.notty ? FALSE : TRUE))
+    if ((action = traceAction(buf, port, frame, GD->cmdline.notty ? FALSE : TRUE))
 							== ACTION_AGAIN)
       goto again;
   } else
@@ -276,11 +307,11 @@ setupFind(char *buf)
   for(s = buf; *s && isBlank(*s); s++)	/* Skip blanks */
     ;
   if ( *s == EOS )			/* No specification: repeat */
-  { if ( find.port == 0 )
+  { if ( !LD->trace.find || !LD->trace.find->port )
     { Putf("[No previous search]\n");
       fail;
     }
-    find.searching = TRUE;
+    LD->trace.find->searching = TRUE;
     succeed;
   }
   for( ; *s && !isBlank(*s); s++ )	/* Parse the port specification */
@@ -301,11 +332,16 @@ setupFind(char *buf)
 
   if ( *s == EOS )			/* Nothing is a variable */
   { s = buf;
-    strcpy(buf, "_");
+    buf[0] = '_',
+    buf[1] = EOS;
   }
 
   { fid_t cid = PL_open_foreign_frame();
     term_t t = PL_new_term_ref();
+    FindData find;
+
+    if ( !(find = LD->trace.find) )
+      find = LD->trace.find = allocHeap(sizeof(find_data));
 
     seeString(s);
     rval = pl_read(t);
@@ -316,11 +352,23 @@ setupFind(char *buf)
       fail;
     }
 
-    if ( find.goal )
-      freeRecord(find.goal);
-    find.port      = port;
-    find.goal      = copyTermToHeap(t);
-    find.searching = TRUE;
+    if ( find->type == TRACE_FIND_TERM && find->goal.term.term )
+      freeRecord(find->goal.term.term);
+
+    if ( PL_is_variable(t) )
+    { find->type = TRACE_FIND_ANY;
+    } else if ( PL_get_atom(t, &find->goal.name) )
+    { find->type = TRACE_FIND_NAME;
+    } else if ( PL_get_functor(t, &find->goal.term.functor) )
+    { find->type = TRACE_FIND_TERM;
+      find->goal.term.term    = compileTermToHeap(t);
+    } else
+    { Putf("[Illegal goal specification]\n");
+      fail;
+    }
+
+    find->port      = port;
+    find->searching = TRUE;
 
     DEBUG(2, Sdprintf("setup ok, port = 0x%x, goal = ", port);
 	  pl_write(t);
@@ -368,9 +416,10 @@ traceAction(char *cmd, int port, LocalFrame frame, bool interactive)
 		  return ACTION_CONTINUE;
 		}
 		return ACTION_AGAIN;    		
-    case '.':   if ( find.goal != NULL )
+    case '.':   if ( LD->trace.find &&
+		     LD->trace.find->type != TRACE_FIND_NONE )
       	        { FeedBack("repeat search\n");
-		  find.searching = TRUE;
+		  LD->trace.find->searching = TRUE;
 		  clear(frame, FR_SKIPPED);
 		  return ACTION_CONTINUE;
 		} else
@@ -395,7 +444,7 @@ traceAction(char *cmd, int port, LocalFrame frame, bool interactive)
 		} else
 		  Warn("Can't ignore goal at this port\n");
 		return ACTION_CONTINUE;
-    case 'r':	if (port & (REDO_PORT|FAIL_PORT|EXIT_PORT))
+    case 'r':	if (port & (REDO_PORT|FAIL_PORT|EXIT_PORT|EXCEPTION_PORT))
 		{ FeedBack("retry\n[retry]\n");
 		  return ACTION_RETRY;
 		} else
@@ -414,7 +463,7 @@ traceAction(char *cmd, int port, LocalFrame frame, bool interactive)
     case 'p':	FeedBack("print\n");
 		debugstatus.style = W_PRINT;
 		return ACTION_AGAIN;
-    case 'd':	FeedBack("display\n");
+    case 'd':	FeedBack("write canonical\n");
 		debugstatus.style = W_DISPLAY;
 		return ACTION_AGAIN;
     case 'l':	FeedBack("leap\n");
@@ -449,7 +498,7 @@ traceAction(char *cmd, int port, LocalFrame frame, bool interactive)
     case '?': 
     case 'h':	helpTrace();
 		return ACTION_AGAIN;
-    case 'D':   status.debugLevel = num_arg;
+    case 'D':   GD->debug_level = num_arg;
 		FeedBack("Debug level\n");
 		return ACTION_AGAIN;
     default:	Warn("Unknown option (h for help)\n");
@@ -514,7 +563,7 @@ writeFrameGoal(LocalFrame frame, int how)
        (false(def->module, SYSTEM) || SYSTEM_MODE))
     Putf("%s:", stringAtom(def->module->name));
 
-  PL_unify_functor(goal, def->functor);
+  PL_unify_functor(goal, def->functor->functor);
   if ( argc > 0 )
   { Word argp = valTermRef(goal);
     int i;
@@ -533,7 +582,7 @@ writeFrameGoal(LocalFrame frame, int how)
   switch(how)
   { case W_PRINT:
 	debugstatus.debugging = FALSE;
-	if ( status.boot )
+	if ( GD->bootsession )
 	  pl_write(goal);
 	else
 	  pl_print(goal);
@@ -546,7 +595,7 @@ writeFrameGoal(LocalFrame frame, int how)
 	pl_writeq(goal);
 	break;
     case W_DISPLAY:
-	pl_display(goal);
+	pl_write_canonical(goal);
 	break;
   }
 
@@ -579,9 +628,9 @@ listProcedure(Definition def)
   term_t argv = PL_new_term_refs(1);
   Procedure proc = lookupProcedure(FUNCTOR_listing1, MODULE_system);
 
-  unify_definition(argv, def, 0);	/* module:name(args) */
+  unify_definition(argv, def, 0, 0);	/* module:name(args) */
   Output = 1;
-  qid = PL_open_query(MODULE_user, FALSE, proc, argv);
+  qid = PL_open_query(MODULE_user, PL_Q_NODEBUG, proc, argv);
   PL_next_solution(qid);
   PL_close_query(qid);
   Output = OldOut;
@@ -657,31 +706,32 @@ This predicate is supposed to return one of the following atoms:
 static int
 traceInterception(LocalFrame frame, LocalFrame bfr, int port, Code PC)
 { int rval = -1;			/* Default C-action */
-  static predicate_t proc;
+  predicate_t proc;
 
-  if ( !proc )
-    proc = PL_predicate("prolog_trace_interception", 4, "user");
+  proc = _PL_predicate("prolog_trace_interception", 4, "user",
+		       &GD->procedures.prolog_trace_interception4);
   if ( !proc->definition->definition.clauses )
     return rval;
 
-  if ( !status.boot && status.debugLevel == 0 )
+  if ( !GD->bootsession && GD->debug_level == 0 )
   { fid_t cid = PL_open_foreign_frame();
     qid_t qid;
     term_t argv = PL_new_term_refs(4);
     term_t rarg = argv+3;
     atom_t portname = NULL_ATOM;
-    functor_t portfunc = NULL;
+    functor_t portfunc = 0;
     int nodebug = FALSE;
 
     switch(port)
-    { case CALL_PORT:	  portname = ATOM_call;		break;
-      case REDO_PORT:	  portname = ATOM_redo;		break;
-      case EXIT_PORT:	  portname = ATOM_exit;		break;
-      case FAIL_PORT:	  portname = ATOM_fail;		break;
-      case UNIFY_PORT:	  portname = ATOM_unify;	break;
-      case BREAK_PORT:    portfunc = FUNCTOR_break1;	break;
-      case CUT_CALL_PORT: portfunc = FUNCTOR_cut_call1; break;
-      case CUT_EXIT_PORT: portfunc = FUNCTOR_cut_exit1; break;
+    { case CALL_PORT:	   portname = ATOM_call;         break;
+      case REDO_PORT:	   portname = ATOM_redo;         break;
+      case EXIT_PORT:	   portname = ATOM_exit;         break;
+      case FAIL_PORT:	   portname = ATOM_fail;         break;
+      case UNIFY_PORT:	   portname = ATOM_unify;	 break;
+      case EXCEPTION_PORT: portname = ATOM_exception; 	 break;
+      case BREAK_PORT:     portfunc = FUNCTOR_break1;	 break;
+      case CUT_CALL_PORT:  portfunc = FUNCTOR_cut_call1; break;
+      case CUT_EXIT_PORT:  portfunc = FUNCTOR_cut_exit1; break;
       default:
 	assert(0);
         return rval;
@@ -706,7 +756,7 @@ traceInterception(LocalFrame frame, LocalFrame bfr, int port, Code PC)
     PL_put_integer(argv+2, PrologRef(bfr));
     PL_put_variable(rarg);
 
-    qid = PL_open_query(MODULE_user, FALSE, proc, argv);
+    qid = PL_open_query(MODULE_user, PL_Q_NODEBUG, proc, argv);
     if ( PL_next_solution(qid) )
     { atom_t a;
 
@@ -794,7 +844,7 @@ interruptHandler(int sig)
   LocalFrame oldltop = lTop;
   Char c; 
 
-  if ( status.initialised == FALSE )
+  if ( !GD->initialised )
   { Sfprintf(Serror, "Interrupt during startup. Cannot continue\n");
     Halt(1);
   }  
@@ -889,12 +939,12 @@ initTracer(void)
 
   debugstatus.visible      = 
   debugstatus.leashing     = CALL_PORT|FAIL_PORT|REDO_PORT|EXIT_PORT|
-			     BREAK_PORT;
+			     BREAK_PORT|EXCEPTION_PORT;
   debugstatus.tracing      =
   debugstatus.debugging    = FALSE;
   debugstatus.suspendTrace = FALSE;
   debugstatus.skiplevel    = 0;
-  debugstatus.style        = status.boot ? W_WRITE : W_PRINT; 
+  debugstatus.style        = GD->bootsession ? W_WRITE : W_PRINT; 
   debugstatus.showContext  = FALSE;
   debugstatus.retryFrame   = NULL;
 }
@@ -917,7 +967,8 @@ tracemode(int doit, int *old)
   { if ( doit )
     { debugstatus.debugging = TRUE;
       debugstatus.skiplevel = VERY_DEEP;
-      find.searching = FALSE;
+      if ( LD->trace.find )
+	LD->trace.find->searching = FALSE;
     }
     debugstatus.tracing = doit;
     callEventHook(PLEV_TRACING, doit);
@@ -1052,7 +1103,7 @@ pl_visible(term_t old, term_t new)
 
 word
 pl_debuglevel(term_t old, term_t new)
-{ return setInteger(&status.debugLevel, "$visible", old, new);
+{ return setInteger(&GD->debug_level, "$debuglevel", old, new);
 }
 
 
@@ -1127,7 +1178,7 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
 
 #ifdef O_DEBUGLOCAL			/* see pl-wam.c */
     assert( *argFrameP(fr, argn-1) != (word)(((char*)ATOM_nil) + 1) );
-    checkData(argFrameP(fr, argn-1), FALSE);
+    checkData(argFrameP(fr, argn-1));
 #endif
 
    deRef(p);
@@ -1186,7 +1237,7 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
     } else
     { term_t a = PL_new_term_ref();
 
-      PL_unify_functor(arg, fr->predicate->functor);
+      PL_unify_functor(arg, fr->predicate->functor->functor);
       for(n=0; n < arity; n++)
       { PL_get_arg(n+1, arg, a);
 	unify_ptrs(valTermRef(a), argFrameP(fr, n));
